// Copyright (C) 2010 By Paul Dunn
//
// This file is part of the SpecBAS BASIC Interpreter, which is in turn
// part of the SpecOS project.
//
// SpecBAS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// SpecBAS is distributed in the hope that it will be entertaining,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SpecBAS.  If not, see <http://www.gnu.org/licenses/>.

unit SP_Main;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}
{$INCLUDE SpecBAS.inc}

{

  The main loop of the interpreter and editor. This should be run as a low-priority
  thread separate from the main app (vcl) thread. This is to allow a 50hz (20ms)
  keyboard poll via the message queue.

}

interface

Uses

  Math, Classes, SysUtils, SP_Graphics32,
  SP_Graphics, SP_SysVars, SP_Input, SP_Errors, SP_Util, SP_Tokenise,
  SP_InfixToPostFix, SP_Interpret_PostFix, SP_Variables, SP_BankManager,
  SP_FileIO, SP_Sound, SP_Editor, SP_Package{$IFNDEF FPC}, JclAnsiStrings{$ENDIF};

Type

  TCB_YieldProc = Procedure;
  TCB_QuitProc = Procedure;
  TCB_GetTicks = Function: LongWord;
  TCB_GetAxis = Function(NubID, Axis: Integer): Integer;
  TCB_InitSticks = Procedure;
  TCB_ReleaseSticks = Procedure;

Procedure SP_FrameUpdate;
Procedure DoAutoSave(Var Error: TSP_ErrorCode);
Procedure SP_MainLoop;
Procedure SP_CleanUp;
Procedure SP_Display_Cursor;
Procedure SP_ListProg(AutoList: Boolean; StartLine: Integer; var Error: TSP_ErrorCode);
Function  SP_TestScroll(Height: Integer; var Error: TSP_ErrorCode): Boolean;
Function  SP_FindLine(LineNum: Integer; Exact: Boolean): Integer;
Function  SP_FindStatement(Tokens: paString; Statement: Integer): Integer;
Procedure SP_StackLine(Line, Statement, St, Source: Integer);
Function  SP_ConvertLineStatement(Line, Statement: Integer): TSP_GOSUB_Item;
Procedure SP_BuildLabelList;
Procedure SP_AddLabel(const LabelText: aString; Line, Statement, St, DLine, DStatement, Dst: Integer);
Function  SP_FindLabel(const LabelText: aString; Var Error: TSP_ErrorCode): TSP_Label;
Procedure SP_ClearLabels;
Procedure SP_MergeLines(Start, Finish: Integer; var Error: TSP_ErrorCode);
Procedure SP_Renumber(Start, Finish, Line, Step: Integer; var Error: TSP_ErrorCode);
Procedure SP_DeleteLines(Start, Finish: Integer; var Error: TSP_ErrorCode);
Function  SP_LIST(Start, Finish: Integer): aString;

Var

  CB_Yield: TCB_YieldProc;
  CB_Quit: TCB_QuitProc;
  CB_GetTicks: TCB_GetTicks;
  CB_GetAxis: TCB_GetAxis;
  CB_InitSticks: TCB_InitSticks;
  CB_ReleaseSticks: TCB_ReleaseSticks;
  SP_GOSUB_Stack: Array of TSP_GOSUB_Item;
  SP_GOSUB_StackPtr,
  SP_GOSUB_StackLen: Integer;
  SP_LabelList: Array of TSP_Label;
  SP_Interpreter_Ready: Boolean = False;
  CauseUpdate: Boolean = False;
  FrameElapsed: Boolean = False;
  TerminateInterpreter: Boolean = False;
  SP_KeyWordID: LongWord;

implementation

Procedure SP_FrameUpdate;
Begin

  // Changes the FLASH sysvar once every 16 frames.

  If FRAMES Mod 16 = 0 Then Begin
    FLASHSTATE := 1 - FLASHSTATE;
    SP_NeedDisplayUpdate := True;
  End Else
    If NUMSPRITES > 0 Then
      SP_NeedDisplayUpdate := True;

  If SP_NeedDisplayUpdate Then Begin
    If Not SCREENLOCK And (DisableCount = 0) Then Begin
      CauseUpdate := True;
      SP_NeedDisplayUpdate := False;
    End Else
      SP_NeedDisplayUpdate := False;
  End;

  FrameElapsed := True;

End;

Procedure DoAutoSave(Var Error: TSP_ErrorCode);
Begin
  If AUTOSAVE And (Error.Code = SP_ERR_OK) Then Begin
    SP_SaveProgram('s:autosave', -1, Error);
    Error.Code := SP_ERR_OK;
  End;
End;

Procedure SP_MainLoop;
Var
  Error: TSP_ErrorCode;
Begin

  HARDWARE := 'PC Windows/x86';

  {$IFDEF FPC}
  HARDWARE := 'PC Linux/x86';
  {$ENDIF}

  {$IFDEF PANDORA}
  HARDWARE := 'Pandora Linux/ARM';
  {$ENDIF}

  {$IFDEF RASPI}
  HARDWARE := 'Pi Linux/ARM';
  {$ENDIF}

  FPS := 50;
  FRAME_MS := 20;

  SetLength(SP_Program, 0);
  SP_Program_Count := 0;
  SetLength(SP_GOSUB_Stack, 0);
  SP_GOSUB_STACKPTR := 0;
  SP_GOSUB_STACKLEN := 0;
  SP_PreParse(True, Error);

  MATHMODE := 0;

  Randomize;

  FLASHSTATE := 0;
  FRAMES := 0;
  LASTFRAME := FRAMES;
  REPDEL := 20;
  REPPER := 2;
  INSERT := True;
  CURSORX := -FONTWIDTH;
  CURSORCHAR := 32;
  PRPOSX := 0;
  PRPOSY := 0;
  DRPOSX := 0;
  DRPOSY := 0;
  DRHEADING := 0;
  TABSIZE := 3;
  SHOWLINE := 0;
  PROGLINE := 0;
  QUITMSG := FALSE;
  LISTWINDOW := -1;
  COMMANDWINDOW := -1;
  PROGNAME := 'New program';
  FILENAMED := False;
  FDIST := 300;
  EDFONTSCALEX := 1;
  EDFONTSCALEY := 1;

  ERROR_LineNum := -1;

  // Set up the initial copyright message.

  SP_Interpreter_Ready := True;

  SP_CLS(CPAPER);

  SetLength(SP_Ass_List, 6);
  SP_Ass_List[0] := 'SYS'#255'/';
  SP_Ass_List[1] := 'TEMP'#255'/temp/';
  SP_ASS_LIST[2] := 'S'#255'/s/';
  SP_ASS_LIST[3] := 'FONTS'#255'/fonts/';
  SP_ASS_LIST[4] := 'KEYBOARDS'#255'/keyboards/';
  SP_ASS_LIST[5] := 'INCLUDE'#255'/include/';

  SP_MakeDir('SYS:s', Error);
  SP_MakeDir('SYS:fonts', Error);
  SP_MakeDir('SYS:keyboards', Error);
  SP_MakeDir('SYS:include', Error);

  SP_MakeSystemSounds;

  SP_SetCurrentDir('SYS:', Error);

  SP_ConvertToOrigin_i(PRPOSX, PRPOSY);
  SP_ConvertToOrigin_i(DRPOSX, DRPOSY);

  SP_EditLoop(Error);

  DoAutoSave(Error);
  AUTOSAVE := False;
  SP_CleanUp;

End;

Procedure SP_CleanUp;
Begin

  SP_DeleteIncludes;
  If PackageIsOpen Then SP_ClosePackage;
  SetLength(SP_Program, 0);
  SP_Program_Count := 0;
  SP_DeleteAllWindows;
  SP_DeleteAllBanks;
  SP_DeleteTempFiles;

End;

Procedure SP_Display_Cursor;
Var
  Fg, Bg: Byte;
  WindowId, Font, PosX, PosY, Over: Integer;
  Err: TSP_ErrorCode;
Begin

  // Repaints the cursor without repainting the whole edit line. Usually called
  // when the flash state changes. Note that CPos is "1" - because this is a one
  // character aString.

  Over := T_OVER;
  T_OVER := 0;

  If FlashState = 1 Then Begin
    Fg := CURSORFG; Bg := CURSORBG;
  End Else Begin
    Fg := CURSORBG; Bg := CURSORFG;
  End;

  PosX := PRPOSX; PosY := PRPOSY;

  If CCOMMANDWINDOW Then Begin

    WindowID := SCREENID;
    Font := FONTBANKID;

    SP_SetSystemFont(EDITORFONT, Err);
    SP_SetDrawingWindow(SCREENBANK, COMMANDWINDOW);

  End Else Begin

    WindowID := 0;
    Font := 0;

  End;

  SP_TEXTOUT(-1, CURSORX, CURSORY, EdSc + aChar(CURSORCHAR), Fg, Bg, True);

  If CCOMMANDWINDOW Then Begin

    SP_SetSystemFont(Font, Err);
    SP_SetDrawingWindow(SCREENBANK, WindowID);

  End;

  PRPOSX := PosX;
  PRPOSY := PosY;

  T_OVER := Over;

  SP_NeedDisplayUpdate := True;

End;

Procedure SP_LISTPROG(AutoList: Boolean; StartLine: Integer; var Error: TSP_ErrorCode);
Var
  Idx, cPos, CharWidth, LineNum, sLine, ProgLen: Integer;
  Tokens, LineNumStr: aString;
  Done: Boolean;
Begin

  // lists the current program in memory to the screen. If listing from a LIST command,
  // it will display the program in its entirety, even if too large to show (uses the "Scroll?"
  // message), but if it's an autolist, will stop at the bottom of the screen.

  SP_CLS(CPAPER);

  // First, find the line we want to list from.

  If INCLUDEFROM = -1 Then
    ProgLen := Length(SP_Program) -1
  Else
    ProgLen := INCLUDEFROM -1;

  Idx := 0;
  While (Idx <= ProgLen) And (pInteger(@SP_Program[Idx][2])^ < StartLine) Do
    Inc(Idx);

  cPos := -1;
  CharWidth := SCREENWIDTH Div FONTWIDTH;
  sLine := Idx;
  If Idx <= ProgLen Then Begin
    Done := False;
    Repeat

      Tokens := SP_Program[Idx];
      LineNum := pLongWord(@Tokens[2])^;
      LineNumStr := IntToString(LineNum);
      Tokens := SP_DeTokenise(Tokens, cPos, False, False);

      // Format the line number - pad with spaces until 5 spaces used.

      For cPos := 1 to 5 - Length(LineNumStr) Do
        Tokens := ' ' + Tokens;

      If LineNum = PROGLINE Then
        Tokens := Copy(Tokens, 1, 5)+'>'+Copy(Tokens, 7, Length(Tokens));

      While Tokens <> '' Do Begin
        SP_PRINT(-1, 0, PRPOSY, -1, Copy(Tokens, 1, CharWidth), CINK, CPAPER, Error);
        Inc(PRPOSY, FONTHEIGHT);
        If PRPOSY > SCREENHEIGHT - (((MAXLOWER +1)* FONTHEIGHT)+4) Then Begin
          If AutoList Then Begin
            If Idx <= SP_FindLine(PROGLINE, True) Then Begin
              SP_Scroll(FONTHEIGHT);
              Dec(PRPOSY, FONTHEIGHT);
              Inc(sLine);
              SHOWLINE := pLongWord(@SP_Program[sLine][2])^;
            End Else Begin
              Done := True;
              Break;
            End;
          End Else Begin
            // Scroll the screen, and loop round to display the line that didn't make it.
            SP_TestScroll(FONTHEIGHT, Error);
          End;
        End;
        Tokens := Copy(Tokens, CharWidth +1, Length(Tokens));
      End;

      Inc(Idx);
      If Idx > ProgLen Then Done := True;

    Until Done;

    PRPOSX := 0;

  End;

End;

Function SP_TestScroll(Height: Integer; var Error: TSP_ErrorCode): Boolean;
Begin

  Inc(SCROLLCNT);
  Result := True;
  CONTLINE := Error.Line;
  CONTSTATEMENT := Error.Statement;

  If SCREENID < 0 Then Exit;

  If SCROLLCNT > (SCREENHEIGHT Div Height) -3 Then Begin

    SCROLLCNT := 0;
    SP_CLS_Lower(CPAPER);
    If SCREENBPP = 8 Then
      SP_TextOut(-1, 2, SCREENHEIGHT - 2 - Height, 'Scroll?', CINK, CPAPER, True)
    Else
      SP_TextOut32(-1, 2, SCREENHEIGHT - 2 - Height, 'Scroll?', CINK, CPAPER, True);

    LASTKEY := 0;
    While (LASTKEY = 0) And not QUITMSG Do
      CB_Yield;

    If QUITMSG Then Exit;
    SP_Restore_Lower;

    If Ord(LASTKEY) in [K_N, K_Space, K_Escape] Then
      Result := False
    Else Begin
      Result := True;
      SP_Scroll(Height);
      CB_Yield;
    End;

  End Else Begin
    SP_Scroll(Height);
    CB_Yield;
  End;

  SP_NeedDisplayUpdate := True;

End;

Function SP_FindLine(LineNum: Integer; Exact: Boolean): Integer;
Begin

  If PROGCHANGED Then Begin
    Result := 0;
    While Result < SP_Program_Count Do Begin
      If Exact Then Begin
        If pInteger(@SP_Program[Result][2])^ = LineNum Then
          Break;
      End Else
        If pInteger(@SP_Program[Result][2])^ >= LineNum Then
          Break;
      Inc(Result);
    End;
    If Result = SP_Program_Count Then
      Result := -1;
  End Else Begin
    If LineNum < Length(LineLUT) Then
      Result := LineLUT[LineNum]
    Else
      Result := -1;
  End;

End;

Function SP_FindStatement(Tokens: paString; Statement: Integer): Integer; Inline;
Var
  NumStatements: Integer;
  Idx: pByte;
Begin

  Idx := pByte(pLongWord(Tokens)^);

  If pByte(Idx)^ = SP_LINE_NUM Then
    Inc(Idx, 1 + SizeOf(LongWord));

  If pByte(Idx)^ = SP_STATEMENTS Then
    Inc(Idx);

  NumStatements := pLongWord(Idx)^;

  If Statement > NumStatements Then
    Result := -1
  Else Begin
    Inc(Idx, (Statement * SizeOf(LongWord)));
    Result := pLongWord(Idx)^;
  End;

End;

Procedure SP_StackLine(Line, Statement, St, Source: Integer);
Begin

  Inc(SP_GOSUB_STACKPTR);
  If SP_GOSUB_STACKPTR >= SP_GOSUB_STACKLEN Then Begin
    Inc(SP_GOSUB_STACKLEN);
    SetLength(SP_GOSUB_STACK, SP_GOSUB_STACKLEN);
  End;

  SP_GOSUB_STACK[SP_GOSUB_STACKPTR -1].Line := Line;
  SP_GOSUB_STACK[SP_GOSUB_STACKPTR -1].Statement := Statement;
  SP_GOSUB_STACK[SP_GOSUB_STACKPTR -1].St := St;
  SP_GOSUB_STACK[SP_GOSUB_STACKPTR -1].Source := Source;

End;

Function SP_ConvertLineStatement(Line, Statement: Integer): TSP_GOSUB_Item;
Begin

  If Line > -1 Then Begin
    If Line < SP_Program_Count Then Begin
      Result.Line := Line;
      Result.Statement := SP_FindStatement(@SP_Program[Line], Statement);
      If Result.Statement = -1 Then Begin
        Inc(Result.Line);
        If Result.Line >= SP_Program_Count Then
          Result.Line := -1
        Else Begin
          Result.Statement := SP_FindStatement(@SP_Program[Result.Line], 1);
          Statement := 1;
          If Result.Line = SP_Program_Count Then
            Result.Line := -1;
        End;
      End;
    End Else Begin
      Result.Line := -1;
      Result.Statement := -1;
    End;
  End Else Begin
    Result.Line := -1;
    Result.Statement := -1;
  End;
  Result.St := Statement;

End;

Procedure SP_AddLabel(const LabelText: aString; Line, Statement, St, DLine, DStatement, Dst: Integer);
Var
  Idx: Integer;
Begin

  Idx := 0;
  While Idx < Length(SP_LabelList) Do Begin
    If SP_LabelList[Idx].Name = Lower(LabelText) Then
      Break;
    Inc(Idx);
  End;

  If Idx = Length(SP_LabelList) Then
    SetLength(SP_LabelList, Length(SP_LabelList)+1);

  If Line = -1 Then Begin
    Line := DLine;
    St := DSt;
    Statement := DStatement;
  End;

  SP_LabelList[Idx].Name := Lower(LabelText);
  SP_LabelList[Idx].Line := Line;
  SP_LabelList[Idx].Statement := Statement;
  SP_LabelList[Idx].St := St;

  SP_LabelList[Idx].DLine := DLine;
  SP_LabelList[Idx].DStatement := DStatement;
  SP_LabelList[Idx].DSt := DSt;

End;

Procedure SP_BuildLabelList;
Var
  Idx, Idx2, pStatement, LabelPos, LabelLen: Integer;
  TempLine, DLine: TSP_GOSUB_Item;
  Tokens: aString;
  Token: pToken;
Begin

  For Idx := 0 To SP_Program_Count -1 Do Begin

    Tokens := SP_Program[Idx];
    pStatement := 1;
    Idx2 := 1;

    If Tokens <> '' Then Begin

      If Tokens[Idx2] = aChar(SP_LINE_NUM) Then Inc(Idx2, 1 + SizeOf(LongWord));
      If Tokens[Idx2] = aChar(SP_STATEMENTS) Then Idx2 := pLongWord(@Tokens[1 + Idx2 + SizeOf(LongWord)])^;

      While Idx2 <= Length(Tokens) Do Begin

        Token := @Tokens[Idx2];
        If Token^.Token = SP_TERMINAL Then Break;
        Inc(Idx2, SizeOf(TToken));
        If Token^.Token = SP_KEYWORD Then Begin
          If pLongWord(@Tokens[Idx2])^ = SP_KW_LABEL Then Begin
            TempLine := SP_ConvertLineStatement(Idx, pStatement +1);
            DLine := SP_ConvertLineStatement(Idx, pStatement);
            SP_AddLabel(Copy(Tokens, LabelPos, LabelLen), TempLine.Line, TempLine.Statement, TempLine.St, DLine.Line, DLine.Statement, DLine.St);
          End;
        End Else
          If Token^.Token = SP_LABEL Then Begin
            LabelPos := Idx2;
            LabelLen := Token^.TokenLen
          End Else
            If Token^.Token = SP_SYMBOL Then
              If Tokens[Idx2] = ':' Then
                Inc(pStatement);

        Inc(Idx2, Token^.TokenLen);

      End;

    End;

  End;

End;

Function SP_FindLabel(const LabelText: aString; Var Error: TSP_ErrorCode): TSP_Label;
Var
  Idx: Integer;
  Token: pToken;
  Statement, LabelPos, LabelLen: Integer;
  nLabelText: aString;
Begin

  Result.Line := -1;
  Result.Dline := -1;

  LabelPos := 0;
  LabelLen := 0;
  nLabelText := LabelText;

  If Error.Line >= 0 Then Begin

    Idx := 0;
    While Idx < Length(SP_LabelList) Do Begin
      If SP_LabelList[Idx].Name = nLabelText Then
        Break;
      Inc(Idx);
    End;
    If Idx = Length(SP_LabelList) Then Begin
      Error.Code := SP_ERR_LABEL_NOT_FOUND;
      Result.Line := -1;
    End Else
      Result := SP_LabelList[Idx];

  End Else Begin

    Idx := 1;
    If COMMAND_TOKENS[Idx] = aChar(SP_LINE_NUM) Then Inc(Idx, 1 + SizeOf(LongWord));
    If COMMAND_TOKENS[Idx] = aChar(SP_STATEMENTS) Then Idx := pLongWord(@COMMAND_TOKENS[1 + Idx + SizeOf(LongWord)])^;
    Statement := 1;

    While Idx <= Length(COMMAND_TOKENS) Do Begin

      Token := @COMMAND_TOKENS[Idx];
      If Token^.Token = SP_TERMINAL Then Break;
      Inc(Idx, SizeOf(TToken));
      Case Token^.Token of
        SP_SYMBOL:
          Begin
            If COMMAND_TOKENS[Idx+1] = ':' Then
              Inc(Statement);
            Inc(Idx, Token^.TokenLen);
          End;
        SP_LABEL:
          Begin
            LabelPos := Idx;
            LabelLen := Token^.TokenLen;
            Inc(Idx, Token^.TokenLen);
          End;
        SP_KEYWORD:
          Begin
            If pLongWord(@COMMAND_TOKENS[Idx])^ = SP_KW_LABEL Then Begin
              If Lower(Copy(COMMAND_TOKENS, LabelPos, LabelLen)) = nLabelText Then Begin
                Result.Line := -2;
                Result.Statement := SP_FindStatement(@COMMAND_TOKENS, Statement +1);
                Result.St := Statement +1;
                Break;
              End;
            End;
            Inc(Idx, Token^.TokenLen);
          End;
      Else
        Inc(Idx, Token^.TokenLen);
      End;
    End;

    If Result.Line = -1 Then Begin

      Idx := 0;
      While Idx < Length(SP_LabelList) Do Begin
        If SP_LabelList[Idx].Name = nLabelText Then
          Break;
        Inc(Idx);
      End;
      If Idx = Length(SP_LabelList) Then
        Result.Line := -1
      Else
        Result := SP_LabelList[Idx];

    End;

  End;

End;

Procedure SP_ClearLabels;
Begin

  SetLength(SP_LabelList, 0);

End;

Procedure SP_MergeLines(Start, Finish: Integer; var Error: TSP_ErrorCode);
Var
  cPos: Integer;
  Idx, LineNum: Integer;
  NeedStart: Boolean;
  NewLine: aString;
Begin

  SP_DeleteIncludes;

  If Finish <= Start Then

    Error.Code := SP_ERR_INTEGER_OUT_OF_RANGE

  Else Begin

    Idx := 0;
    cPos := 0;
    NewLine := '';
    NeedStart := True;

    While Idx < SP_Program_Count Do Begin

      LineNum := pLongWord(@SP_Program[Idx][2])^;
      If (LineNum >= Start) And (LineNum <= Finish) then Begin

        If NeedStart Then Begin
          NewLine := SP_DeTokenise(SP_Program[Idx], cPos, False, False);
          NeedStart := False;
          Inc(Idx);
        End Else Begin
          NewLine := NewLine + ':' + SP_Detokenise(Copy(SP_Program[Idx], 6, Length(SP_Program[Idx])), cPos, False, False);
          SP_Program_Delete(Idx);
        End;

      End Else

        Inc(Idx);

    End;

    NewLine := SP_TokeniseLine(NewLine, False, True) + SP_TERMINAL_SEQUENCE;
    Error.Position := 1;
    SP_Convert_ToPostFix(NewLine, Error.Position, Error);
    PROGLINE := dLongWord(@NewLine[2]);
    SP_MakeEditorList;
    SP_ScrollInView(SP_Store_Line(NewLine));
    PROGCHANGED := True;

  End;

End;

Procedure SP_Renumber(Start, Finish, Line, Step: Integer; var Error: TSP_ErrorCode);
Var
  Token, Idx, Idx2, Idx3, Idx5, LineNum, TokenLen, MaxLineNum, CeilingNum: Integer;
  NeedStart, NeedFinish: Boolean;
  Changed, CanTest: Boolean;
  Tokens, TempStr, Empty: aString;
  CurLines: TAnsiStringlist;
  NewLines: Array of aString;
  taFloat: aFloat;
Begin

  SP_DeleteIncludes;
  MaxLineNum := pLongWord(@SP_Program[SP_Program_Count -1][2])^;

  Empty := '';
  If (Line <= 0) or (Step <= 0) Then Begin
    Error.Code := SP_ERR_INTEGER_OUT_OF_RANGE;
    Exit;
  End;

  CurLines := TAnsiStringlist.Create;

  If (Start = -1) and (Finish = -1) Then Begin

    Start := 0;
    Finish := SP_Program_Count -1;
    NeedStart := False;
    NeedFinish := False;

  End Else Begin

    NeedStart := True;
    NeedFinish := True;

  End;

  For Idx := 0 To SP_Program_Count -1 Do Begin

    LineNum := pLongWord(@SP_Program[Idx][2])^;
    If NeedStart Then
      If LineNum >= Start then Begin
        Start := Idx;
        NeedStart := False;
      End;
    If NeedFinish Then
      If LineNum > Finish Then Begin
        Finish := Idx -1;
        NeedFinish := False;
      End;
    CurLines.Add(LongWordToString(LineNum));

  End;

  If NeedFinish Then
    Finish := SP_Program_Count -1;

  Idx := 0;
  SetLength(NewLines, (Finish - Start) +1);
  While Finish >= Start Do Begin
    NewLines[Idx] := SP_Program[Start];
    SP_Program_Delete(Start);
    Dec(Finish);
    Inc(Idx);
  End;

  LineNum := Line;
  For Idx := 0 To Length(NewLines) -1 Do Begin

    TempStr := NewLines[Idx];
    TempStr := aChar(SP_LINE_NUM) + LongWordToString(LineNum) + Copy(TempStr, 2 + SizeOf(LongWord), Length(TempStr));
    SP_Store_Line(TempStr);
    CurLines.Objects[Idx + Start] := TObject(LineNum);
    Inc(LineNum, Step);

  End;

  CeilingNum := LongWord(CurLines.Objects[CurLines.Count -1]) + Step;

  // Iterate through program, check GOTO/GOSUB etc commands with old numbers, change to new line numbers

  For Idx := 0 To SP_Program_Count -1 Do Begin

    Idx2 := 1;
    Changed := False;
    Tokens := SP_Program[Idx];

    If Tokens[Idx2] = aChar(SP_LINE_NUM) Then
      Inc(Idx2, 1 + SizeOf(LongWord));

    If Tokens[Idx2] = aChar(SP_LINE_LEN) Then
      Inc(Idx2, SizeOf(LongWord) +1);

    If Tokens[Idx2] = aChar(SP_STATEMENTS) Then
      Inc(Idx2, 1 +((1 + pLongWord(@Tokens[Idx2 +1])^) * SizeOf(LongWord)));

    While Idx2 <= Length(Tokens) Do Begin

      CanTest := False;

      Token := Byte(Tokens[Idx2]);
      Inc(Idx2);

      Case Token of

        SP_KEYWORD, SP_FUNCTION:
          Begin
            Case pLongWord(@Tokens[Idx2])^ of
              SP_KW_GO:
                Begin
                  Inc(Idx2, SizeOf(LongWord));
                  If Byte(Tokens[Idx2]) = SP_KEYWORD Then Begin
                    Inc(Idx2);
                    Case pLongWord(@Tokens[Idx2])^ of
                      SP_KW_TO, SP_KW_SUB, SP_KW_GOSUB:
                        Begin
                          Inc(Idx2, SizeOf(LongWord));
                          CanTest := True;
                        End;
                    Else
                      Begin
                        Inc(Idx2, SizeOf(LongWord));
                      End;
                    End;
                  End Else
                    If (Byte(Tokens[Idx2]) = SP_SYMBOL) and (Tokens[Idx2 +1] = SP_CHAR_SUB) Then Begin
                      Inc(Idx2, 2);
                      CanTest := True;
                    End;
                End;
              SP_KW_RESTORE, SP_KW_RUN:
                Begin
                  CanTest := True;
                  Inc(Idx2, SizeOf(LongWord));
                End;
            Else
              Begin
                Inc(Idx2, SizeOf(LongWord));
              End;
            End;
          End;
        SP_LABEL, SP_STRING, SP_STRUCT_MEMBER_N, SP_STRUCT_MEMBER_S:
          Inc(Idx2, SizeOf(LongWord) + pLongWord(@Tokens[Idx2])^);
        SP_VALUE, SP_STRINGCHAR:
          Inc(Idx2, SizeOf(aFloat));
        SP_VALUE10:
          Inc(Idx2, 10);
        SP_NUMVAR, SP_STRVAR:
          Inc(Idx2, (SizeOf(LongWord)*2) + pLongWord(@Tokens[Idx2 + SizeOf(LongWord)])^);
        SP_TEXT:
          Inc(Idx2, SizeOf(LongWord) + pLongWord(@Tokens[Idx2])^);
        SP_SYMBOL:
          Inc(Idx2);
        SP_TERMINAL:
          Break;
      End;

      If CanTest Then Begin
        // Now pointing at the token which *should* be an SP_VALUE type - check for:
        // 1. Is actually an SP_VALUE
        // 2. Only token that follows is either SP_SYMBOL(:), SP_KW_ELSE or SP_TERMINAL.
        Token := Byte(Tokens[Idx2]);
        If Token = SP_VALUE Then Begin
          LineNum := Trunc(gaFloat(@Tokens[Idx2 +1]));
          Inc(Idx2);
          Idx3 := Idx2;
          Inc(Idx2, 1 + SizeOf(aFloat));
          Inc(Idx2, SizeOf(LongWord) + pLongWord(@Tokens[Idx2])^);
          Token := Byte(Tokens[Idx2]);
          If ((Token = SP_SYMBOL) And (Tokens[Idx2 + 1] = ':')) or (Token = SP_TERMINAL) or ((Token = SP_KEYWORD) And (pLongWord(@Tokens[Idx2 + 1])^ = SP_KW_ELSE)) Then Begin
            If LineNum > MaxLineNum Then Begin
              taFloat := CeilingNum;
              WriteaFloat(@Tokens[Idx3], taFloat);
              TempStr := IntToString(CeilingNum);
              Tokens := Copy(Tokens, 1, Idx3 + SizeOf(aFloat)) + LongWordToString(Length(TempStr)) + TempStr + Copy(Tokens, Idx2, Length(Tokens));
              Idx2 := Idx3 + SizeOf(aFloat) + 1 + SizeOf(LongWord) + Length(TempStr);
              Changed := True;
              Break;
            End Else Begin
              For Idx5 := 0 To CurLines.Count -1 Do Begin
                TempStr := CurLines[Idx5];
                If pLongWord(@TempStr[1])^ >= LongWord(LineNum) Then Begin
                  If CurLines.Objects[Idx5] <> nil Then Begin
                    taFloat := LongWord(CurLines.Objects[Idx5]);
                    WriteaFloat(@Tokens[Idx3], taFloat);
                    TempStr := IntToString(LongWord(CurLines.Objects[Idx5]));
                    Tokens := Copy(Tokens, 1, Idx3 + SizeOf(aFloat)) + LongWordToString(Length(TempStr)) + TempStr + Copy(Tokens, Idx2, Length(Tokens));
                    Idx2 := Idx3 + SizeOf(aFloat) + 1 + SizeOf(LongWord) + Length(TempStr);
                    Changed := True;
                  End;
                  Break;
                End;
              End;
            End;
            If Token = SP_TERMINAL Then Break;
          End;
        End;
      End;

    End;

    If Changed Then
      SP_Program[Idx] := Tokens;

  End;

  Idx2 := -1;
  For Idx := 0 To SP_Program_Count -1 Do Begin
    Error.Position := 1;
    TempStr := SP_Detokenise(SP_Program[Idx], Idx2, False, False);
    TempStr := SP_TokeniseLine(TempStr, False, True) + SP_TERMINAL_SEQUENCE;
    SP_Convert_ToPostFix(TempStr, Error.Position, Error);
    SP_Store_Line(TempStr);
  End;

  PROGCHANGED := True;
  SP_MakeEditorList;
  SP_PreParse(False, Error);
  CurLines.Free;

End;

Procedure SP_DeleteLines(Start, Finish: Integer; var Error: TSP_ErrorCode);
Var
  LineNum, Idx, ProgLen: Integer;
  NeedStart, NeedFinish: Boolean;
Begin

  SP_DeleteIncludes;

  If (Start > Finish) Then Begin
    Error.Code := SP_ERR_INTEGER_OUT_OF_RANGE;
    Exit;
  End;

  If INCLUDEFROM = -1 Then
    ProgLen := Length(SP_Program) -1
  Else
    ProgLen := INCLUDEFROM -1;

  NeedStart := True;
  NeedFinish := True;
  For Idx := 0 To ProgLen Do Begin

    LineNum := pLongWord(@SP_Program[Idx][2])^;
    If NeedStart Then
      If LineNum >= Start then Begin
        Start := Idx;
        NeedStart := False;
      End;
    If NeedFinish Then
      If LineNum >= Finish Then Begin
        Finish := Idx;
        NeedFinish := False;
      End;

  End;

  If NeedFinish Then
    Finish := ProgLen;

  While (Finish >= Start) And (Start <= ProgLen) Do Begin
    SP_Program_Delete(Start);
    Dec(Finish);
  End;

  SP_MakeEditorList;

End;

Function SP_LIST(Start, Finish: Integer): aString;
Var
  Idx, LineNum, cPos, ProgLen: Integer;
  LineNumStr, Tokens: aString;
Begin

  Result := '';
  If INCLUDEFROM = -1 Then
    ProgLen := Length(SP_Program) -1
  Else
    ProgLen := INCLUDEFROM -1;

  Idx := 0;
  While (Idx <= ProgLen) And (pLongWord(@SP_Program[Idx][2])^ < LongWord(Start)) Do Inc(Idx);

  While (Idx <= ProgLen) And (pLongWord(@SP_Program[Idx][2])^ <= LongWord(Finish)) Do Begin

    cPos := -1;
    Tokens := SP_Program[Idx];
    LineNum := pLongWord(@Tokens[2])^;
    LineNumStr := IntToString(LineNum);
    Tokens := SP_DeTokenise(Tokens, cPos, False, False);
    For cPos := 1 to 5 - Length(LineNumStr) Do Tokens := ' ' + Tokens;
    If LineNum = PROGLINE Then Tokens := Copy(Tokens, 1, 5)+'>'+Copy(Tokens, 7, Length(Tokens));

    Result := Result + Tokens + #13;
    Inc(Idx);

  End;

End;

end.




