// Copyright (C) 2010 By Paul Dunn
//
// This file is part of the SpecBAS BASIC Interpreter, which is in turn
// part of the SpecOS project.
//
// SpecBAS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// SpecBAS is distributed in the hope that it will be entertaining,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SpecBAS.  If not, see <http://www.gnu.org/licenses/>.

unit SP_SysVars;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}
{$INCLUDE SpecBAS.inc}

interface

Uses SP_Util, SP_Errors;

Type

  TConstant = Packed Record Name: aString; Value: aFloat; End;
  TSysVar = Record
    Name: aString;
    svType: Integer;
    Size: Integer;
    Data: Pointer;
  End;

Var

  // System-wide sysvars

  BUILDSTR:                 aString;      // Build number of SpecBAS
  HARDWARE:                 aString;      // The hardware platform SpecBAS is running on (Windows/Linux x86/Pandora/Pi etc).
  SCREENCHANGE:             Boolean;      // Screen resolution changing?
  SPFULLSCREEN:             Boolean;      // True if we're in Fullscreen mode.
  SCRUPDATENOW:             Boolean;
  DRAWING:                  Boolean;
  SCMINX, SCMINY,
  SCMAXX, SCMAXY:           Integer;      // "Dirty rectangle" display coordinates for the SP_Composite routine
  SORGX, SORGY, SORGW,
  SORGH, WINSCALEX,
  WINSCALEY:                aFloat;       // The current window's origin and extent metrics
  BORDERSIZE:               Integer;      // The size of the screen border.
  WINSCALE, WINORIGIN:      Boolean;      // Are the origin or scaling of a window active?
  NUMSPRITES:               Integer;      // Global number of sprites defined.
  PROGNAME:                 aString;      // Name of current program
  FILENAMED:                Boolean;      // The current program has a filename (ie, it's been saved).
  TIMERES:                  Integer;      // The timer resolution
  MOUSESPRITE:              Integer;      // The sprite used by the mouse
  MOUSEVISIBLE:             Boolean;      // Is the mouse pointer visible?
  MOUSEIMAGE:               aString;      // The image that the mouse pointer covers
  MOUSESTOREW,
  MOUSESTOREH,
  MOUSESTOREX,
  MOUSESTOREY:              LongWord;     // The last known mouse coordinates for restoring the display under the pointer.
  MOUSEISGRAPHIC:           Boolean;      // Is the mouse image a graphic bank?
  MOUSESTR:                 aString;      // If not, then this string holds the mouse pointer image.
  MOUSEHSX, MOUSEHSY:       Integer;      // Mouse hotspot offset coordinates
  NUBMODE1:                 Integer;      // PANDORA ONLY - nub setting - 0/mouse 1/absolute 2/scroll 3/mousebuttons
  NUBMODE2:                 Integer;      // Second (right hand) nub.
  BATTLEVEL:                Integer;      // Pandora battery level
  DRAWINGSPRITES:           Boolean;      // Are we drawing sprites?
  BLOCKSPRITES:             Boolean;      // Temporarily stop sprite updates

  DISPLAYPOINTER:           Pointer;      // A pointer to the display - the one external to SpecOS.
  DISPLAYWIDTH:             LongWord;     // The current display surface width
  DISPLAYHEIGHT:            LongWord;     // The current display surface height
  DISPLAYSTRIDE:            LongWord;     // The *real* width of the display surface
  DISABLECOUNT:             Integer;      // Number of calls to SP_DisableDisplay

  CBLACK,
  CBLUE,
  CRED,
  CMAGENTA,
  CGREEN,
  CCYAN,
  CYELLOW,
  CWHITE,
  CGREY,
  CBRIGHTBLUE,
  CBRIGHTRED,
  CBRIGHTMAGENTA,
  CBRIGHTGREEN,
  CBRIGHTCYAN,
  CBRIGHTYELLOW,
  CBRIGHTWHITE:             LongWord;     // The default palette.

  FDIST:                    aFloat;       // The focal distance used in 3D calculations.

  NUMBANKS:                 Integer;      // The number of memory banks in use.
  SYSFONT:                  Integer;      // The index of the system font - cannot be deleted.
  EDITORFONT:               Integer;      // The index of the editor font.
  EDFONTSCALEX,
  EDFONTSCALEY:             aFloat;       // The editor font's size (as scalars for width and height)
  EDINVERSE:                Byte;
  DOUBLEFONT:               Integer;      // Index of the aFloat-size font, optional for the editor
  NUMSTREAMS:               Integer;      // The number of streams allocated
  NUMCONSTS:                Integer;      // The number of constants currently allocated

  PROGBANK:                 Integer;      // The Bank ID of the program bank.
  COMMANDBANK:              Integer;      // The Bank ID of the bank which holds direct commands.
  CURPROGBANK:              Integer;      // The currently executing bank - usually the same as one of the above.

  SYSTEMSTATE:              Integer;      // What the system is doing now - SS_INPUT, SS_PAUSE or SS_INTERPRET
  PROGCHANGED:              Boolean;      // Set if the program in memory has changed.
  MAXLOWER:                 Integer;      // Maximum number of rows in the editing area. Minimum of 2.
  EDITLINE:                 aString;      // The current edit line
  CURSORPOS:                Integer;      // The position within the edit line of the cursor during editing.
  CURSORX:                  Integer;      // The X Position of the cursor.
  CURSORY:                  Integer;      // The Y Position of the cursor - used to quickly redraw the cursor on FLASH state-change.
  CURSORCHAR:               Byte;         // The character that is currently being covered by the cursor.
  CURSORFG:                 Longword;     // The foreground colour of the cursor
  CURSORBG:                 Longword;     // The background colour of the cursor. Flipped while flashing.
  INPUTLINE:                aString;      // The current line of input
  INPUTCSR1:                LongWord;     // Two colours used by the INPUT cursor
  INPUTCSR2:                LongWord;
  INPUTPOSX,
  INPUTPOSY:                Integer;      // Print position prior to calling an INPUT.
  INPUTERROR:               Boolean;      // Was there an error in INPUT? Set if yes.
  INPUTERRORHANDLED:        Boolean;      // Are errors in INPUT handled by the system?
  INPUTERRVAL:              aFloat;       // Value to return in case of error in INPUT in a number
  INFORMAT:                 aString;      // The INPUT command's current MASK.
  INPUTBACK:                aString;      // Backup (in a string!) of the window used by INPUT
  ERRORSTATE:               Boolean;      // Indicates an error state in the edit line
  LINK:                     Integer;      // The colour of the INK used in the lower screen for editing. Not user-changeable.
  SCREENLOCK:               Boolean;      // True if the screen is locked and cannot be redrawn.
  LASTERROR:                Integer;      // The last error code produced - 0 if the line executed successfully else see sp_errors.pas
  LASTERRORLINE:            Integer;      // The line that the last error occurred on
  LASTERRORSTATEMENT:       Integer;      // The statement within the line that the last error occurred on
  CONTLINE:                 Integer;      // Line that CONTINUE jumps to
  CONTSTATEMENT:            Integer;      // Statement that CONTINUE jumps to
  COMMAND_TOKENS:           aString;      // The current command line.
  PROGSTATE:                Integer;      // Program state - running, or editing.
  IGNORE_ON_ERROR:          Boolean;      // If an ON ERROR command is run, then allow errors in the handler.
  INCLUDEFROM:              Integer;      // The line number that INCLUDEs are appended from, if present (-1 if not)
  INPROC:                   Boolean;      // Are we executing code inside a PROC?

  FRAMES:                   LongWord;     // Elapsed frame counter, 50Hz
  LASTFRAME:                LongWord;
  FLASHSTATE:               Integer;      // 1 or 0 for the cursor flash
  EDITERROR:                Boolean;      // True if the syntax check caught an editing error
  EDITERRORPOS:             Integer;      // Position in the EDITLINE aString of the error
  PROGLINE:                 Integer;      // The currently selected line in the listing
  SHOWLINE:                 Integer;      // The top line of an AUTOLIST

  DBGSTR:                   aString;      // Debugging purposes
  FPS:                      aFloat;       // Frames per second - default 50.
  FRAME_MS:                 Integer;      // Frame size in Ms.

  CAPSLOCK:                 Integer;      // State of the CAPSLOCK key
  NUMLOCK:                  Integer;      // State of the NUMLOCK key
  INSERT:                   Boolean;      // State of the Insert/Overwrite flag
  BREAKSIGNAL:              Boolean;      // BREAK pressed?
  KEYSTATE:   array[0..255] of Byte;      // The state of the keyboard
  LASTKEY:                  Byte;         // The last key that went down. 0 if key in previous LASTKEY goes up.
  LASTKEYCHAR:              aString;      // Decoded key - takes international keymaps into account.
  REPDEL:                   LongWord;     // The delay in frames before a key repeats when held down
  REPPER:                   LongWord;     // Delay between successive repeats of a key that is held down
  REPCOUNT:                 LongWord;     // The current counter for key repeats
  DEBUGVAL:                 LongWord;
  VOLUME:                   aFloat;       // The current volume of sounds in SpecBAS
  CLICKBANK,
  OKSNDBANK,
  ERRSNDBANK:               Integer;      // Sample banks for the system sounds
  CLICKCHAN,
  OKCHAN,
  ERRORCHAN:                LongWord;     // Channel numbers for the system sounds

  COLLIDE_FLAG:             Boolean;      // For the ON COLLIDE command
  M_MOVEFLAG:               Boolean;      // For the ON MOUSEMOVE command
  M_DOWNFLAG:               Boolean;      // For the ON MOUSEDOWN command
  M_UPFLAG:                 Boolean;      // For the ON MOUSEUP command
  M_WHEELUPFLAG:            Boolean;      // For the ON WHEELUP command
  M_WHEELDNFLAG:            Boolean;      // For the ON WHEELDOWN command
  K_DOWNFLAG:               Boolean;      // For the ON MOUSEDOWN command
  K_UPFLAG:                 Boolean;      // For the ON MOUSEUP command

  MOUSEX:                   Integer;      // The current mouse x-coordinate
  MOUSEY:                   Integer;      // The current mouse y-coordinate
  MOUSEW:                   Integer;
  MOUSEH:                   Integer;
  MOUSEWHEEL:               Integer;      // The current mouse wheel position.
  MOUSEBTN:                 LongWord;     // The current state of the mouse buttons - bits 0,1,2 for left/right/middle
  M_DELTAX:                 Integer;      // Amount the mouse has moved in the X direction
  M_DELTAY:                 Integer;      // Amount the mouse has moved in the Y direction

  AUTOSAVE:                 Boolean;      // Automatically save your work after each direct command/Line Entry?
  HOMEFOLDER:               aString;      // The folder that specbas lives in

  QUITMSG:                  Boolean;      // Signals that the interpreter is to quit when true.

  MP3AVAILABLE:             Boolean;      // Libraries for MP3 decoding are available
  MODAVAILABLE:             Boolean;      // Libraries for MOD/S3M/IT/XM decoding are available
  FLACAVAILABLE:            Boolean;      // Libraries for decoding FLAC files are available
  OGGAVAILABLE:             Boolean;      // Libraries for decoding OGG files are available
  MAXRATE:                  LongWord;     // Maximum sample rate supported
  MINRATE:                  LongWord;     // Minimum sample rate supported
  MUSICHANDLE:              LongWord;     // The handle to the currently loaded music file. 0 if no music loaded.
  MUSICISSTREAM:            Boolean;      // Is the music currently loaded a stream?

  LOGGING:                  Boolean;      // enable debug logging?

  LISTWINDOW:               Integer;      // The Window Index of the Listing window
  COMMANDWINDOW:            Integer;      // The Window ID of the direct command window
  SPLITSTATEMENTS:          Boolean;      // Split statements in the editor?
  SPLITREMS:                Boolean;      // Word-wrap REM statements?
  EDITORWRAP:               Boolean;      // Wrap long statements 128k editor style?
  LISTTOPINDEX:             aFloat;       // The topindex of the list window - not the line, but the text index
  LISTLEFTINDEX:            aFloat;       // The Leftmost column in the list window
  LISTSELLINE:              Integer;      // The currently selected linenumber in the LIST window
  LISTLINECOUNT:            Integer;      // Number of lines in the listing, counting statements.
  LISTCOLCOUNT:             Integer;      // The maximum number of characters the listing's longest line holds
  LISTVISLINES:             Integer;      // The number of lines visible in the listing window.
  LISTVISCOLS:              Integer;      // Number of characters visible per line
  LISTWINX:                 Integer;      // The List window metrics
  LISTWINY:                 Integer;
  LISTWINW:                 Integer;
  LISTWINH:                 Integer;
  LISTVSCUPENABLED:         Boolean;      // Is the list window's vertical scrollbar "up" arrow enabled?
  LISTVSCUPTLX:             Integer;
  LISTVSCUPTLY:             Integer;
  LISTVSCUPBRX:             Integer;
  LISTVSCUPBRY:             Integer;      // The vertical scrollbar in the listing window's "up" arrow's bounding box
  LISTVSCDNENABLED:         Boolean;      // Is the list window's vertical scrollbar "down" arrow enabled?
  LISTVSCDNTLX:             Integer;
  LISTVSCDNTLY:             Integer;
  LISTVSCDNBRX:             Integer;
  LISTVSCDNBRY:             Integer;      // The vertical scrollbar in the listing window's "down" arrow's bounding box
  LISTVSTRENABLED:          Boolean;      // Is the tracking region of the list window's vertical scrollbar enabled?
  LISTVSTRTLX:              Integer;      // The list window's vertical scrollbar track region
  LISTVSTRTLY:              Integer;
  LISTVSTRBRX:              Integer;
  LISTVSTRBRY:              Integer;
  LISTVTHP:                 Integer;      // The vertical position of the listing window's vertical scrollbar thumb.
  LISTVTHS:                 Integer;      // The physical size of the listing window's vertical scrollbar thumb.
  LISTVSCTRKH:              Integer;      // The height of the vertical scrollbar's trackbar

  LISTHSCUPENABLED:         Boolean;      // Is the list window's horizontal scrollbar "up" arrow enabled?
  LISTHSCUPTLX:             Integer;
  LISTHSCUPTLY:             Integer;
  LISTHSCUPBRX:             Integer;
  LISTHSCUPBRY:             Integer;      // The horizontal scrollbar in the listing window's "up" arrow's bounding box
  LISTHSCDNENABLED:         Boolean;      // Is the list window's horizontal scrollbar "down" arrow enabled?
  LISTHSCDNTLX:             Integer;
  LISTHSCDNTLY:             Integer;
  LISTHSCDNBRX:             Integer;
  LISTHSCDNBRY:             Integer;      // The horizontal scrollbar in the listing window's "down" arrow's bounding box
  LISTHSTRENABLED:          Boolean;      // Is the tracking region of the list window's horizontal scrollbar enabled?
  LISTHSTRTLX:              Integer;      // The list window's horizontal scrollbar track region
  LISTHSTRTLY:              Integer;
  LISTHSTRBRX:              Integer;
  LISTHSTRBRY:              Integer;
  LISTHTHP:                 Integer;      // The horizontal position of the listing window's horizontal scrollbar thumb.
  LISTHTHS:                 Integer;      // The physical size of the listing window's horizontal scrollbar thumb.

  LISTSIZETLX:              Integer;
  LISTSIZETLY:              Integer;
  LISTSIZEBRX:              Integer;
  LISTSIZEBRY:              Integer;
  CCOMMANDWINDOW:           Boolean;      // True when the command window exists
  CLISTWINDOW:              Boolean;      // True when the LIST window exists.
  COMMANDWINX:              Integer;      // The Command window metrics
  COMMANDWINY:              Integer;
  COMMANDWINW:              Integer;
  COMMANDWINH:              Integer;
  COMMANDSIZETLX:           Integer;
  COMMANDSIZETLY:           Integer;
  COMMANDSIZEBRX:           Integer;
  COMMANDSIZEBRY:           Integer;
  SHOWLIST:                 Boolean;
  EDITORHIDDEN:             Integer;
  HELPBANK:                 Integer;      // The Bank-ID of the help file.
  HELPWINDOW:               Integer;      // The Window-ID of the help window in the editor.

  TEMPDIR:                  aString;      // the location of the TEMP directory in the host filesystem.

  // Thread-specific sysvars

  MULTITHREAD:              Boolean;      // Is the interpreter multi-threaded? False for single-core devices.
  SCREENPOINTER:            Pointer;      // A pointer to the current drawing surface
  SCREENBANK:               Integer;      // The bank ID number of the current screen being drawn to
  SCREENWIDTH:              Integer;      // The current drawing surface width
  SCREENHEIGHT:             Integer;      // The current drawing surface height
  SCREENX:                  Integer;      // The current drawing surface X position
  SCREENY:                  Integer;      // The current drawing surface Y position
  SCREENBPP:                Integer;      // The current drawing surface's bitdepth
  SCREENSTRIDE:             Integer;      // The *real* width of the drawing surface - some bitmap formats are 32bit aligned.
  SCREENID:                 Integer;      // The WindowID of the current drawing surface.
  SCREENVISIBLE:            Boolean;      // Is the current surface (potentially) visible?
  WINDOWPOINTER:            Pointer;      // Pointer to the SP_Window_Info structure
  SCRBANKPOINTER:           Pointer;      // As above, but for the display bank.
  FONTBANKID:               Integer;      // The ID number of the bank to use as the current font.
  FONTWIDTH:                LongWord;     // The Width of the current font
  FONTHEIGHT:               LongWord;     // The Height of the current font
  FONTTRANSPARENT:          Integer;      // The colour index to be used as a transparent colour in the current font
  FONTTYPE:                 Integer;      // The type of font being used (graphical or mono)

  COLOURSCHANGED:           Boolean;      // Are the temporary colours changed?
  LOWERSAVE:                aString;      // Storage so that the Scroll? prompt doesn't destroy the screen.
  CPAPER:                   LongWord;     // Current PAPER set by the PAPER command
  CINK:                     LongWord;     // Current INK set by the INK command
  CINVERSE:                 Integer;      // Current state of the INVERSE ... state?
  COVER:                    Integer;      // Current state of the OVER flag
  CCLIPX1:                  Integer;      // Current clip rect
  CCLIPY1:                  Integer;
  CCLIPX2:                  Integer;
  CCLIPY2:                  Integer;
  CSCALEX:                  aFloat;
  CSCALEY:                  aFloat;
  COUTEXPR:                 aString;
  T_PAPER:                  LongWord;     // Temporary PAPER used by PRINT and INPUT colour items
  T_INK:                    LongWord;     // Temporary INK
  T_INVERSE:                Integer;      // Temporary INVERSE
  T_OVER:                   Integer;      // Temporary OVER
  T_CLIPX1:                 Integer;      // Temporary cliprect
  T_CLIPY1:                 Integer;
  T_CLIPX2:                 Integer;
  T_CLIPY2:                 Integer;
  T_OUTEXPR:                aString;
  T_OUTASSIGN:              aString;
  T_OUTMODE:                Integer;
  T_OUTSTRM:                Integer;
  T_SCALEX:                 aFloat;
  T_SCALEY:                 aFloat;
  T_USINGMASK:              aString;
  T_TRANSPARENT:            Boolean;
  T_CENTRE:                 Boolean;
  T_CENTRETEXT:             aString;
  T_CENTRE_Y:               Integer;
  OUTBUFFER:                aString;
  OUTSET:                   Boolean;
  OUTWORKSP:                aString;
  COUTSTRM:                 Integer;
  COUTMODE:                 Integer;      // OUT redirects to screen (0), variable (1) or stream (2)


  DRPOSX:                   Integer;      // The x-coordinate of the last point plotted
  DRPOSY:                   Integer;      // The y-coordinate of the last point plotted
  DRHEADING:                aFloat;       // The heading of the "turtle", in whichever angular system is in use (deg, rad)
  XORG:                     Integer;      // The x-coordinate (in screen space) of the graphics origin
  YORG:                     Integer;      // As above, but for y.
  PRPOSX:                   Integer;      // The X-Coordinate of the PRINT position
  PRPOSY:                   Integer;      // The Y Coordinate of the PRINT position
  TABSIZE:                  Integer;      // The size in characters of tab stops, starting at the far left of the screen.
  SCROLLCNT:                Integer;      // The scroll counter - when this reaches the bottom of the screen, the "Scroll?" message is triggered.
  NXTLINE:                  Integer;      // The next line to be executed. If 0, then stop.
  NXTSTATEMENT:             Integer;      // The statement to be executed. Used by the RETURN and NEXT commands.
  MATHMODE:                 Integer;      // 0 - Radians, 1 - Degrees
  BASE:                     LongWord;     // The base of arrays. Default to 1.

  INITMS:                   LongWord;     // The time that SpecBAS was started, in MS.

Const

  SS_INPUT =            0;
  SS_PAUSE =            1;
  SS_INTERPRET =        2;
  SS_ERROR =            3;
  SS_IDLE =             4;
  SS_EDITOR =           5;

  BlockGraphics: Array[0..127] of Byte =
    (0, 0, 0, 0, 0, 0, 0, 0,
     15, 15, 15, 15, 0, 0, 0, 0,
     240, 240, 240, 240, 0, 0, 0, 0,
     255, 255, 255, 255, 0, 0, 0, 0,
     0, 0, 0, 0, 15, 15, 15, 15,
     15, 15, 15, 15, 15, 15, 15, 15,
     240, 240, 240, 240, 15, 15, 15, 15,
     255, 255, 255, 255, 15, 15, 15, 15,
     0, 0, 0, 0, 240, 240, 240, 240,
     15, 15, 15, 15, 240, 240, 240, 240,
     240, 240, 240, 240, 240, 240, 240, 240,
     255, 255, 255, 255, 240, 240, 240, 240,
     0, 0, 0, 0, 255, 255, 255, 255,
     15, 15, 15, 15, 255, 255, 255, 255,
     240, 240, 240, 240, 255, 255, 255, 255,
     255, 255, 255, 255, 255, 255, 255, 255);

  Slope: Array[0..7] of Byte =
    (1, 3, 7, 15, 31, 63, 127, 255);

  Arrows: Array[0..31] of Byte =
    (0, 0, 24, 60, 126, 126, 0, 0,
     0, 0, 126, 126, 60, 24, 0, 0,
     0, 48, 56, 60, 60, 56, 48, 0,
     0, 12, 28, 60, 60, 28, 12, 0);

  SizeGrip: Array[0..7] of Byte =
    (0, 0, 2, 0, 10, 0, 42, 0);

  TitleBar: Array[0..7] of Byte =
    (0, 255, 255, 0, 0, 255, 255, 0);

  Lock: Array[0..7] of Byte =
    (0, 96, 144, 159, 149, 97, 0, 0);

//    (48, 72, 136, 144, 104, 4, 10, 4);

  DefaultFont: Array[0..(96*64)-1] Of Byte =
  (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0);

   svString = 0;
   svArray = 1;
   svBoolean = 2;
   svLongWord = 3;
   svaFloat = 4;
   svInteger = 5;
   svPointer = 6;
   svByte = 7;

   SP_Constants: Array[0..117] of TConstant =
   ((Name: 'TRUE'; Value: 1),
    (Name: 'FALSE'; Value: 0),

    (Name: 'KEY_BACKSPACE'; Value: 8),
    (Name: 'KEY_TAB'; Value: 9),
    (Name: 'KEY_RETURN'; Value: 13),
    (Name: 'KEY_SHIFT'; Value: 16),
    (Name: 'KEY_CONTROL'; Value: 17),
    (Name: 'KEY_ALT'; Value: 18),
    (Name: 'KEY_PAUSE'; Value: 19),
    (Name: 'KEY_CAPS_LOCK'; Value: 20),
    (Name: 'KEY_ESCAPE'; Value: 27),
    (Name: 'KEY_SPACE'; Value: 32),
    (Name: 'KEY_PG_UP'; Value: 33),
    (Name: 'KEY_PG_DOWN'; Value: 34),
    (Name: 'KEY_END'; Value: 35),
    (Name: 'KEY_HOME'; Value: 36),
    (Name: 'KEY_LEFT'; Value: 37),
    (Name: 'KEY_UP'; Value: 38),
    (Name: 'KEY_RIGHT'; Value: 39),
    (Name: 'KEY_DOWN'; Value: 40),
    (Name: 'KEY_PRINT'; Value: 42),
    (Name: 'KEY_INSERT'; Value: 45),
    (Name: 'KEY_DELETE'; Value: 46),
    (Name: 'KEY_0'; Value: 48),
    (Name: 'KEY_1'; Value: 49),
    (Name: 'KEY_2'; Value: 50),
    (Name: 'KEY_3'; Value: 51),
    (Name: 'KEY_4'; Value: 52),
    (Name: 'KEY_5'; Value: 53),
    (Name: 'KEY_6'; Value: 54),
    (Name: 'KEY_7'; Value: 55),
    (Name: 'KEY_8'; Value: 56),
    (Name: 'KEY_9'; Value: 57),
    (Name: 'KEY_A'; Value: 65),
    (Name: 'KEY_B'; Value: 66),
    (Name: 'KEY_C'; Value: 67),
    (Name: 'KEY_D'; Value: 68),
    (Name: 'KEY_E'; Value: 69),
    (Name: 'KEY_F'; Value: 70),
    (Name: 'KEY_G'; Value: 71),
    (Name: 'KEY_H'; Value: 72),
    (Name: 'KEY_I'; Value: 73),
    (Name: 'KEY_J'; Value: 74),
    (Name: 'KEY_K'; Value: 75),
    (Name: 'KEY_L'; Value: 76),
    (Name: 'KEY_M'; Value: 77),
    (Name: 'KEY_N'; Value: 78),
    (Name: 'KEY_O'; Value: 79),
    (Name: 'KEY_P'; Value: 80),
    (Name: 'KEY_Q'; Value: 81),
    (Name: 'KEY_R'; Value: 82),
    (Name: 'KEY_S'; Value: 83),
    (Name: 'KEY_T'; Value: 84),
    (Name: 'KEY_U'; Value: 85),
    (Name: 'KEY_V'; Value: 86),
    (Name: 'KEY_W'; Value: 87),
    (Name: 'KEY_X'; Value: 88),
    (Name: 'KEY_Y'; Value: 89),
    (Name: 'KEY_Z'; Value: 90),
    (Name: 'KEY_LWIN'; Value: 91),
    (Name: 'KEY_RWIN'; Value: 92),
    (Name: 'KEY_NUMPAD_0'; Value: 96),
    (Name: 'KEY_NUMPAD_1'; Value: 97),
    (Name: 'KEY_NUMPAD_2'; Value: 98),
    (Name: 'KEY_NUMPAD_3'; Value: 99),
    (Name: 'KEY_NUMPAD_4'; Value: 100),
    (Name: 'KEY_NUMPAD_5'; Value: 101),
    (Name: 'KEY_NUMPAD_6'; Value: 102),
    (Name: 'KEY_NUMPAD_7'; Value: 103),
    (Name: 'KEY_NUMPAD_8'; Value: 104),
    (Name: 'KEY_NUMPAD_9'; Value: 105),
    (Name: 'KEY_MULTIPLY'; Value: 106),
    (Name: 'KEY_ADD'; Value: 107),
    (Name: 'KEY_SEPARATOR'; Value: 108),
    (Name: 'KEY_SUBTRACT'; Value: 109),
    (Name: 'KEY_DECIMAL'; Value: 110),
    (Name: 'KEY_DIVIDE'; Value: 111),
    (Name: 'KEY_F1'; Value: 112),
    (Name: 'KEY_F2'; Value: 113),
    (Name: 'KEY_F3'; Value: 114),
    (Name: 'KEY_F4'; Value: 115),
    (Name: 'KEY_F5'; Value: 116),
    (Name: 'KEY_F6'; Value: 117),
    (Name: 'KEY_F7'; Value: 118),
    (Name: 'KEY_F8'; Value: 119),
    (Name: 'KEY_F9'; Value: 120),
    (Name: 'KEY_F10'; Value: 121),
    (Name: 'KEY_F11'; Value: 122),
    (Name: 'KEY_F12'; Value: 123),
    (Name: 'KEY_NUM_LOCK'; Value: 144),
    (Name: 'KEY_SCROLL_LOCK'; Value: 145),

    (Name: 'FONT_MONO'; Value: 0),
    (Name: 'FONT_COLOUR'; Value: 1),

    (Name: 'dtNONE'; Value: 0),
    (Name: 'dtRANDOM'; Value: 1),
    (Name: 'dtDIAGONAL5'; Value: 2),
    (Name: 'dtDIAGONAL9'; Value: 3),
    (Name: 'dtCLUSTERED'; Value: 4),
    (Name: 'dtDISPERSED'; Value: 5),
    (Name: 'dtHALFTONE'; Value: 6),
    (Name: 'dtORDERED16'; Value: 7),
    (Name: 'dtORDERED64'; Value: 8),
    (Name: 'dtFLOYD'; Value: 9),
    (Name: 'dtJARVIS'; Value: 10),
    (Name: 'dtSTUCKI'; Value: 11),
    (Name: 'dtBURKES'; Value: 12),
    (Name: 'dtSIERRA3'; Value: 13),
    (Name: 'dtSIERRA2'; Value: 14),
    (Name: 'dtSIERRA'; Value: 15),
    (Name: 'dtATKINSON'; Value: 16),
    (Name: 'dtSTEVENSON'; Value: 17),

    (Name: 'tpLOWER'; Value: -1),
    (Name: 'tpNORMAL'; Value: 0),
    (Name: 'tpHIGHER'; Value: 1),

    (Name: 'nubMOUSE'; Value: 0),
    (Name: 'nubSTICK'; Value: 1),
    (Name: 'nubSCROLL'; Value: 2),
    (Name: 'nubBUTTONS'; Value: 3));

  SysVars: Array[0..222] of TSysVar =
  ((Name: 'BUILDSTR'; svType: svString; Size: 0; Data: @BUILDSTR),
   (Name: 'HARDWARE'; svType: svString; Size: 0; Data: @HARDWARE),
   (Name: 'SPFULLSCREEN'; svType: svBoolean; Size: 1; Data: @SPFULLSCREEN),
   (Name: 'SCMINX'; svType: svInteger; Size: 4; Data: @SCMINX),
   (Name: 'SCMINY'; svType: svInteger; Size: 4; Data: @SCMINY),
   (Name: 'SCMAXX'; svType: svInteger; Size: 4; Data: @SCMAXX),
   (Name: 'SCMAXY'; svType: svInteger; Size: 4; Data: @SCMAXY),
   (Name: 'NUMSPRITES'; svType: svInteger; Size: 4; Data: @NUMSPRITES),
   (Name: 'PROGNAME'; svType: svString; Size: 0; Data: @PROGNAME),
   (Name: 'MOUSEVISIBLE'; svType: svBoolean; Size: 1; Data: @MOUSEVISIBLE),
   (Name: 'MOUSEIMAGE'; svType: svString; Size: 0; Data: @MOUSEIMAGE),
   (Name: 'MOUSESTOREX'; svType: svInteger; Size: 4; Data: @MOUSESTOREX),
   (Name: 'MOUSESTOREY'; svType: svInteger; Size: 4; Data: @MOUSESTOREY),
   (Name: 'MOUSEISGRAPHIC'; svType: svBoolean; Size: 1; Data: @MOUSEISGRAPHIC),
   (Name: 'MOUSESTR'; svType: svString; Size: 0; Data: @MOUSESTR),
   (Name: 'MOUSEHSX'; svType: svInteger; Size: 4; Data: @MOUSEHSX),
   (Name: 'MOUSEHSY'; svType: svInteger; Size: 4; Data: @MOUSEHSY),
   (Name: 'DISPLAYPOINTER'; svType: svPointer; Size: 4; Data: @DISPLAYPOINTER),
   (Name: 'DISPLAYWIDTH'; svType: svInteger; Size: 4; Data: @DISPLAYWIDTH),
   (Name: 'DISPLAYHEIGHT'; svType: svInteger; Size: 4; Data: @DISPLAYHEIGHT),
   (Name: 'DISABLECOUNT'; svType: svInteger; Size: 4; Data: @DISABLECOUNT),
   (Name: 'NUMBANKS'; svType: svInteger; Size: 4; Data: @NUMBANKS),
   (Name: 'SYSFONT'; svType: svInteger; Size: 4; Data: @SYSFONT),
   (Name: 'EDITORFONT'; svType: svInteger; Size: 4; Data: @EDITORFONT),
   (Name: 'NUMSTREAMS'; svType: svInteger; Size: 4; Data: @NUMSTREAMS),
   (Name: 'SYSTEMSTATE'; svType: svInteger; Size: 4; Data: @SYSTEMSTATE),
   (Name: 'PROGCHANGED'; svType: svBoolean; Size: 1; Data: @PROGCHANGED),
   (Name: 'EDITLINE'; svType: svString; Size: 0; Data: @EDITLINE),
   (Name: 'CURSORPOS'; svType: svInteger; Size: 4; Data: @CURSORPOS),
   (Name: 'CURSORX'; svType: svInteger; Size: 4; Data: @CURSORX),
   (Name: 'CURSORY'; svType: svInteger; Size: 4; Data: @CURSORY),
   (Name: 'CURSORCHAR'; svType: svByte; Size: 1; Data: @CURSORCHAR),
   (Name: 'CURSORFG'; svType: svByte; Size: 1; Data: @CURSORFG),
   (Name: 'CURSORBG'; svType: svByte; Size: 1; Data: @CURSORBG),
   (Name: 'INPUTLINE'; svType: svString; Size: 0; Data: @INPUTLINE),
   (Name: 'INPUTCSR1'; svType: svByte; Size: 1; Data: @INPUTCSR1),
   (Name: 'INPUTCSR2'; svType: svByte; Size: 1; Data: @INPUTCSR2),
   (Name: 'INPUTPOSX'; svType: svInteger; Size: 4; Data: @INPUTPOSX),
   (Name: 'INPUTPOSY'; svType: svInteger; Size: 4; Data: @INPUTPOSY),
   (Name: 'INPUTERROR'; svType: svBoolean; Size: 1; Data: @INPUTERROR),
   (Name: 'INPUTERRORHANDLED'; svType: svBoolean; Size: 1; Data: @INPUTERRORHANDLED),
   (Name: 'INPUTERRVAL'; svType: svaFloat; Size: 8; Data: @INPUTERRVAL),
   (Name: 'INFORMAT'; svType: svString; Size: 0; Data: @INFORMAT),
   (Name: 'INPUTBACK'; svType: svString; Size: 0; Data: @INPUTBACK),
   (Name: 'ERRORSTATE'; svType: svBoolean; Size: 1; Data: @ERRORSTATE),
   (Name: 'LINK'; svType: svInteger; Size: 4; Data: @LINK),
   (Name: 'SCREENLOCK'; svType: svBoolean; Size: 1; Data: @SCREENLOCK),
   (Name: 'LASTERROR'; svType: svInteger; Size: 4; Data: @LASTERROR),
   (Name: 'LASTERRORLINE'; svType: svInteger; Size: 4; Data: @LASTERRORLINE),
   (Name: 'LASTERRORSTATEMENT'; svType: svInteger; Size: 4; Data: @LASTERRORSTATEMENT),
   (Name: 'CONTLINE'; svType: svInteger; Size: 4; Data: @CONTLINE),
   (Name: 'CONTSTATEMENT'; svType: svInteger; Size: 4; Data: @CONTSTATEMENT),
   (Name: 'COMMAND_TOKENS'; svType: svString; Size: 0; Data: @COMMAND_TOKENS),
   (Name: 'PROGSTATE'; svType: svInteger; Size: 4; Data: @PROGSTATE),
   (Name: 'FRAMES'; svType: svLongWord; Size: 4; Data: @FRAMES),
   (Name: 'FLASHSTATE'; svType: svInteger; Size: 4; Data: @FLASHSTATE),
   (Name: 'EDITERROR'; svType: svBoolean; Size: 1; Data: @EDITERROR),
   (Name: 'EDITERRORPOS'; svType: svInteger; Size: 4; Data: @EDITERRORPOS),
   (Name: 'PROGLINE'; svType: svInteger; Size: 4; Data: @PROGLINE),
   (Name: 'SHOWLINE'; svType: svInteger; Size: 4; Data: @SHOWLINE),
   (Name: 'DBGSTR'; svType: svString; Size: 0; Data: @DBGSTR),
   (Name: 'FPS'; svType: svaFloat; Size: 8; Data: @FPS),
   (Name: 'FRAME_MS'; svType: svInteger; Size: 4; Data: @FRAME_MS),
   (Name: 'CAPSLOCK'; svType: svInteger; Size: 4; Data: @CAPSLOCK),
   (Name: 'NUMLOCK'; svType: svInteger; Size: 4; Data: @NUMLOCK),
   (Name: 'INSERT'; svType: svBoolean; Size: 1; Data: @INSERT),
   (Name: 'KEYSTATE'; svType: svArray; Size: 256; Data: @KEYSTATE[0]),
   (Name: 'LASTKEY'; svType: svByte; Size: 1; Data: @LASTKEY),
   (Name: 'REPDEL'; svType: svLongWord; Size: 4; Data: @REPDEL),
   (Name: 'REPPER'; svType: svLongWord; Size: 4; Data: @REPPER),
   (Name: 'REPCOUNT'; svType: svLongWord; Size: 4; Data: @REPCOUNT),
   (Name: 'DEBUGVAL'; svType: svLongWord; Size: 4; Data: @DEBUGVAL),
   (Name: 'VOLUME'; svType: svaFloat; Size: 8; Data: @VOLUME),
   (Name: 'CLICKBANK'; svType: svInteger; Size: 4; Data: @CLICKBANK),
   (Name: 'OKSNDBANK'; svType: svInteger; Size: 4; Data: @OKSNDBANK),
   (Name: 'ERRSNDBANK'; svType: svInteger; Size: 4; Data: @ERRSNDBANK),
   (Name: 'MOUSEX'; svType: svInteger; Size: 4; Data: @MOUSEX),
   (Name: 'MOUSEY'; svType: svInteger; Size: 4; Data: @MOUSEY),
   (Name: 'MOUSEWHEEL'; svType: svInteger; Size: 4; Data: @MOUSEWHEEL),
   (Name: 'MOUSEBTN'; svType: svLongWord; Size: 4; Data: @MOUSEBTN),
   (Name: 'M_DELTAX'; svType: svInteger; Size: 4; Data: @M_DELTAX),
   (Name: 'M_DELTAY'; svType: svInteger; Size: 4; Data: @M_DELTAY),
   (Name: 'AUTOSAVE'; svType: svBoolean; Size: 1; Data: @AUTOSAVE),
   (Name: 'HOMEFOLDER'; svType: svString; Size: 0; Data: @HOMEFOLDER),
   (Name: 'QUITMSG'; svType: svBoolean; Size: 1; Data: @QUITMSG),
   (Name: 'MP3AVAILABLE'; svType: svBoolean; Size: 1; Data: @MP3AVAILABLE),
   (Name: 'MODAVAILABLE'; svType: svBoolean; Size: 1; Data: @MODAVAILABLE),
   (Name: 'FLACAVAILABLE'; svType: svBoolean; Size: 1; Data: @FLACAVAILABLE),
   (Name: 'OGGAVAILABLE'; svType: svBoolean; Size: 1; Data: @OGGAVAILABLE),
   (Name: 'MAXRATE'; svType: svLongWord; Size: 4; Data: @MAXRATE),
   (Name: 'MINRATE'; svType: svLongWord; Size: 4; Data: @MINRATE),
   (Name: 'MUSICHANDLE'; svType: svLongWord; Size: 4; Data: @MUSICHANDLE),
   (Name: 'MUSICISSTREAM'; svType: svBoolean; Size: 1; Data: @MUSICISSTREAM),
   (Name: 'LOGGING'; svType: svBoolean; Size: 1; Data: @LOGGING),
   (Name: 'LISTWINDOW'; svType: svInteger; Size: 4; Data: @LISTWINDOW),
   (Name: 'COMMANDWINDOW'; svType: svInteger; Size: 4; Data: @COMMANDWINDOW),
   (Name: 'LISTTOPINDEX'; svType: svaFloat; Size: 8; Data: @LISTTOPINDEX),
   (Name: 'LISTLEFTINDEX'; svType: svaFloat; Size: 8; Data: @LISTLEFTINDEX),
   (Name: 'LISTSELLINE'; svType: svInteger; Size: 4; Data: @LISTSELLINE),
   (Name: 'LISTLINECOUNT'; svType: svInteger; Size: 4; Data: @LISTLINECOUNT),
   (Name: 'LISTCOLCOUNT'; svType: svInteger; Size: 4; Data: @LISTCOLCOUNT),
   (Name: 'LISTVISLINES'; svType: svInteger; Size: 4; Data: @LISTVISLINES),
   (Name: 'LISTVISCOLS'; svType: svInteger; Size: 4; Data: @LISTVISCOLS),
   (Name: 'LISTWINX'; svType: svInteger; Size: 4; Data: @LISTWINX),
   (Name: 'LISTWINY'; svType: svInteger; Size: 4; Data: @LISTWINY),
   (Name: 'LISTWINW'; svType: svInteger; Size: 4; Data: @LISTWINW),
   (Name: 'LISTWINH'; svType: svInteger; Size: 4; Data: @LISTWINH),
   (Name: 'LISTVSCUPENABLED'; svType: svBoolean; Size: 1; Data: @LISTVSCUPENABLED),
   (Name: 'LISTVSCUPTLX'; svType: svInteger; Size: 4; Data: @LISTVSCUPTLX),
   (Name: 'LISTVSCUPTLY'; svType: svInteger; Size: 4; Data: @LISTVSCUPTLY),
   (Name: 'LISTVSCUPBRX'; svType: svInteger; Size: 4; Data: @LISTVSCUPBRX),
   (Name: 'LISTVSCUPBRY'; svType: svInteger; Size: 4; Data: @LISTVSCUPBRY),
   (Name: 'LISTVSCDNENABLED'; svType: svBoolean; Size: 1; Data: @LISTVSCDNENABLED),
   (Name: 'LISTVSCDNTLX'; svType: svInteger; Size: 4; Data: @LISTVSCDNTLX),
   (Name: 'LISTVSCDNTLY'; svType: svInteger; Size: 4; Data: @LISTVSCDNTLY),
   (Name: 'LISTVSCDNBRX'; svType: svInteger; Size: 4; Data: @LISTVSCDNBRX),
   (Name: 'LISTVSCDNBRY'; svType: svInteger; Size: 4; Data: @LISTVSCDNBRY),
   (Name: 'LISTVSTRENABLED'; svType: svBoolean; Size: 1; Data: @LISTVSTRENABLED),
   (Name: 'LISTVSTRTLX'; svType: svInteger; Size: 4; Data: @LISTVSTRTLX),
   (Name: 'LISTVSTRTLY'; svType: svInteger; Size: 4; Data: @LISTVSTRTLY),
   (Name: 'LISTVSTRBRX'; svType: svInteger; Size: 4; Data: @LISTVSTRBRX),
   (Name: 'LISTVSTRBRY'; svType: svInteger; Size: 4; Data: @LISTVSTRBRY),
   (Name: 'LISTVTHP'; svType: svInteger; Size: 4; Data: @LISTVTHP),
   (Name: 'LISTVTHS'; svType: svInteger; Size: 4; Data: @LISTVTHS),
   (Name: 'LISTVSCTRKH'; svType: svInteger; Size: 4; Data: @LISTVSCTRKH),
   (Name: 'LISTHSCUPENABLED'; svType: svBoolean; Size: 1; Data: @LISTHSCUPENABLED),
   (Name: 'LISTHSCUPTLX'; svType: svInteger; Size: 4; Data: @LISTHSCUPTLX),
   (Name: 'LISTHSCUPTLY'; svType: svInteger; Size: 4; Data: @LISTHSCUPTLY),
   (Name: 'LISTHSCUPBRX'; svType: svInteger; Size: 4; Data: @LISTHSCUPBRX),
   (Name: 'LISTHSCUPBRY'; svType: svInteger; Size: 4; Data: @LISTHSCUPBRY),
   (Name: 'LISTHSCDNENABLED'; svType: svBoolean; Size: 1; Data: @LISTHSCDNENABLED),
   (Name: 'LISTHSCDNTLX'; svType: svInteger; Size: 4; Data: @LISTHSCDNTLX),
   (Name: 'LISTHSCDNTLY'; svType: svInteger; Size: 4; Data: @LISTHSCDNTLY),
   (Name: 'LISTHSCDNBRX'; svType: svInteger; Size: 4; Data: @LISTHSCDNBRX),
   (Name: 'LISTHSCDNBRY'; svType: svInteger; Size: 4; Data: @LISTHSCDNBRY),
   (Name: 'LISTHSTRENABLED'; svType: svBoolean; Size: 1; Data: @LISTHSTRENABLED),
   (Name: 'LISTHSTRTLX'; svType: svInteger; Size: 4; Data: @LISTHSTRTLX),
   (Name: 'LISTHSTRTLY'; svType: svInteger; Size: 4; Data: @LISTHSTRTLY),
   (Name: 'LISTHSTRBRX'; svType: svInteger; Size: 4; Data: @LISTHSTRBRX),
   (Name: 'LISTHSTRBRY'; svType: svInteger; Size: 4; Data: @LISTHSTRBRY),
   (Name: 'LISTHTHP'; svType: svInteger; Size: 4; Data: @LISTHTHP),
   (Name: 'LISTHTHS'; svType: svInteger; Size: 4; Data: @LISTHTHS),
   (Name: 'LISTSIZETLX'; svType: svInteger; Size: 4; Data: @LISTSIZETLX),
   (Name: 'LISTSIZETLY'; svType: svInteger; Size: 4; Data: @LISTSIZETLY),
   (Name: 'LISTSIZEBRX'; svType: svInteger; Size: 4; Data: @LISTSIZEBRX),
   (Name: 'LISTSIZEBRY'; svType: svInteger; Size: 4; Data: @LISTSIZEBRY),
   (Name: 'CCOMMANDWINDOW'; svType: svBoolean; Size: 1; Data: @CCOMMANDWINDOW),
   (Name: 'CLISTWINDOW'; svType: svBoolean; Size: 1; Data: @CLISTWINDOW),
   (Name: 'COMMANDWINX'; svType: svInteger; Size: 4; Data: @COMMANDWINX),
   (Name: 'COMMANDWINY'; svType: svInteger; Size: 4; Data: @COMMANDWINY),
   (Name: 'COMMANDWINW'; svType: svInteger; Size: 4; Data: @COMMANDWINW),
   (Name: 'COMMANDWINH'; svType: svInteger; Size: 4; Data: @COMMANDWINH),
   (Name: 'COMMANDSIZETLX'; svType: svInteger; Size: 4; Data: @COMMANDSIZETLX),
   (Name: 'COMMANDSIZETLY'; svType: svInteger; Size: 4; Data: @COMMANDSIZETLY),
   (Name: 'COMMANDSIZEBRX'; svType: svInteger; Size: 4; Data: @COMMANDSIZEBRX),
   (Name: 'COMMANDSIZEBRY'; svType: svInteger; Size: 4; Data: @COMMANDSIZEBRY),
   (Name: 'SHOWLIST'; svType: svBoolean; Size: 1; Data: @SHOWLIST),
   (Name: 'TEMPDIR'; svType: svString; Size: 0; Data: @TEMPDIR),
   (Name: 'SCREENPOINTER'; svType: svPointer; Size: 4; Data: @SCREENPOINTER),
   (Name: 'SCREENBANK'; svType: svInteger; Size: 4; Data: @SCREENBANK),
   (Name: 'SCREENWIDTH'; svType: svInteger; Size: 4; Data: @SCREENWIDTH),
   (Name: 'SCREENHEIGHT'; svType: svInteger; Size: 4; Data: @SCREENHEIGHT),
   (Name: 'SCREENX'; svType: svInteger; Size: 4; Data: @SCREENX),
   (Name: 'SCREENY'; svType: svInteger; Size: 4; Data: @SCREENY),
   (Name: 'SCREENID'; svType: svInteger; Size: 4; Data: @SCREENID),
   (Name: 'WINDOWPOINTER'; svType: svPointer; Size: 4; Data: @WINDOWPOINTER),
   (Name: 'SCRBANKPOINTER'; svType: svPointer; Size: 4; Data: @SCRBANKPOINTER),
   (Name: 'FONTBANKID'; svType: svInteger; Size: 4; Data: @FONTBANKID),
   (Name: 'FONTWIDTH'; svType: svInteger; Size: 4; Data: @FONTWIDTH),
   (Name: 'FONTHEIGHT'; svType: svInteger; Size: 4; Data: @FONTHEIGHT),
   (Name: 'FONTTRANSPARENT'; svType: svInteger; Size: 4; Data: @FONTTRANSPARENT),
   (Name: 'FONTTYPE'; svType: svInteger; Size: 4; Data: @FONTTYPE),
   (Name: 'LOWERSAVE'; svType: svString; Size:0; Data: @LOWERSAVE),
   (Name: 'CPAPER'; svType: svInteger; Size: 4; Data: @CPAPER),
   (Name: 'CINK'; svType: svInteger; Size: 4; Data: @CINK),
   (Name: 'CINVERSE'; svType: svInteger; Size: 4; Data: @CINVERSE),
   (Name: 'COVER'; svType: svInteger; Size: 4; Data: @COVER),
   (Name: 'T_PAPER'; svType: svInteger; Size: 4; Data: @T_PAPER),
   (Name: 'T_INK'; svType: svInteger; Size: 4; Data: @T_INK),
   (Name: 'T_INVERSE'; svType: svInteger; Size: 4; Data: @T_INVERSE),
   (Name: 'T_OVER'; svType: svInteger; Size: 4; Data: @T_OVER),
   (Name: 'DRPOSX'; svType: svInteger; Size: 4; Data: @DRPOSX),
   (Name: 'DRPOSY'; svType: svInteger; Size: 4; Data: @DRPOSY),
   (Name: 'DRHEADING'; svType: svaFloat; Size: 8; Data: @DRHEADING),
   (Name: 'XORG'; svType: svInteger; Size: 4; Data: @XORG),
   (Name: 'YORG'; svType: svInteger; Size: 4; Data: @YORG),
   (Name: 'PRPOSX'; svType: svInteger; Size: 4; Data: @PRPOSX),
   (Name: 'PRPOSY'; svType: svInteger; Size: 4; Data: @PRPOSY),
   (Name: 'TABSIZE'; svType: svInteger; Size: 4; Data: @TABSIZE),
   (Name: 'SCROLLCNT'; svType: svInteger; Size: 4; Data: @SCROLLCNT),
   (Name: 'NXTLINE'; svType: svInteger; Size: 4; Data: @NXTLINE),
   (Name: 'NXTSTATEMENT'; svType: svInteger; Size: 4; Data: @NXTSTATEMENT),
   (Name: 'MATHMODE'; svType: svInteger; Size: 4; Data: @MATHMODE),
   (Name: 'BASE'; svType: svLongWord; Size: 4; Data: @BASE),
   (Name: 'SPLITSTATEMENTS'; svType: svBoolean; Size: 4; Data: @SPLITSTATEMENTS),
   (Name: 'SPLITREMS'; svType: svBoolean; Size: 1; Data: @SPLITREMS),
   (Name: 'EDITORWRAP'; svType: svBoolean; Size: 1; Data: @EDITORWRAP),
   (Name: 'NUBMODE1'; svType: svInteger; Size: 4; Data: @NUBMODE1),
   (Name: 'NUBMODE2'; svType: svInteger; Size: 4; Data: @NUBMODE1),
   (Name: 'BATTLEVEL'; svType: svInteger; Size: 4; Data: @BATTLEVEL),
   (Name: 'CBLACK'; svType: svLongWord; Size: 4; Data: @CBLACK),
   (Name: 'CBLUE'; svType: svLongWord; Size: 4; Data: @CBLUE),
   (Name: 'CRED'; svType: svLongWord; Size: 4; Data: @CRED),
   (Name: 'CMAGENTA'; svType: svLongWord; Size: 4; Data: @CMAGENTA),
   (Name: 'CGREEN'; svType: svLongWord; Size: 4; Data: @CGREEN),
   (Name: 'CCYAN'; svType: svLongWord; Size: 4; Data: @CCYAN),
   (Name: 'CYELLOW'; svType: svLongWord; Size: 4; Data: @CYELLOW),
   (Name: 'CWHITE'; svType: svLongWord; Size: 4; Data: @CWHITE),
   (Name: 'CGREY'; svType: svLongWord; Size: 4; Data: @CGREY),
   (Name: 'CBRIGHTBLUE'; svType: svLongWord; Size: 4; Data: @CBRIGHTBLUE),
   (Name: 'CBRIGHTRED'; svType: svLongWord; Size: 4; Data: @CBRIGHTRED),
   (Name: 'CBRIGHTMAGENTA'; svType: svLongWord; Size: 4; Data: @CBRIGHTMAGENTA),
   (Name: 'CBRIGHTGREEN'; svType: svLongWord; Size: 4; Data: @CBRIGHTGREEN),
   (Name: 'CBRIGHTCYAN'; svType: svLongWord; Size: 4; Data: @CBRIGHTCYAN),
   (Name: 'CBRIGHTYELLOW'; svType: svLongWord; Size: 4; Data: @CBRIGHTYELLOW),
   (Name: 'CBRIGHTWHITE'; svType: svLongWord; Size: 4; Data: @CBRIGHTWHITE),
   (Name: 'FDIST'; svType: svaFloat; Size: 8; Data: @FDIST),
   (Name: 'SORGX'; svType: svaFloat; Size: 8; Data: @SORGX),
   (Name: 'SORGY'; svType: svaFloat; Size: 8; Data: @SORGY),
   (Name: 'SORGW'; svType: svaFloat; Size: 8; Data: @SORGW),
   (Name: 'SORGH'; svType: svaFloat; Size: 8; Data: @SORGH),
   (Name: 'EDFONTSCALEX'; svType: svaFloat; Size: 8; Data: @EDFONTSCALEX),
   (Name: 'EDFONTSCALEY'; svType: svaFloat; Size: 8; Data: @EDFONTSCALEY));

  Function  SP_GetSysVarN(ID: aString; Var Error: TSP_ErrorCode): aFloat;
  Function  SP_GetSysVarS(ID: aString; Var Error: TSP_ErrorCode): aString;
  Procedure SP_SetSysVar(ID: aString; Val: aFloat; Str: aString; Var Error: TSP_ErrorCode);

implementation

Uses SP_Editor;

Function SP_FindSysVar(ID: aString): Integer;
Begin

  ID := Upper(ID);

  Result := 0;
  While Result <= High(SysVars) Do Begin
    If SysVars[Result].Name = ID Then
      Break;
    Inc(Result);
  End;

  If Result > High(SysVars) then
    Result := -1;

End;

Function SP_GetSysVarN(ID: aString; Var Error: TSP_ErrorCode): aFloat;
Var
  Idx: Integer;
Begin

  // Find the SysVar

  Result := 0.0;
  Idx := SP_FindSysVar(ID);

  If Idx >= 0 Then Begin

    Case SysVars[Idx].svType Of
      svString, svArray:
        Begin
          Error.Code := SP_ERR_INVALID_SYSVAR_TYPE;
        End;
      svBoolean:
        Begin
          Result := Byte(pBoolean(SysVars[Idx].Data)^);
        End;
      svLongWord:
        Begin
          Result := pLongWord(SysVars[Idx].Data)^;
        End;
      svaFloat:
        Begin
          Result := gaFloat(SysVars[Idx].Data);
        End;
      svInteger:
        Begin
          Result := pInteger(SysVars[Idx].Data)^;
        End;
      svPointer:
        Begin
          Result := pLongWord(SysVars[Idx].Data)^;
        End;
      svByte:
        Begin
          Result := pByte(SysVars[Idx].Data)^;
        End;
    End;

  End Else

    Error.Code := SP_ERR_INVALID_SYSVAR;

End;

Function SP_GetSysVarS(ID: aString; Var Error: TSP_ErrorCode): aString;
Var
  Idx: Integer;
  Ptr: pByte;
Begin

  // Find the SysVar

  Idx := SP_FindSysVar(ID);

  If Idx >= 0 Then Begin

    Case SysVars[Idx].svType Of
      svString:
        Begin
          Result := paString(SysVars[Idx].Data)^;
        End;
      svArray:
        Begin
          SetLength(Result, SysVars[Idx].Size);
          Ptr := pByte(SysVars[Idx].Data);
          For Idx := 0 To SysVars[Idx].Size -1 Do Begin
            pByte(@Result[Idx +1])^ := Ptr^;
            Inc(Ptr);
          End;
        End;
      svBoolean, svLongWord, svaFloat, svInteger, svPointer, svByte:
        Begin
          Error.Code := SP_ERR_INVALID_SYSVAR_TYPE;
        End;
    End;

  End Else

    Error.Code := SP_ERR_INVALID_SYSVAR;

End;

Procedure SP_SetSysVar(ID: aString; Val: aFloat; Str: aString; Var Error: TSP_ErrorCode);
Var
  Idx: Integer;
Begin

  // Find the SysVar.

  Idx := SP_FindSysVar(ID);

  // Set the data parameter according to the value passed

  If Idx >= 0 Then Begin

    Case SysVars[Idx].svType Of
      svString:
        Begin
          paString(SysVars[Idx].Data)^ := Str;
        End;
      svArray:
        Begin
          If Length(Str) = SysVars[Idx].Size Then
            CopyMem(SysVars[Idx].Data, @Str[1], SysVars[Idx].Size)
          Else
            Error.Code := SP_ERR_INVALID_SYSVAR_SIZE;
        End;
      svBoolean:
        Begin
          pBoolean(SysVars[Idx].Data)^ := Trunc(Val) <> 0;
        End;
      svLongWord:
        Begin
          If (Val >= 0) And (Val <= $FFFFFFFF) Then
            pLongWord(SysVars[Idx].Data)^ := Trunc(Val)
          Else
            Error.Code := SP_ERR_INTEGER_OUT_OF_RANGE;
        End;
      svaFloat:
        Begin
          WriteaFloat(SysVars[Idx].Data, Val);
        End;
      svInteger:
        Begin
          pInteger(SysVars[Idx].Data)^ := Trunc(Val);
        End;
      svPointer:
        Begin
          If (Val >= 0) And (Val <= $FFFFFFFF) Then
            pLongWord(SysVars[Idx].Data)^ := Trunc(Val)
          Else
            Error.Code := SP_ERR_INTEGER_OUT_OF_RANGE;
        End;
      svByte:
        Begin
          If (Val >= 0) And (Val <= $FF) Then
            pByte(SysVars[Idx].Data)^ := Trunc(Val) And 255
          Else
            Error.Code := SP_ERR_INTEGER_OUT_OF_RANGE;
        End;
    End;

  End Else
    Error.Code := SP_ERR_INVALID_SYSVAR;

  SP_MakeEditorList;

End;

end.
