// Copyright (C) 2010 By Paul Dunn
//
// This file is part of the SpecBAS BASIC Interpreter, which is in turn
// part of the SpecOS project.
//
// SpecBAS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// SpecBAS is distributed in the hope that it will be entertaining,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SpecBAS.  If not, see <http://www.gnu.org/licenses/>.

unit SP_Tokenise;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}
{$INCLUDE SpecBAS.inc}

interface

Uses SysUtils, Math, Classes, SP_Util;

Type
  TToken = Packed Record
    Token: Byte;
    Handler: Pointer;
    TokenPos: LongWord;
    TokenLen: LongWord;
  End;
  pToken = ^TToken;

Function  SP_TokeniseLine(Line: aString; IsExpression, AddLineNum: Boolean): aString;
Function  SP_IsConstant(Line: aString): Integer;
Function  SP_IsKeyWord(Line: aString): Integer;
Function  SP_IsFunction(Line: aString): Integer;
Function  SP_IsFunctionEx(Line: aString): Integer;
Function  SP_GetNumber(Line: aString; Var Idx: Integer; Var Number: aFloat): Boolean;
Function  SP_Detokenise(Tokens: aString; Var cPos: Integer; Highlight, UseDoubles: Boolean): aString;
Function  SP_Store_Line(Tokens: aString): Integer;
Procedure SP_Program_Add(Str: aString);
Procedure SP_Program_AddStrings(List: TStringlist);
Procedure SP_Program_AddStringsArray(Var List: Array of aString);
Procedure SP_Program_Insert(Index: Integer; Str: aString);
Procedure SP_Program_Delete(Index: Integer);
Procedure SP_Program_Delete_Line(Line: Integer);
Procedure SP_Program_Clear;

Var

  Idx_t: Integer;
  TempStr: aString;
  SP_KEYWORDS: Array of aString;
  SP_FUNCTIONS: Array of aString;
  SP_FUNCTIONS_EX: Array of aString;
  SP_OPTIMISE_FLAGS: Array of Boolean;
  SP_Program: Array of aString;
  SP_Program_Count: Integer;
  SP_User_Program_Count: Integer;

Const

  // List of keywords that are used in statements.
  // MUST Be in this order - add new commands to the end of the list.

  SP_KEYWORDS_EXTRA: Array[0..224] of aString =
    ('SPECTRUM ', 'PLAY ', 'AT ', 'TAB ', 'LINE ', ' THEN ', ' TO ', ' STEP ',
     'DEF ', 'CAT ', ' FORMAT ', 'MOVE ', 'ERASE ', 'OPEN ', 'CLOSE ', 'MERGE ', 'BEEP ',
     'CIRCLE ', 'INK ', 'PAPER ', 'INVERSE ', 'OUT ', 'STOP ', 'READ ', ' DATA ',
     'RESTORE ', 'NEW ', 'CONTINUE ', 'DIM ', 'REM ', 'FOR ', 'GO ', ' s** ', 'INPUT ',
     'LOAD ', 'LIST ', 'LET ', ' PAUSE ', 'NEXT ', 'POKE ', 'TEXT ', 'PLOT ', 'RUN ',
     'SAVE ', 'RANDOMIZE ', 'IF ', 'CLS ', 'DRAW ', 'CLEAR ', 'RETURN ', 'COPY ',
     'ELLIPSE ', 'CURVE ', 'INC ', 'DEC ', 'SWAP ', ' PALETTE ', 'DO ', 'WHILE ', 'UNTIL ',
     ' LOOP ', 'EXIT ', 'QUIT ', 'SCREEN ', 'LOCK ', 'UNLOCK ', 'UPDATE ', 'WINDOW ', 'SIZE ',
     'FRONT ', 'BACK ', 'SHOW ', 'HIDE ', 'GRAB ', 'PUT ', ' TRANSPARENT ', 'TRANS ', 'DPOKE ',
     'QPOKE ', 'OVER ', 'SCROLL ', 'ROLL ', 'ON ', 'EVERY ', 'OFF', 'ERROR ', 'YIELD', 'FULL ',
     'FILL ', 'RECTANGLE ', 'POLYGON ', 'DEGREES', 'RADIANS', 'RECOVER', 'FONT ', 'COLOUR',
     'MONO', ' BANK ', 'WAIT ', 'STREAM ', 'WRITE ','SEEK ', 'LABEL ', 'SETDIR ', 'MOUSEDOWN',
     'MOUSEUP', 'MOUSEMOVE', 'KEYDOWN', 'KEYUP', 'DEFAULT', 'STACK ', 'PUSH ', 'POP ', 'EXECUTE ',
     ' ROTATE ', 'FACE ', ' ASCII ', ' SAMPLE ', ' RATE ', ' VOLUME ', ' PAN ', ' BITS ',
     ' CHANNEL ', ' STEREO', ' RESUME', 'MUSIC ', ' ELSE ', 'END IF', 'END PROC', 'END STRUCT',
     ' PROC ', 'STRUCT ', 'DEF PROC ', 'DEF FN ', 'DEF STRUCT ', 'GLOBAL ', ' PROTECT ',
     ' UNPROTECT ', 'EXIT PROC ', 'REF ', 'POKE$ ', 'PRINT ', 'SCALE ', 'SPRITE ', ' ADDFRAME ',
     ' INSFRAME ', ' DELFRAME ', ' CW ', ' CCW ', 'WHEELUP', 'WHEELDOWN', 'GRAPHIC ', ' FLIP ',
     ' MIRROR ', ' REMAP ', ' DITHER ', 'FPOKE ', 'RENUMBER ', 'DELETE ', 'LAUNCH ', 'THREAD ',
     'PRIORITY ', 'PACKAGE ', ' a** ', 'ASSIGN ', ' ALL ', 'MAKEDIR ', 'RENAME ', 'CONST ',
     'CD ', 'TILEMAP ', ' SET ', ' VAR ', 'CURSOR', 'EACH ', 'MENU ', 'MOUSE ', 'DEBUG ', 'FPS ',
     ' BASE ', 'OPTION ', 'SETNUB ', 'SETCPU ', 'MULTIPLOT ', 'PROJECT3D ', 'TRANSFORM3D ',
     'TRANSFORM2D ', 'RAINBOW ', 'LOCAL ', 'CENTRE ', 'KEYBOARD ', 'ORIGIN ', 'CLIP ', ' CLONE ',
     'SHOWGUIDE ', 'HELP ', 'BORDER ', 'CASE ', 'WHEN ', 'OTHERWISE ', 'END CASE', 'INCLUDE ',
     'HALT', 'SORT ', 'KEY ', 'MAT ', 'ZER', 'IDN', 'CON', 'INV ', 'TRN ', 'ANIM ', 'FRAME ',
     'COLLIDE ', 'WRAP ', 'ZONE ', 'USING ', 'SPLIT ', ' BPP ', ' ALPHA ', ' DEPTH ', 'UNDIM ',
     'BOLD ', 'ITALIC ', 'FILTER ');

  // Constants used to quickly identify keywords when in token form. Each keyword listed
  // above has a corresponding constant, which must be SP_KEYWORD_BASE + (Index of Keyword above).

  SP_KEYWORD_BASE           = 1000;

  SP_KW_SPECTRUM            = 1000;
  SP_KW_PLAY                = 1001;
  SP_KW_AT                  = 1002;
  SP_KW_TAB                 = 1003;
  SP_KW_LINE                = 1004;
  SP_KW_THEN                = 1005;
  SP_KW_TO                  = 1006;
  SP_KW_STEP                = 1007;
  SP_KW_DEF                 = 1008;
  SP_KW_CAT                 = 1009;
  SP_KW_FORMAT              = 1010;
  SP_KW_MOVE                = 1011;
  SP_KW_ERASE               = 1012;
  SP_KW_OPEN                = 1013;
  SP_KW_CLOSE               = 1014;
  SP_KW_MERGE               = 1015;
  SP_KW_BEEP                = 1016;
  SP_KW_CIRCLE              = 1017;
  SP_KW_INK                 = 1018;
  SP_KW_PAPER               = 1019;
  SP_KW_INVERSE             = 1020;
  SP_KW_OUT                 = 1021;
  SP_KW_STOP                = 1022;
  SP_KW_READ                = 1023;
  SP_KW_DATA                = 1024;
  SP_KW_RESTORE             = 1025;
  SP_KW_NEW                 = 1026;
  SP_KW_CONTINUE            = 1027;
  SP_KW_DIM                 = 1028;
  SP_KW_REM                 = 1029;
  SP_KW_FOR                 = 1030;
  SP_KW_GO                  = 1031;
  SP_KW_SUB                 = 1032;
  SP_KW_INPUT               = 1033;
  SP_KW_LOAD                = 1034;
  SP_KW_LIST                = 1035;
  SP_KW_LET                 = 1036;
  SP_KW_PAUSE               = 1037;
  SP_KW_NEXT                = 1038;
  SP_KW_POKE                = 1039;
  SP_KW_TEXT                = 1040;
  SP_KW_PLOT                = 1041;
  SP_KW_RUN                 = 1042;
  SP_KW_SAVE                = 1043;
  SP_KW_RANDOMIZE           = 1044;
  SP_KW_IF                  = 1045;
  SP_KW_CLS                 = 1046;
  SP_KW_DRAW                = 1047;
  SP_KW_CLEAR               = 1048;
  SP_KW_RETURN              = 1049;
  SP_KW_COPY                = 1050;
  SP_KW_ELLIPSE             = 1051;
  SP_KW_CURVE               = 1052;
  SP_KW_INC                 = 1053;
  SP_KW_DEC                 = 1054;
  SP_KW_SWAP                = 1055;
  SP_KW_PALETTE             = 1056;
  SP_KW_DO                  = 1057;
  SP_KW_WHILE               = 1058;
  SP_KW_UNTIL               = 1059;
  SP_KW_LOOP                = 1060;
  SP_KW_EXIT                = 1061;
  SP_KW_QUIT                = 1062;
  SP_KW_SCREEN              = 1063;
  SP_KW_LOCK                = 1064;
  SP_KW_UNLOCK              = 1065;
  SP_KW_UPDATE              = 1066;
  SP_KW_WINDOW              = 1067;
  SP_KW_SIZE                = 1068;
  SP_KW_FRONT               = 1069;
  SP_KW_BACK                = 1070;
  SP_KW_SHOW                = 1071;
  SP_KW_HIDE                = 1072;
  SP_KW_GRAB                = 1073;
  SP_KW_PUT                 = 1074;
  SP_KW_TRANSPARENT         = 1075;
  SP_KW_TRANS               = 1076;
  SP_KW_DPOKE               = 1077;
  SP_KW_QPOKE               = 1078;
  SP_KW_OVER                = 1079;
  SP_KW_SCROLL              = 1080;
  SP_KW_ROLL                = 1081;
  SP_KW_ON                  = 1082;
  SP_KW_EVERY               = 1083;
  SP_KW_OFF                 = 1084;
  SP_KW_ERROR               = 1085;
  SP_KW_YIELD               = 1086;
  SP_KW_FULL                = 1087;
  SP_KW_FILL                = 1088;
  SP_KW_RECTANGLE           = 1089;
  SP_KW_POLYGON             = 1090;
  SP_KW_DEGREES             = 1091;
  SP_KW_RADIANS             = 1092;
  SP_KW_RECOVER             = 1093;
  SP_KW_FONT                = 1094;
  SP_KW_COLOUR              = 1095;
  SP_KW_MONO                = 1096;
  SP_KW_BANK                = 1097;
  SP_KW_WAIT                = 1098;
  SP_KW_STREAM              = 1099;
  SP_KW_WRITE               = 1100;
  SP_KW_SEEK                = 1101;
  SP_KW_LABEL               = 1102;
  SP_KW_SETDIR              = 1103;
  SP_KW_MOUSEDOWN           = 1104;
  SP_KW_MOUSEUP             = 1105;
  SP_KW_MOUSEMOVE           = 1106;
  SP_KW_KEYDOWN             = 1107;
  SP_KW_KEYUP               = 1108;
  SP_KW_DEFAULT             = 1109;
  SP_KW_STACK               = 1110;
  SP_KW_PUSH                = 1111;
  SP_KW_POP                 = 1112;
  SP_KW_EXECUTE             = 1113;
  SP_KW_ROTATE              = 1114;
  SP_KW_FACE                = 1115;
  SP_KW_ASCII               = 1116;
  SP_KW_SAMPLE              = 1117;
  SP_KW_RATE                = 1118;
  SP_KW_VOLUME              = 1119;
  SP_KW_PAN                 = 1120;
  SP_KW_BITS                = 1121;
  SP_KW_CHANNEL             = 1122;
  SP_KW_STEREO              = 1123;
  SP_KW_RESUME              = 1124;
  SP_KW_MUSIC               = 1125;
  SP_KW_ELSE                = 1126;
  SP_KW_ENDIF               = 1127;
  SP_KW_ENDPROC             = 1128;
  SP_KW_ENDSTRUCT           = 1129;
  SP_KW_PROC                = 1130;
  SP_KW_STRUCT              = 1131;
  SP_KW_DEF_PROC            = 1132;
  SP_KW_DEF_FN              = 1133;
  SP_KW_DEF_STRUCT          = 1134;
  SP_KW_GLOBAL              = 1135;
  SP_KW_PROTECT             = 1136;
  SP_KW_DEPROTECT           = 1137;
  SP_KW_EXIT_PROC           = 1138;
  SP_KW_REF                 = 1139;
  SP_KW_POKES               = 1140;
  SP_KW_PRINT               = 1141;
  SP_KW_SCALE               = 1142;
  SP_KW_SPRITE              = 1143;
  SP_KW_ADDFRAME            = 1144;
  SP_KW_INSFRAME            = 1145;
  SP_KW_DELFRAME            = 1146;
  SP_KW_CW                  = 1147;
  SP_KW_CCW                 = 1148;
  SP_KW_WHEELUP             = 1149;
  SP_KW_WHEELDOWN           = 1150;
  SP_KW_GRAPHIC             = 1151;
  SP_KW_FLIP                = 1152;
  SP_KW_MIRROR              = 1153;
  SP_KW_REMAP               = 1154;
  SP_KW_DITHER              = 1155;
  SP_KW_FPOKE               = 1156;
  SP_KW_RENUMBER            = 1157;
  SP_KW_DELETE              = 1158;
  SP_KW_LAUNCH              = 1159;
  SP_KW_THREAD              = 1160;
  SP_KW_PRIORITY            = 1161;
  SP_KW_PACKAGE             = 1162;
  SP_KW_ADD                 = 1163;
  SP_KW_ASSIGN              = 1164;
  SP_KW_ALL                 = 1165;
  SP_KW_MAKEDIR             = 1166;
  SP_KW_RENAME              = 1167;
  SP_KW_CONST               = 1168;
  SP_KW_CD                  = 1169;
  SP_KW_TILEMAP             = 1170;
  SP_KW_SET                 = 1171;
  SP_KW_VAR                 = 1172;
  SP_KW_CURSOR              = 1173;
  SP_KW_EACH                = 1174;
  SP_KW_MENU                = 1175;
  SP_KW_MOUSE               = 1176;
  SP_KW_DEBUG               = 1177;
  SP_KW_FPS                 = 1178;
  SP_KW_BASE                = 1179;
  SP_KW_OPTION              = 1180;
  SP_KW_SETNUB              = 1181;
  SP_KW_SETCPU              = 1182;
  SP_KW_MULTIPLOT           = 1183;
  SP_KW_PROJECT3D           = 1184;
  SP_KW_TRANSFORM3D         = 1185;
  SP_KW_TRANSFORM2D         = 1186;
  SP_KW_RAINBOW             = 1187;
  SP_KW_LOCAL               = 1188;
  SP_KW_CENTRE              = 1189;
  SP_KW_KEYBOARD            = 1190;
  SP_KW_ORIGIN              = 1191;
  SP_KW_CLIP                = 1192;
  SP_KW_CLONE               = 1193;
  SP_KW_SHOWGUIDE           = 1194;
  SP_KW_HELP                = 1195;
  SP_KW_BORDER              = 1196;
  SP_KW_CASE                = 1197;
  SP_KW_WHEN                = 1198;
  SP_KW_OTHERWISE           = 1199;
  SP_KW_END_CASE            = 1200;
  SP_KW_INCLUDE             = 1201;
  SP_KW_HALT                = 1202;
  SP_KW_SORT                = 1203;
  SP_KW_KEY                 = 1204;
  SP_KW_MAT                 = 1205;
  SP_KW_MAT_ZER             = 1206;
  SP_KW_MAT_IDN             = 1207;
  SP_KW_MAT_CON             = 1208;
  SP_KW_MAT_INV             = 1209;
  SP_KW_MAT_TRN             = 1210;
  SP_KW_ANIM                = 1211;
  SP_KW_FRAME               = 1212;
  SP_KW_COLLIDE             = 1213;
  SP_KW_WRAP                = 1214;
  SP_KW_ZONE                = 1215;
  SP_KW_USING               = 1216;
  SP_KW_SPLIT               = 1217;
  SP_KW_BPP                 = 1218;
  SP_KW_ALPHA               = 1219;
  SP_KW_DEPTH               = 1220;
  SP_KW_UNDIM               = 1221;
  SP_KW_BOLD                = 1222;
  SP_KW_ITALIC              = 1223;
  SP_KW_FILTER              = 1224;

  // These are meta-commands; they do not appear in listings, and are used during
  // execution only, having been inserted by the pre-processor.

  SP_KW_PR_INK              = 4051;
  SP_KW_PR_PAPER            = 4052;
  SP_KW_PR_INVERSE          = 4053;
  SP_KW_PR_TAB              = 4054;
  SP_KW_PR_AT               = 4055;
  SP_KW_PR_MOVE             = 4056;
  SP_KW_GOTO                = 4057;
  SP_KW_GOSUB               = 4058;
  SP_KW_PALSHIFT            = 4059;
  SP_KW_READ_ASSIGN         = 4060;
  SP_KW_DRAWTO              = 4061;
  SP_KW_SCR_LOCK            = 4062;
  SP_KW_SCR_UNLOCK          = 4063;
  SP_KW_SCR_UPDATE          = 4064;
  SP_KW_SCR_RES             = 4065;
  SP_KW_WIN_NEW             = 4066;
  SP_KW_WIN_DEL             = 4067;
  SP_KW_WIN_MOVE            = 4068;
  SP_KW_WIN_SIZE            = 4069;
  SP_KW_WIN_FRONT           = 4070;
  SP_KW_WIN_BACK            = 4071;
  SP_KW_WIN_SHOW            = 4072;
  SP_KW_WIN_HIDE            = 4073;
  SP_KW_SCR_GRAB            = 4074;
  SP_KW_WIN_GRAB            = 4075;
  SP_KW_PAL_HSV             = 4076;
  SP_KW_WIN_PUT             = 4077;
  SP_KW_PR_OVER             = 4078;
  SP_KW_WIN_SCROLL          = 4079;
  SP_KW_WIN_ROLL            = 4080;
  SP_KW_ONERR               = 4081;
  SP_KW_CLEAR_VAR           = 4082;
  SP_KW_CLEAR_ERR           = 4083;
  SP_KW_ON_ERR              = 4084;
  SP_KW_DBG                 = 4085;
  SP_KW_ON_ERR_OFF          = 4086;
  SP_KW_PR_TRANSPARENT      = 4087;
  SP_KW_FULLSCREEN          = 4088;
  SP_KW_WINDOWED            = 4089;
  SP_KW_POLYFILL            = 4090;
  SP_KW_POLYLINE            = 4091;
  SP_KW_RECTFILL            = 4092;
  SP_KW_CIRCLEFILL          = 4093;
  SP_KW_ELLIPSEFILL         = 4094;
  SP_KW_SCRCOPY             = 4095;
  SP_KW_FONT_NEW            = 4096;
  SP_KW_FONT_TRANS          = 4097;
  SP_KW_FONT_ERASE          = 4098;
  SP_KW_LOAD_BANK           = 4099;
  SP_KW_SAVE_BANK           = 4100;
  SP_KW_INCRANGE            = 4101;
  SP_KW_DECRANGE            = 4102;
  SP_KW_BANK_SIZE           = 4103;
  SP_KW_BANK_ERASE          = 4104;
  SP_KW_BANK_COPY           = 4105;
  SP_KW_STREAM_NEW          = 4106;
  SP_KW_STREAM_WRITE        = 4107;
  SP_KW_STREAM_READ         = 4108;
  SP_KW_STREAM_CLOSE        = 4109;
  SP_KW_STREAM_SEEK         = 4110;
  SP_KW_MOUSEDOWN_OFF       = 4111;
  SP_KW_MOUSEUP_OFF         = 4112;
  SP_KW_MOUSEMOVE_OFF       = 4113;
  SP_KW_KEYDOWN_OFF         = 4114;
  SP_KW_KEYUP_OFF           = 4115;
  SP_KW_ON_MOUSEDOWN        = 4116;
  SP_KW_ON_MOUSEUP          = 4117;
  SP_KW_ON_MOUSEMOVE        = 4118;
  SP_KW_ON_KEYDOWN          = 4119;
  SP_KW_ON_KEYUP            = 4120;
  SP_KW_PAL_LOAD            = 4121;
  SP_KW_PAL_SAVE            = 4122;
  SP_KW_PAL_DEFAULT         = 4123;
  SP_KW_STACK_NEW           = 4124;
  SP_KW_STACK_PUSH          = 4125;
  SP_KW_STACK_POP           = 4126;
  SP_KW_DRMOVETO            = 4127;
  SP_KW_ROTATETO            = 4128;
  SP_KW_DRAWTURTLE          = 4129;
  SP_KW_DRMOVEXY            = 4130;
  SP_KW_SAVEASCII           = 4131;
  SP_KW_SAMPLE_LOAD         = 4132;
  SP_KW_SAMPLE_PLAY         = 4133;
  SP_KW_SAMPLE_NEW          = 4134;
  SP_KW_SAMPLE_RATE         = 4135;
  SP_KW_SAMPLE_DEPTH        = 4136;
  SP_KW_SAMPLE_STEREO       = 4137;
  SP_KW_SAMPLE_VOLUME       = 4138;
  SP_KW_SAMPLE_PAN          = 4139;
  SP_KW_SAMPLE_ERASE        = 4140;
  SP_KW_CHANNEL_PAUSE       = 4141;
  SP_KW_CHANNEL_STOP        = 4142;
  SP_KW_CHANNEL_VOLUME      = 4143;
  SP_KW_CHANNEL_PAN         = 4144;
  SP_KW_CHANNEL_RESUME      = 4145;
  SP_KW_CHANNEL_SEEK        = 4146;
  SP_KW_CHANNEL_RATE        = 4147;
  SP_KW_MUSIC_PLAY          = 4148;
  SP_KW_MUSIC_PAUSE         = 4149;
  SP_KW_MUSIC_STOP          = 4150;
  SP_KW_MUSIC_VOLUME        = 4151;
  SP_KW_MUSIC_RESUME        = 4152;
  SP_KW_MUSIC_SEEK          = 4153;
  SP_KW_BANK_COPY_EX        = 4154;
  SP_KW_BANK_PROTECT        = 4155;
  SP_KW_BANK_DEPROTECT      = 4156;
  SP_KW_BANK_DEFAULT        = 4157;
  SP_KW_WIN_PUT_EX          = 4158;
  SP_KW_NEW_SPRITE          = 4159;
  SP_KW_SPRITE_ADDFR        = 4160;
  SP_KW_SPRITE_SHOW         = 4161;
  SP_KW_SPRITE_HIDE         = 4162;
  SP_KW_SPRITE_MOVE         = 4163;
  SP_KW_SPRITE_MOVED        = 4164;
  SP_KW_SPRITE_MOVE_T       = 4165;
  SP_KW_SPRITE_MOVED_T      = 4166;
  SP_KW_SPRITE_MOVE_S       = 4167;
  SP_KW_SPRITE_MOVED_S      = 4168;
  SP_KW_SPRITE_MOVE_WIN     = 4169;
  SP_KW_SPRITE_STOP         = 4170;
  SP_KW_SPRITE_ROT          = 4171;
  SP_KW_SPRITE_ROT_TO       = 4172;
  SP_KW_SPRITE_ROT_T        = 4173;
  SP_KW_SPRITE_ROT_TO_T     = 4174;
  SP_KW_SPRITE_SCALE        = 4175;
  SP_KW_SPRITE_SCALE_T      = 4176;
  SP_KW_SPRITE_ERASE        = 4177;
  SP_KW_SPRITE_CLEAR        = 4178;
  SP_KW_WHEELUP_OFF         = 4179;
  SP_KW_WHEELDOWN_OFF       = 4180;
  SP_KW_ON_WHEELUP          = 4181;
  SP_KW_ON_WHEELDOWN        = 4182;
  SP_KW_GFX_NEW_SIZE        = 4183;
  SP_KW_GFX_NEW_LOAD        = 4184;
  SP_KW_GFX_NEW_GFXS        = 4185;
  SP_KW_GFX_LOAD            = 4186;
  SP_KW_GFX_GRAB            = 4187;
  SP_KW_GFX_PUT             = 4188;
  SP_KW_GFX_PUT_EX          = 4189;
  SP_KW_GFX_ROTATE          = 4190;
  SP_KW_GFX_SCALE           = 4191;
  SP_KW_GFX_FLIP            = 4192;
  SP_KW_GFX_MIRROR          = 4193;
  SP_KW_GFX_ERASE           = 4194;
  SP_KW_GFX_TRANSPARENT     = 4195;
  SP_KW_GFX_REMAP           = 4196;
  SP_KW_GFX_PAL_DEFAULT     = 4197;
  SP_KW_GFX_PAL_HSV         = 4198;
  SP_KW_GFX_PALETTE         = 4199;
  SP_KW_PAL_COPY            = 4200;
  SP_KW_WINDOW_GFX          = 4201;
  SP_KW_READLINE            = 4202;
  SP_KW_ERASE_LINES         = 4203;
  SP_KW_LIST_LINES          = 4204;
  SP_KW_PACK_NEW            = 4205;
  SP_KW_PACK_OPEN           = 4206;
  SP_KW_PACK_CLOSE          = 4207;
  SP_KW_PACK_ADD            = 4208;
  SP_KW_MOVE_FILES          = 4209;
  SP_KW_PACK_PROTECT        = 4210;
  SP_KW_PACK_UNPROTECT      = 4211;
  SP_KW_DATA_LOAD           = 4212;
  SP_KW_DATA_SAVE           = 4213;
  SP_KW_MERGE_LINES         = 4214;
  SP_KW_GFX_ROLL            = 4215;
  SP_KW_GFX_SCROLL          = 4216;
  SP_KW_TILEMAP_NEW         = 4217;
  SP_KW_TILEMAP_DRAW        = 4218;
  SP_KW_TILEMAP_SET         = 4219;
  SP_KW_TILEMAP_GRAPHIC     = 4220;
  SP_KW_TILEMAP_CLEAR       = 4221;
  SP_KW_FILLTEX             = 4222;
  SP_KW_SAVE_BANK_DATA      = 4223;
  SP_KW_LIST_WINDOW         = 4224;
  SP_KW_LIST_BANK           = 4225;
  SP_KW_LIST_VAR            = 4226;
  SP_KW_LIST_ARRAY          = 4227;
  SP_KW_LIST_ASSIGN         = 4228;
  SP_KW_LIST_BANK_EX        = 4229;
  SP_KW_BANK_ERASE_ALL      = 4230;
  SP_KW_LIST_STRUCT         = 4231;
  SP_KW_STRUCT_ERASE        = 4232;
  SP_KW_INPUT_NUM           = 4233;
  SP_KW_INPUT_STR           = 4234;
  SP_KW_INIT_INPUT          = 4235;
  SP_KW_INPUT_FORMAT        = 4236;
  SP_KW_PR_CURSOR           = 4237;
  SP_KW_INPUT_FINAL         = 4238;
  SP_KW_INPUT_ERROR         = 4239;
  SP_KW_FOR_EACH            = 4240;
  SP_KW_MENU_NEW            = 4241;
  SP_KW_MENU_INSITEM        = 4242;
  SP_KW_MENU_DELITEM        = 4243;
  SP_KW_MENU_SUB            = 4244;
  SP_KW_MENU_DELSUB         = 4245;
  SP_KW_MENU_ERASE          = 4246;
  SP_KW_MOUSE_SHOW          = 4247;
  SP_KW_MOUSE_HIDE          = 4248;
  SP_KW_MOUSE_GRAPHIC       = 4249;
  SP_KW_MOUSE_GFXS          = 4250;
  SP_KW_MOUSE_DEFAULT       = 4251;
  SP_KW_PR_CENTRE           = 4252;
  SP_KW_WIN_ORIGIN          = 4253;
  SP_KW_GFX_ORIGIN          = 4254;
  SP_KW_WIN_ORG_NO_EXT      = 4255;
  SP_KW_GFX_ORG_NO_EXT      = 4256;
  SP_KW_WIN_ORG_OFF         = 4257;
  SP_KW_GFX_ORG_OFF         = 4258;
  SP_KW_PR_CLIP             = 4259;
  SP_KW_ORG_NO_EXT          = 4260;
  SP_KW_ORG_OFF             = 4261;
  SP_KW_WIN_CLIP            = 4262;
  SP_KW_GFX_CLIP            = 4263;
  SP_KW_WIN_CLIP_OFF        = 4264;
  SP_KW_GFX_CLIP_OFF        = 4265;
  SP_KW_SPRITE_CLONE        = 4266;
  SP_KW_SPRITE_CLONE_ERASE  = 4267;
  SP_KW_SPRITE_CLONE_MOVE   = 4268;
  SP_KW_DRAW_CTO            = 4269;
  SP_KW_OUT_SCREEN          = 4270;
  SP_KW_OUT_VAR             = 4271;
  SP_KW_OUT_STREAM          = 4272;
  SP_KW_PR_OUT_SCREEN       = 4273;
  SP_KW_PR_OUT_VAR          = 4274;
  SP_KW_PR_OUT_STREAM       = 4275;
  SP_KW_CLIP_OFF            = 4276;
  SP_KW_PR_CLIP_OFF         = 4277;
  SP_KW_FOR_EACH_RANGE      = 4278;
  SP_KW_WIN_MERGE           = 4279;
  SP_KW_WIN_MERGEALL        = 4280;
  SP_KW_AUTODIM             = 4281;
  SP_KW_WHEN_IN             = 4282;
  SP_KW_KEY_NUM             = 4283;
  SP_KW_KEY_STR             = 4284;
  SP_KW_KEY_CLEAR           = 4285;
  SP_KW_KEY_CLEAR_ALL       = 4286;
  SP_KW_DYNAMIC_DIM         = 4287;
  SP_KW_MAT_ADD             = 4288;
  SP_KW_MAT_SUB             = 4289;
  SP_KW_MAT_PROD            = 4290;
  SP_KW_MAT_SCALE           = 4291;
  SP_KW_MAT_IDN_PARAM       = 4292;
  SP_KW_MAT_CON_PARAM       = 4293;
  SP_KW_PR_SCALE            = 4294;
  SP_KW_SPRITE_POINT        = 4295;
  SP_KW_SPRITE_SET_ANIM     = 4296;
  SP_KW_SPRITE_ANIM_STOP    = 4297;
  SP_KW_SPRITE_FRAME        = 4298;
  SP_KW_SPRITE_MIRROR       = 4299;
  SP_KW_SPRITE_FLIP         = 4300;
  SP_KW_COLLIDE_OFF         = 4301;
  SP_KW_ON_COLLIDE          = 4302;
  SP_KW_ON_COLLIDE_OFF      = 4303;
  SP_KW_SPRITE_COLLIDE_CLEAR= 4304;
  SP_KW_NEW_BANK            = 4305;
  SP_KW_SPRITE_PUT          = 4306;
  SP_KW_SPRITE_OVER         = 4307;
  SP_KW_ZONE_NEW            = 4308;
  SP_KW_ZONE_ERASE          = 4309;
  SP_KW_LIST_ZONE           = 4310;
  SP_KW_ZONE_MOVE           = 4311;
  SP_KW_ZONE_MOVETO         = 4312;
  SP_KW_ZONE_UPDATE         = 4313;
  SP_KW_ZONE_NEW_B          = 4314;
  SP_KW_ZONE_UPDATE_B       = 4315;
  SP_KW_RECTANGLE2          = 4316;
  SP_KW_RECTFILL2           = 4317;
  SP_KW_PR_USING            = 4318;
  SP_KW_DIM_SPLIT           = 4319;
  SP_KW_PR_CENTRE_OFF       = 4320;
  SP_KW_WIN_TRANS           = 4321;
  SP_KW_WIN_ALPHA           = 4322;
  SP_KW_WIN_DEPTH           = 4323;
  SP_KW_ADRAWTO             = 4324;
  SP_KW_ADRAWTURTLE         = 4325;
  SP_KW_ACIRCLEFILL         = 4326;
  SP_KW_AELLIPSEFILL        = 4327;
  SP_KW_AFILLTEX            = 4328;
  SP_KW_ARECTFILL           = 4329;
  SP_KW_ARECTANGLE2         = 4330;
  SP_KW_ARECTFILL2          = 4331;
  SP_KW_APOLYLINE           = 4332;
  SP_KW_APOLYFILL           = 4333;
  SP_KW_ACIRCLE             = 4334;
  SP_KW_APLOT               = 4335;
  SP_KW_ADRAW               = 4336;
  SP_KW_AELLIPSE            = 4337;
  SP_KW_ACURVE              = 4338;
  SP_KW_AFILL               = 4339;
  SP_KW_ARECTANGLE          = 4340;
  SP_KW_APOLYGON            = 4341;
  SP_KW_AMULTIPLOT          = 4342;
  SP_KW_ADRAW_CTO           = 4343;
  SP_KW_APRINT              = 4344;
  SP_KW_ATEXT               = 4345;
  SP_KW_DO_COUNT            = 4346;
  SP_KW_SPRITE_COLL_CLR_ID  = 4347;
  SP_KW_CURVE_EX            = 4348;
  SP_KW_ACURVE_EX           = 4349;
  SP_KW_FILTER_STR          = 4350;
  SP_KW_FILTER_NUM_R        = 4351;
  SP_KW_FILTER_NUM          = 4352;

  // Names of the above meta-keywords - for use by the DEBUG command.

  SP_Keyword_Names: Array[0..300] of aString =
    ('PR INK', 'PR PAPER', 'PR INVERSE', 'PR TAB', 'PR AT', 'PR MOVE', 'GOTO', 'GOSUB', 'PALSHIFT',
     'READ ASSIGN', 'DRAWTO', 'SCR LOCK', 'SCR UNLOCK', 'SCR UPDATE', 'SCR RES', 'WIN NEW', 'WIN DEL',
     'WIN MOVE', 'WIN SIZE', 'WIN FRONT', 'WIN BACK', 'WIN SHOW', 'WIN HIDE', 'SCR GRAB', 'WIN GRAB',
     'PAL HSV', 'WIN PUT', 'PR OVER', 'WIN SCROLL', 'WIN ROLL', 'ONERR', 'CLEAR VAR', 'CLEAR ERR',
     'ON ERR', 'DBG', 'ON ERR OFF', 'PR TRANSPARENT', 'FULLSCREEN', 'WINDOWED', 'POLYFILL', 'POLYLINE',
     'RECTFILL', 'CIRCLEFILL', 'ELLIPSEFILL', 'SCRCOPY', 'FONT NEW', 'FONT TRANS', 'FONT ERASE',
     'LOAD BANK', 'SAVE BANK', 'INCRANGE', 'DECRANGE', 'BANK SIZE', 'BANK ERASE', 'BANK COPY',
     'STREAM NEW', 'STREAM WRITE', 'STREAM READ', 'STREAM CLOSE', 'STREAM SEEK', 'MOUSEDOWN OFF',
     'MOUSEUP OFF', 'MOUSEMOVE OFF', 'KEYDOWN OFF', 'KEYUP OFF', 'ON MOUSEDOWN', 'ON MOUSEUP', 'ON MOUSEMOVE',
     'ON KEYDOWN', 'ON KEYUP', 'PAL LOAD', 'PAL SAVE', 'PAL DEFAULT', 'STACK NEW', 'STACK PUSH',
     'STACK POP', 'DRMOVETO', 'ROTATETO', 'DRAWTURTLE', 'DRMOVEXY', 'SAVEASCII', 'SAMPLE LOAD', 'SAMPLE PLAY',
     'SAMPLE NEW', 'SAMPLE RATE', 'SAMPLE DEPTH', 'SAMPLE STEREO', 'SAMPLE VOLUME', 'SAMPLE PAN', 'SAMPLE ERASE',
     'CHANNEL PAUSE', 'CHANNEL STOP', 'CHANNEL VOLUME', 'CHANNEL PAN', 'CHANNEL RESUME', 'CHANNEL SEEK',
     'CHANNEL RATE', 'MUSIC PLAY', 'MUSIC PAUSE', 'MUSIC STOP', 'MUSIC VOLUME', 'MUSIC RESUME', 'MUSIC SEEK',
     'BANK COPY EX', 'BANK PROTECT', 'BANK DEPROTECT', 'BANK DEFAULT', 'WIN PUT EX', 'NEW SPRITE', 'SPRITE ADDFR',
     'SPRITE SHOW', 'SPRITE HIDE', 'SPRITE MOVE', 'SPRITE MOVED', 'SPRITE MOVE T', 'SPRITE MOVED T',
     'SPRITE MOVE S', 'SPRITE MOVED S', 'SPRITE MOVE WIN', 'SPRITE STOP', 'SPRITE ROT', 'SPRITE ROT TO',
     'SPRITE ROT T', 'SPRITE ROT TO T', 'SPRITE SCALE', 'SPRITE SCALE T', 'SPRITE ERASE', 'SPRITE CLEAR',
     'WHEELUP OFF', 'WHEELDOWN OFF', 'ON WHEELUP', 'ON WHEELDOWN', 'GFX NEW SIZE', 'GFX NEW LOAD', 'GFX NEW GFXS',
     'GFX LOAD', 'GFX GRAB', 'GFX PUT', 'GFX PUT EX', 'GFX ROTATE', 'GFX SCALE', 'GFX FLIP', 'GFX MIRROR',
     'GFX ERASE', 'GFX TRANSPARENT', 'GFX REMAP', 'GFX PAL DEFAULT', 'GFX PAL HSV', 'GFX PALETTE', 'PAL COPY',
     'WINDOW GFX', 'READLINE', 'ERASE LINES', 'LIST LINES', 'PACK NEW', 'PACK OPEN', 'PACK CLOSE', 'PACK ADD',
     'MOVE FILES', 'PACK PROTECT', 'PACK UNPROTECT', 'DATA LOAD', 'DATA SAVE', 'MERGE LINES', 'GFX ROLL',
     'GFX SCROLL', 'TILEMAP NEW', 'TILEMAP DRAW', 'TILEMAP SET', 'TILEMAP GRAPHIC', 'TILEMAP CLEAR', 'FILLTEX',
     'SAVE BANK DATA', 'LIST WINDOW', 'LIST BANK', 'LIST VAR', 'LIST ARRAY', 'LIST ASSIGN', 'LIST BANK EX',
     'BANK ERASE ALL', 'LIST STRUCT', 'STRUCT ERASE', 'INPUT NUM', 'INPUT STR', 'INIT INPUT', 'INPUT FORMAT',
     'PR CURSOR', 'INPUT FINAL', 'INPUT ERROR', 'FOR EACH', 'MENU NEW', 'MENU INSITEM', 'MENU DELITEM', 'MENU SUB',
     'MENU DELSUB', 'MENU ERASE', 'MOUSE SHOW', 'MOUSE HIDE', 'MOUSE GRAPHIC', 'MOUSE GFXS', 'MOUSE DEFAULT',
     'PR CENTRE', 'WIN ORIGIN', 'GFX ORIGIN', 'WIN ORG NO EXT', 'GFX ORG NO EXT', 'WIN ORG OFF', 'GFX ORG OFF',
     'PR CLIP', 'ORG NO EXT', 'ORG OFF', 'WIN CLIP', 'GFX CLIP', 'WIN CLIP OFF', 'GFX CLIP OFF', 'SPRITE CLONE',
     'SPRITE CLONE ERASE', 'SPRITE CLONE MOVE', 'DRAW CTO', 'OUT SCREEN', 'OUT VAR', 'OUT STREAM', 'PR OUT SCREEN',
     'PR OUT VAR', 'PR OUT STREAM', 'CLIP OFF', 'PR CLIP OFF', 'FOR EACH RANGE', 'WIN MERGE', 'WIN MERGEALL',
     'AUTODIM', 'WHEN IN', 'KEY NUM', 'KEY STR', 'KEY CLEAR', 'KEY CLEAR ALL', 'DYNAMIC DIM', 'MAT ADD',
     'MAT SUB', 'MAT PROD', 'MAT SCALE', 'MAT IDN PARAM', 'MAT CON PARAM', 'PR SCALE', 'SPRITE POINT',
     'SPRITE SET ANIM', 'SPRITE ANIM STOP', 'SPRITE FRAME', 'SPRITE MIRROR', 'SPRITE FLIP', 'COLLIDE OFF',
     'ON COLLIDE', 'ON COLLIDE OFF', 'SPRITE COLLIDE CLEAR', 'NEW BANK', 'SPRITE PUT', 'SPRITE OVER',
     'ZONE NEW', 'ZONE ERASE', 'LIST ZONE', 'ZONE MOVE', 'ZONE MOVE TO', 'ZONE UPDATE', 'ZONE NEW SIZE',
     'ZONE UPDATE SIZE', 'RECTANGLE_S', 'RECTFILL_S', 'PR USING', 'DIM SPLIT', 'PR CENTRE OFF', 'WIN TRANSPARENT',
     'WIN ALPHA', 'WIN DEPTH', 'A-DRAWTO', 'A-DRAW TURTLE', 'A-CIRCLE FILL', 'A-ELLIPSE FILL', 'A-TEXTURE FILL',
     'A-RECTFILL', 'A-RECTANGLE2', 'A-RECTFILL2', 'A-POLY LINE', 'A-POLY FILL', 'A-CIRCLE', 'A-PLOT', 'A_DRAW',
     'A-ELLIPSE', 'A-CURVE', 'A-FILL', 'A-RECTANGLE', 'A-PLOYGON', 'A-MULTIPLOT', 'A-PRINT', 'A-TEXT',
     'DO COUNT', 'SP COLL CLR ID', 'CURVEX', 'A-CURVEX', 'FILTER_STR', 'FILTER_NUM_R', 'FILTER_NUM');

  // List of Functions that are used in expressions. Again, MUST be in order.

  SP_FUNCTIONS_EXTRA: Array[0..228] of aString =
    ('nRND', 'nINKEY$', 'oPI', 'nVAL$ ', 'oCODE ', 'oVAL ', 'oLEN ', 'nSIN ', 'nCOS ',
     'nTAN ', 'nASN ', 'nACS ', 'nATN ', 'oLN ', 'oEXP ', 'oINT ', 'oSQR ', 'oSGN ', 'oABS ', 'n IN ',
     'nUSR ', 'oSTR$ ','oCHR$ ', 'nPEEK ', 'oNOT ', 'o OR ', 'o AND ', 'o MOD ', 'o XOR ', 'o SHL ',
     'o SHR ', 'oFRAC ', 'oCEIL ', 'oFLOOR ', 'oMAX', 'oMIN', 'oROUND', 'oODD ', 'oEVEN ', 'oPOS',
     'oTRUNC ', 'oLEFT$', 'oRIGHT$', 'oMID$', 'oUP$ ', 'oLOW$ ', 'oREP$', 'oRED ', 'oGREEN ',
     'oBLUE ', 'nRGB ', 'oHEX$ ', 'oBIN$ ', 'oRGBf ', 'oRGBn ', 'nKEYST ', 'nHSV ', 'oRGBc ', 'oHUE ',
     'oSATURATION ', 'oVALUE ', 'nWINw', 'nWINh', 'nSCRw', 'nSCRh', 'ncWIN', 'oRGBtoINT', 'oHSVtoINT',
     'oRGBtoHSV ', 'oHSVtoRGB ', 'oTRIM$ ', 'oLTRIM$ ', 'oRTRIM$ ', 'nUDG ', 'oUDG$ ', 'nDPEEK ',
     'nQPEEK ', 'nMOUSEx', 'nMOUSEy', 'nMOUSEBTN', 'nERRORNUM', 'nERRORLINE', 'nERRORSTATEMENT',
     'oDEGTORAD ', 'oRADTODEG ', 'nFONTBANK', 'nFONTw', 'nFONTh', 'nFONTm', 'nFONTt', 'nSTREAMLEN ',
     'nSTREAMPOS ', 'nBANKLEN ', 'nGETDIR', 'nMOUSEdx', 'nMOUSEdy', 'oDCHR$ ', 'oQCHR$ ', 'oDCODE ',
     'oQCODE ', 'oTOKEN$ ', 'nHEADING', 'nDRPOSX', 'nDRPOSY', 'nCHPOS ', 'nWINOFF ', 'nCALL ', 'nPEEK$ ',
     'nLASTK', 'nSPFRADDR ', 'nSPRITEx ', 'nSPRITEy ', 'nSPRITEw ', 'nSPRITEh ', 'nSPRITEv ',
     'nSPFCOUNT ', 'nSPROT ', 'nSPSCALE ', 'nFRAMES', 'nTIME', 'oDAY$ ', 'oMONTH$ ', 'oYEAR ',
     'oHOUR ', 'oMINUTES ', 'oSECONDS ', 'oDAY ', 'oMONTH ', 'nMOUSEWHEEL', 'oGFX$ ', 'nGRGB ',
     'nGHSV ', 'nWINX ', 'nWINY ', 'nGFXw ', 'nGFXh ', 'nGFXTRANS ', 'nGPOINT ', 'nFPEEK ', 'nITEM',
     'oFCODE ', 'oFCHR$ ', 'oPOWER', 'nDIR ', 'nPOINT', 'nFN ', 'nGETTILE', 'oPOWERTWO ', 'oLOGTWO ',
     'oWORDSWAP ', 'oBYTESWAP ', 'oNYBBLESWAP ', 'nPOLAR ', 'oCLAMP ', 'oINRANGE ', 'oINSERT$ ',
     'oITEM$ ', 'oBIT ', 'oHIWORD', 'oLOWORD', 'oHIBYTE', 'oLOBYTE', 'oPOLARDIST ', 'oLPAD$ ', 'oRPAD$ ',
     'nPROCID ', 'nPROCID$ ', 'oBTSET', 'oBTCLR', 'nERROR$', 'nMENUBOX', 'nPOPLINE', 'nPOPST', 'nVOL',
     'nGETOPT ', 'nGETOPT$ ', 'nNUBMODE ', 'nNUBX ', 'nNUBY ', 'nFILEEXISTS ', 'oFPATH ', 'oFNAME ',
     'nDIREXISTS ', 'oPYTH', 'nLOGW', 'nLOGH', 'nORGX', 'nORGY', 'nPTOLX ', 'nPTOLY ', 'nLTOPX ',
     'nLTOPY ', 'nmSECS', 'oINV ', 'oPOLYTERM', 'oUNDER', 'nCOMPSIMPSON', 'nMUSICPOS', 'nMUSICLEN',
     'oBASE$', 'oIIF', 'oIIF$', 'nSEARCH', 'o MUL ', 'o DIV ', 'o ADD ', 'o SUB ', 'oGCD', 'oLCM', 'oDET',
     'oBIN ', 'nSPFRAME ', 'nSPCOLL ', 'nTEXTURE$ ', 'nINZONE', 'nMATCH', 'oDECIMAL', 'oUSING$', 'oJOIN$',
     'oLBOUND', 'oUBOUND', 'oARSIZE', 'oMANDEL', 'oHEX', 'nSCREEN$', 'nDATE$ ', 'nTIME$ ', 'oREPLACE$',
     'oIVAL ');

  // Constants, like above, for identifying Functions in token form

  SP_FUNCTION_BASE          = 2000;

  SP_FN_RND                 = 2000;
  SP_FN_INKEYS              = 2001;
  SP_FN_PI                  = 2002;
  SP_FN_VALS                = 2003;
  SP_FN_CODE                = 2004;
  SP_FN_VAL                 = 2005;
  SP_FN_LEN                 = 2006;
  SP_FN_SIN                 = 2007;
  SP_FN_COS                 = 2008;
  SP_FN_TAN                 = 2009;
  SP_FN_ASN                 = 2010;
  SP_FN_ACS                 = 2011;
  SP_FN_ATN                 = 2012;
  SP_FN_LN                  = 2013;
  SP_FN_EXP                 = 2014;
  SP_FN_INT                 = 2015;
  SP_FN_SQR                 = 2016;
  SP_FN_SGN                 = 2017;
  SP_FN_ABS                 = 2018;
  SP_FN_IN                  = 2019;
  SP_FN_USR                 = 2020;
  SP_FN_STRS                = 2021;
  SP_FN_CHRS                = 2022;
  SP_FN_PEEK                = 2023;
  SP_FN_NOT                 = 2024;
  SP_FN_OR                  = 2025;
  SP_FN_AND                 = 2026;
  SP_FN_MOD                 = 2027;
  SP_FN_XOR                 = 2028;
  SP_FN_SHL                 = 2029;
  SP_FN_SHR                 = 2030;
  SP_FN_FRAC                = 2031;
  SP_FN_CEIL                = 2032;
  SP_FN_FLOOR               = 2033;
  SP_FN_MAX                 = 2034;
  SP_FN_MIN                 = 2035;
  SP_FN_ROUND               = 2036;
  SP_FN_ODD                 = 2037;
  SP_FN_EVEN                = 2038;
  SP_FN_POS                 = 2039;
  SP_FN_TRUNC               = 2040;
  SP_FN_LEFTS               = 2041;
  SP_FN_RIGHTS              = 2042;
  SP_FN_MIDS                = 2043;
  SP_FN_UPS                 = 2044;
  SP_FN_LOWS                = 2045;
  SP_FN_REPS                = 2046;
  SP_FN_RED                 = 2047;
  SP_FN_GREEN               = 2048;
  SP_FN_BLUE                = 2049;
  SP_FN_RGB                 = 2050;
  SP_FN_HEXS                = 2051;
  SP_FN_BINS                = 2052;
  SP_FN_RGBF                = 2053;
  SP_FN_RGBN                = 2054;
  SP_FN_KEY                 = 2055;
  SP_FN_HSV                 = 2056;
  SP_FN_RGBC                = 2057;
  SP_FN_HUE                 = 2058;
  SP_FN_SAT                 = 2059;
  SP_FN_VALUE               = 2060;
  SP_FN_WINW                = 2061;
  SP_FN_WINH                = 2062;
  SP_FN_SCRW                = 2063;
  SP_FN_SCRH                = 2064;
  SP_FN_CWIN                = 2065;
  SP_FN_RGBtoINT            = 2066;
  SP_FN_HSVtoINT            = 2067;
  SP_FN_RGBToHSV            = 2068;
  SP_FN_HSVToRGB            = 2069;
  SP_FN_TRIMS               = 2070;
  SP_FN_LTRIMS              = 2071;
  SP_FN_RTRIMS              = 2072;
  SP_FN_UDG                 = 2073;
  SP_FN_UDGS                = 2074;
  SP_FN_DPEEK               = 2075;
  SP_FN_QPEEK               = 2076;
  SP_FN_MOUSEX              = 2077;
  SP_FN_MOUSEY              = 2078;
  SP_FN_MOUSEBTN            = 2079;
  SP_FN_ERROR_NUM           = 2080;
  SP_FN_ERROR_LIN           = 2081;
  SP_FN_ERROR_STA           = 2082;
  SP_FN_DEGTORAD            = 2083;
  SP_FN_RADTODEG            = 2084;
  SP_FN_FONTBANK            = 2085;
  SP_FN_FONTWIDTH           = 2086;
  SP_FN_FONTHEIGHT          = 2087;
  SP_FN_FONTMODE            = 2088;
  SP_FN_FONTTRANSPARENT     = 2089;
  SP_FN_SLEN                = 2090;
  SP_FN_SPOS                = 2091;
  SP_FN_BSIZE               = 2092;
  SP_FN_GETDIR              = 2093;
  SP_FN_MOUSEDX             = 2094;
  SP_FN_MOUSEDY             = 2095;
  SP_FN_DCHRS               = 2096;
  SP_FN_QCHRS               = 2097;
  SP_FN_DCODE               = 2098;
  SP_FN_QCODE               = 2099;
  SP_FN_TOKENS              = 2100;
  SP_FN_HEADING             = 2101;
  SP_FN_DRPOSX              = 2102;
  SP_FN_DRPOSY              = 2103;
  SP_FN_CHPOS               = 2104;
  SP_FN_WINOFF              = 2105;
  SP_FN_CALL                = 2106;
  SP_FN_PEEKS               = 2107;
  SP_FN_LASTK               = 2108;
  SP_FN_SPFRADDR            = 2109;
  SP_FN_SPRITEX             = 2110;
  SP_FN_SPRITEY             = 2111;
  SP_FN_SPRITEW             = 2112;
  SP_FN_SPRITEH             = 2113;
  SP_FN_SPVISIBLE           = 2114;
  SP_FN_SPFCOUNT            = 2115;
  SP_FN_SPROT               = 2116;
  SP_FN_SPSCALE             = 2117;
  SP_FN_FRAMES              = 2118;
  SP_FN_TIME                = 2119;
  SP_FN_DAYS                = 2120;
  SP_FN_MONTHS              = 2121;
  SP_FN_YEAR                = 2122;
  SP_FN_HOUR                = 2123;
  SP_FN_MINUTES             = 2124;
  SP_FN_SECONDS             = 2125;
  SP_FN_DAY                 = 2126;
  SP_FN_MONTH               = 2127;
  SP_FN_MOUSEWHEEL          = 2128;
  SP_FN_GFXS                = 2129;
  SP_FN_GRGB                = 2130;
  SP_FN_GHSV                = 2131;
  SP_FN_WINX                = 2132;
  SP_FN_WINY                = 2133;
  SP_FN_GFXW                = 2134;
  SP_FN_GFXH                = 2135;
  SP_FN_GFXT                = 2136;
  SP_FN_GPOINT              = 2137;
  SP_FN_FPEEK               = 2138;
  SP_FN_ITEM                = 2139;
  SP_FN_FCODE               = 2140;
  SP_FN_FCHRS               = 2141;
  SP_FN_POWER               = 2142;
  SP_FN_DIR                 = 2143;
  SP_FN_POINT               = 2144;
  SP_FN_FN                  = 2145;
  SP_FN_GETTILE             = 2146;
  SP_FN_POWERTWO            = 2147;
  SP_FN_LOGTWO              = 2148;
  SP_FN_WORDSWAP            = 2149;
  SP_FN_BYTESWAP            = 2150;
  SP_FN_NYBBLESWAP          = 2151;
  SP_FN_POLAR               = 2152;
  SP_FN_CLAMP               = 2153;
  SP_FN_INRANGE             = 2154;
  SP_FN_INSERTS             = 2155;
  SP_FN_ITEMS               = 2156;
  SP_FN_BIT                 = 2157;
  SP_FN_HIWORD              = 2158;
  SP_FN_LOWORD              = 2159;
  SP_FN_HIBYTE              = 2160;
  SP_FN_LOBYTE              = 2161;
  SP_FN_POLARDIST           = 2162;
  SP_FN_LPADS               = 2163;
  SP_FN_RPADS               = 2164;
  SP_FN_PROCID              = 2165;
  SP_FN_PROCIDS             = 2166;
  SP_FN_BTSET               = 2167;
  SP_FN_BTCLR               = 2168;
  SP_FN_ERRORS              = 2169;
  SP_FN_MENUBOX             = 2170;
  SP_FN_POPLINE             = 2171;
  SP_FN_POPST               = 2172;
  SP_FN_VOL                 = 2173;
  SP_FN_GETOPT              = 2174;
  SP_FN_GETOPTS             = 2175;
  SP_FN_NUBMODE             = 2176;
  SP_FN_NUBX                = 2177;
  SP_FN_NUBY                = 2178;
  SP_FN_FEXISTS             = 2179;
  SP_FN_FPATH               = 2180;
  SP_FN_FNAME               = 2181;
  SP_FN_DEXISTS             = 2182;
  SP_FN_PYTH                = 2183;
  SP_FN_LOGW                = 2184;
  SP_FN_LOGH                = 2185;
  SP_FN_ORGX                = 2186;
  SP_FN_ORGY                = 2187;
  SP_FN_PTOLX               = 2188;
  SP_FN_PTOLY               = 2189;
  SP_FN_LTOPX               = 2190;
  SP_FN_LTOPY               = 2191;
  SP_FN_MSECS               = 2192;
  SP_FN_INV                 = 2193;
  SP_FN_POLYTERM            = 2194;
  SP_FN_UNDER               = 2195;
  SP_FN_COMPSIMPSON         = 2196;
  SP_FN_MUSICPOS            = 2197;
  SP_FN_MUSICLEN            = 2198;
  SP_FN_BASES               = 2199;
  SP_FN_IIF                 = 2200;
  SP_FN_IIFS                = 2201;
  SP_FN_SEARCH              = 2202;
  SP_FN_MUL                 = 2203;
  SP_FN_DIV                 = 2204;
  SP_FN_ADD                 = 2205;
  SP_FN_SUB                 = 2206;
  SP_FN_GCD                 = 2207;
  SP_FN_LCM                 = 2208;
  SP_FN_DET                 = 2209;
  SP_FN_BIN                 = 2210;
  SP_FN_SPFRAME             = 2211;
  SP_FN_SPCOLL              = 2212;
  SP_FN_TEXTURES            = 2213;
  SP_FN_INZONE              = 2214;
  SP_FN_MATCH               = 2215;
  SP_FN_DECIMAL             = 2216;
  SP_FN_USINGS              = 2217;
  SP_FN_JOINS               = 2218;
  SP_FN_LBOUND              = 2219;
  SP_FN_UBOUND              = 2220;
  SP_FN_ARSIZE              = 2221;
  SP_FN_MANDEL              = 2222;
  SP_FN_HEX                 = 2223;
  SP_FN_SCREENS             = 2224;
  SP_FN_DATES               = 2225;
  SP_FN_TIMES               = 2226;
  SP_FN_REPLACES            = 2227;
  SP_FN_IVAL                = 2228;

  // Meta-functions

  SP_Function_Names: Array[0..6] of aString =
   ('SEARCH_NEXT', 'SP_MENUBOX_EX', 'SP_iRGBtoHSV', 'SP_iHSVtoRGB', 'SP_iRGBf', 'SP_iRGBn', 'SP_POSN');

  SP_FN_SEARCH_NEXT         = 3000;
  SP_FN_MENUBOX_EX          = 3001;
  SP_FN_iRGBtoHSV           = 3002;
  SP_FN_iHSVtoRGB           = 3003;
  SP_FN_iRGBf               = 3004;
  SP_FN_iRGBn               = 3005;
  SP_FN_POSN                = 3006;

  // List of token types.

  SP_KEYWORD                = 0;
  SP_VALUE10                = 1;
  SP_STRING                 = 2;
  SP_NUMVAR                 = 3;
  SP_STRVAR                 = 4;
  SP_TEXT                   = 5;
  SP_SYMBOL                 = 6;
  SP_FUNCTION               = 7;
  SP_UNKNOWN                = 8;
  SP_OPERATOR               = 9;
  SP_COMMA                  = 10;
  SP_ARRAYPARAM             = 11;
  SP_FUNCTION_EX            = 12;
  SP_LINE_NUM               = 13;
  SP_STATEMENTS             = 14;
  SP_NUMVAR_EVAL            = 15;
  SP_STRVAR_EVAL            = 16;
  SP_ARRAY                  = 17;
  SP_SLICER                 = 18;
  SP_ARRAY_ASSIGN           = 19;
  SP_SLICE_ASSIGN           = 20;
  SP_JUMP                   = 21;
  SP_POINTER                = 22;
  SP_SKIP_STATEMENT         = 23;
  SP_SKIP_LINE              = 24;
  SP_DATA_ITEM              = 25;
  SP_LABEL                  = 26;
  SP_LINE_LEN               = 27;
  SP_DISPLACEMENT           = 28;
  SP_SLICE_VAR              = 29;
  SP_VALUE                  = 30;
  SP_STRINGCHAR             = 31;
  SP_PARAMETER              = 32;
  SP_NUMVAR_LET             = 33;
  SP_STRVAR_LET             = 34;
  SP_STRUCT_MEMBER_N        = 35;
  SP_STRUCT_MEMBER_ASS      = 36;
  SP_STRUCT_MEMBER_S        = 37;
  SP_STRVARPTR              = 38;
  SP_NUMVAR_LET_VALID       = 39;
  SP_STRVAR_LET_VALID       = 40;
  SP_SKIP_DATA              = 41;
  SP_RANGE_LESS             = 42;
  SP_RANGE_GREATER          = 43;
  SP_RANGE                  = 44;
  SP_FUNCTION_MARKER        = 45;
  SP_COMPOUND_NNO           = 46;
  SP_COMPOUND_NVO           = 47;
  SP_COMPOUND_VNO           = 48;
  SP_SPECIAL_SYMBOL         = 49;
  SP_INTEGER                = 50;
  SP_EACHSINGLE             = 51;
  SP_EACHRANGE              = 52;
  SP_EACHSTEP               = 53;
  SP_NUMCONST               = 54;
  SP_STRCONST               = 55;
  SP_INTVAR                 = 56;
  SP_EXIT                   = -1;

  SP_TERMINAL               = 255;
  SP_TERMINAL_CHAR          = #255;
  SP_TERMINAL_SEQUENCE      = #255+#255+#255+#255;

  SP_FUNCTION_ID            = #255;
  SP_CHAR_GTE               = #254;
  SP_CHAR_LTE               = #253;
  SP_CHAR_DNE               = #252;
  SP_CHAR_AND               = #251;
  SP_CHAR_OR                = #250;
  SP_CHAR_UNARYM            = #249;
  SP_CHAR_UNARYP            = #248;
  SP_CHAR_MOD               = #247;
  SP_CHAR_XOR               = #246;
  SP_CHAR_SHL               = #245;
  SP_CHAR_SHR               = #244;
  SP_CHAR_STR_PLUS          = #243;
  SP_CHAR_NUM_PLUS          = #242;
  SP_CHAR_STR_EQU           = #241;
  SP_CHAR_NUM_EQU           = #240;
  SP_CHAR_NUM_LES           = #239;
  SP_CHAR_STR_LES           = #238;
  SP_CHAR_NUM_LTE           = #237;
  SP_CHAR_STR_LTE           = #236;
  SP_CHAR_NUM_DNE           = #235;
  SP_CHAR_STR_DNE           = #234;
  SP_CHAR_NUM_GTE           = #233;
  SP_CHAR_STR_GTE           = #232;
  SP_CHAR_NUM_GTR           = #231;
  SP_CHAR_STR_GTR           = #230;
  SP_CHAR_NUM_AND           = #229;
  SP_CHAR_STR_AND           = #228;
  SP_CHAR_STR_MUL1          = #227;
  SP_CHAR_STR_MUL2          = #226;
  SP_CHAR_DIV               = #225;
  SP_CHAR_MUL               = #224;
  SP_CHAR_ADD               = #223;
  SP_CHAR_SUB               = #222;
  SP_CHAR_INT_PLUS          = #221;

  SP_CHAR_PLUS              = '+';
  SP_CHAR_MINUS             = '-';
  SP_CHAR_MULTIPLY          = '*';
  SP_CHAR_DIVIDE            = '/';
  SP_CHAR_POWER             = '^';
  SP_CHAR_COLON             = ':';
  SP_CHAR_EQUALS            = '=';
  SP_CHAR_GREATER           = '>';
  SP_CHAR_LESS              = '<';
  SP_CHAR_BIT_OR            = '|';
  SP_CHAR_BIT_AND           = '&';
  SP_CHAR_EXCLAIM           = '!';
  SP_CHAR_TILDE             = '~';

  SP_PR_STOP                = 2;
  SP_PR_RUN                 = 1;

  SP_TRUE                   = 1;
  SP_FALSE                  = 0;

implementation

Uses SP_Main, SP_Editor, SP_FileIO, SP_SysVars;

Function SP_TokeniseLine(Line: aString; IsExpression, AddLineNum: Boolean): aString;
Type
  TCompoundWord = Record
    Name: aString;
    KeyWordID: LongWord;
  End;
Var
  TempExtend: aFloat;
  Keyword, SpaceMod: Integer;
  StoreVal, StoreLen, StatementCount,
  TempVal, LineLen, Idx, tIdx, lIdx: Integer;
  StoreText, Statements, tStore, tempStr, tStr: aString;
  Done, FoundBase: Boolean;
Label
  Finish;
Const
  Compounds: Array[0..15] of TCompoundWord =
    ((Name: 'DEFPROC'; KeyWordID: SP_KW_DEF_PROC),
     (Name: 'DEFFN'; KeyWordID: SP_KW_DEF_FN),
     (Name: 'DEFSTRUCT'; KeyWordID: SP_KW_DEF_STRUCT),
     (Name: 'DEF PROC'; KeyWordID: SP_KW_DEF_PROC),
     (Name: 'DEF FN'; KeyWordID: SP_KW_DEF_FN),
     (Name: 'DEF STRUCT'; KeyWordID: SP_KW_DEF_STRUCT),
     (Name: 'ENDPROC'; KeyWordID: SP_KW_ENDPROC),
     (Name: 'ENDIF'; KeyWordID: SP_KW_ENDIF),
     (Name: 'ENDSTRUCT'; KeyWordID: SP_KW_ENDSTRUCT),
     (Name: 'END PROC'; KeyWordID: SP_KW_ENDPROC),
     (Name: 'END IF'; KeyWordID: SP_KW_ENDIF),
     (Name: 'END STRUCT'; KeyWordID: SP_KW_ENDSTRUCT),
     (Name: 'EXIT PROC'; KeyWordID: SP_KW_EXIT_PROC),
     (Name: 'EXITPROC'; KeyWordID: SP_KW_EXIT_PROC),
     (Name: 'END CASE'; KeyWordID: SP_KW_END_CASE),
     (Name: 'ENDCASE'; KeyWordID: SP_KW_END_CASE));
Begin

  // Iterate through the line, picking up keywords and translating them into
  // their LONGWORD equivalents. Values should be converted into aFloat form
  // and stored. Strings go in as-is, with a STRINGID byte followed by a LONGWORD
  // Length byte.

  Result := '';
  Idx := 1;

  While Copy(Line, 1, 1) = ' ' Do
    Line := Copy(Line, 2, Length(Line));

  While (Line <> '') And (Line[Length(Line)] in [#0..#32, ':']) Do
    Line := Copy(Line, 1, Length(Line) -1);

  LineLen := Length(Line);

  If Line <> '' Then Begin

    // Pick up a possible line number - beware though, it could be a calculation.
    // Number on its own - delete line number. Number followed by keyword - command.
    // Number followed by function or operator - calculation.

    SP_SkipSpaces(Line, Idx);

    If AddLineNum Then
      If Not IsExpression Then Begin
        StoreVal := Idx;
        If SP_GetNumber(Line, Idx, TempExtend) Then Begin
          If TempExtend = Round(TempExtend) Then Begin
            If Line[Idx] <> '\' Then // A number followed by a backslash is a base, not a line number.
              Result := Result + aChar(SP_LINE_NUM) + LongWordToString(Round(TempExtend))
            Else
              Idx := StoreVal;
          End Else
            Idx := StoreVal;
        End Else
          Idx := StoreVal;
      End;

    While Idx <= LineLen Do Begin

      SP_SkipSpaces(Line, Idx);
      If Idx > LineLen Then Goto Finish;
      SpaceMod := 0;

      // A number at this position? Might need to look ahead for a base operator - "\"

      tIdx := Idx;
      StoreText := '';
      StoreLen := 0;
      FoundBase := False;
      While (tIdx <= LineLen) And (Line[tIdx] in [' ', 'A'..'Z', 'a'..'z', '0'..'9', '\']) Do Begin
        If Line[tIdx] = '\' Then Break;
        StoreText := StoreText + Line[tIdx];
        Inc(tIdx);
      End;

      If (tIdx <= LineLen) and (Line[tIdx] = '\') Then Begin
        Inc(tIdx);
        StoreVal := 0;
        While (tIdx <= LineLen) And (Line[tIdx] in ['0'..'9']) Do Begin
          StoreVal := (Ord(Line[tIdx]) - 48) + (StoreVal * 10);
          Inc(tIdx);
        End;
        If (StoreVal < 37) And (StoreVal > 1) Then Begin
          tStr := '\'+IntToStr(StoreVal);
          StoreLen := Length(StoreText) + Length(tStr);
          FoundBase := DecodeBase(StoreText, StoreVal);
          If FoundBase Then Begin
            Idx := tIdx;
            Result := Result + aChar(SP_VALUE) + aFloatToString(StoreVal) + aChar(SP_TEXT) + LongWordToString(StoreLen) + StoreText + tStr;
          End;
        End;
      End;

      If Not FoundBase Then
        If Line[Idx] in ['.', '0'..'9', '%', '$'] Then Begin
          StoreVal := Idx;
          If SP_GetNumber(Line, Idx, TempExtend) Then Begin
            StoreLen := Idx - StoreVal;
            StoreText := Copy(Line, StoreVal, StoreLen);

            Result := Result + aChar(SP_VALUE) + aFloatToString(TempExtend) + aChar(SP_TEXT) + LongWordToString(StoreLen) + StoreText;

          End Else Begin
            // Not a number, even though we thought it might be - so store is it a structure member?
            If (line[StoreVal] = '.') and (Line[StoreVal +1] in ['A'..'Z', 'a'..'z']) Then Begin
              Idx := StoreVal +1;
              If Idx <= LineLen Then Begin
                If Line[Idx] in ['A'..'Z', 'a'..'z'] Then Begin
                  StoreText := '';
                  While (Idx <= LineLen) And (Line[Idx] in ['0'..'9', 'A'..'Z', 'a'..'z', '_', '$']) Do Begin
                    StoreText := StoreText + Line[Idx];
                    Inc(Idx);
                    If StoreText[Length(StoreText)] = '$' Then Break;
                  End;
                  If Line[Idx -1] = '$' Then
                    Result := Result + aChar(SP_STRUCT_MEMBER_S) + LongWordToString(Length(StoreText)) + StoreText
                  Else
                    Result := Result + aChar(SP_STRUCT_MEMBER_N) + LongWordToString(Length(StoreText)) + StoreText;
                End;
              End;
            End Else
              // Nope, wasn't a structure member, so just leave it here and let the parser deal with it.
              Result := Result + aChar(SP_SYMBOL) + Line[StoreVal];
          End;
        End Else

          // Could it be a String?

          If Line[Idx] in ['"', '#'] Then Begin

            If Line[Idx] = '"' Then Begin

              StoreText := '';

              While True Do Begin

                Inc(Idx);
                If (Idx <= LineLen) Then Begin
                  If Line[Idx] = '"' Then Begin
                    If (Idx < LineLen) And (Line[Idx+1] = '"') Then Begin
                      StoreText := StoreText + '"';
                      Inc(Idx);
                    End Else Begin
                      Inc(Idx);
                      Break;
                    End;
                  End Else
                    StoreText := StoreText + Line[Idx];
                End Else
                  Break;

              End;

              StoreLen := Length(StoreText);
              Result := Result + aChar(SP_STRING) + LongWordToString(StoreLen) + StoreText;

            End Else Begin
              Inc(Idx);
              If (Idx <= LineLen) And (Line[Idx] in ['0'..'9', '$', '%']) Then Begin
                StoreVal := Idx;
                If SP_GetNumber(Line, Idx, TempExtend) Then Begin
                  StoreLen := Idx - StoreVal;
                  StoreText := Copy(Line, StoreVal, StoreLen);
                  Result := Result + aChar(SP_STRINGCHAR) + aFloatToString(TempExtend) + aChar(SP_TEXT) + LongWordToString(StoreLen) + StoreText;
                End Else Begin
                  // Not a number, even though we thought it might be - so store it as a random char.
                  Result := Result + aChar(SP_SYMBOL) + Line[StoreVal];
                  Inc(Idx);
                End;
              End;
            End;

          End Else

            // A numeric variable? Or a String variable? Or a Keyword? Or a Const?

            If Line[Idx] in ['A'..'Z', 'a'..'z'] Then Begin
              StoreText := '';
              TempVal := Idx;
              While (Idx <= LineLen) And (Line[Idx] in ['0'..'9', 'A'..'Z', 'a'..'z', '_', ' ']) Do Begin
                StoreText := StoreText + Line[Idx];
                Inc(Idx);
              End;
              If Idx <= LineLen Then
                If Line[Idx] = '$' Then Begin
                  StoreText := StoreText + '$';
                  Inc(Idx);
                End;
              tIdx := Idx;
              tStore := StoreText;
              If Copy(StoreText, Length(StoreText), 1) = '_' Then
                While (Idx <= LineLen) And (Line[Idx] in ['0'..'9', '_']) Do Begin
                    StoreText := StoreText + Line[Idx];
                    Inc(Idx);
                  End;

              KeyWord := -1;
              Idx := tIdx;
              StoreText := tStore;
              // Test for compound-word commands - DEF PROC/END PROC/GO TO etc. They will have been picked up as
              // potential variable names ("DEF PROC myproc" would be valid for a variable if DEF PROC didn't exist!)
              If Copy(Upper(StoreText), 1, 2) = 'GO' Then Begin
                // "GO" is a special case, but only if it is followed by "TO" or "SUB".
                If (Copy(Upper(StoreText), 3, 2) = 'TO') or (Copy(Upper(StoreText), 3, 3) = 'SUB') Then Begin
                  Dec(Idx, Length(StoreText) -2);
                  StoreText := Copy(StoreText, 1, 2);
                End;
              End;
              TempStr := Upper(StoreText);
              For lIdx := 0 To High(Compounds) Do Begin
                If Copy(TempStr, 1, Length(Compounds[lIdx].Name)) = Compounds[lIdx].Name Then Begin
                  Dec(Idx, Length(StoreText) - Length(Compounds[lIdx].Name));
                  StoreText := Compounds[lIdx].Name;
                  KeyWord := Compounds[lIdx].KeyWordID;
                  SpaceMod := 0;
                  Break;
                End;
              End;

              // A small spelling check - you can be merkin if you want.

              If Upper(StoreText) = 'RANDOMISE' Then StoreText := 'RANDOMIZE';

              // A bit of a hack - if there's a space in this text, then test each word for a match with a reserved word.
              // Cut off at the first reserved word - this allows the CIRCLE command to work with a FILL command.
              If (Pos(' ', string(StoreText)) <> 0) And (KeyWord = -1) Then Begin
                If StoreText[Length(StoreText)] <> ' ' Then
                  StoreText := StoreText + ' ';
                tStore := Upper(StoreText);
                TempStr := '';
                Dec(Idx, Length(StoreText));
                While Pos(' ', string(tStore)) <> 0 Do Begin
                  tStr := Copy(tStore, 1, Pos(' ', string(tStore)) -1);
                  If (SP_IsFunction(tStr) > -1) or (SP_IsKeyWord(tStr) > -1) Then Begin
                    If TempStr = '' Then Begin
                      TempStr := tStr;
                      Inc(Idx, Length(tStr));
                    End;
                    Break;
                  End Else Begin
                    tStore := Copy(tStore, Pos(' ', string(tStore)), Length(tStore));
                    TempStr := TempStr + tStr;
                    While (tStore <> '') And (tStore[1] = ' ') Do Begin
                      TempStr := TempStr + ' ';
                      tStore := Copy(tStore, 2, Length(tStore));
                      Inc(Idx);
                    End;
                    Inc(Idx, Length(tStr));
                  End;
                End;
                StoreText := Copy(StoreText, 1, Length(TempStr));
                While StoreText[Length(StoreText)] = ' ' Do
                  StoreText := Copy(StoreText, 1, Length(StoreText) -1);
              End;

              lIdx := -1;
              If KeyWord = -1 Then lIdx := SP_IsConstant(Upper(StoreText));
              If lIdx > -1 Then Begin
                StoreText := SP_Constants[lIdx].Name;
                TempExtend := SP_Constants[lIdx].Value;
                StoreLen := Length(StoreText);
                Result := Result + aChar(SP_VALUE) + aFloatToString(TempExtend) + aChar(SP_TEXT) + LongWordToString(StoreLen) + StoreText;
              End Else Begin
                If KeyWord = -1 Then
                  KeyWord := SP_IsKeyWord(Upper(StoreText))
                Else
                  Dec(KeyWord , SP_KEYWORD_BASE);
                If KeyWord > -1 Then Begin
                  Result := Result + aChar(SP_KEYWORD) + LongWordToString(KeyWord + SP_KEYWORD_BASE);
                  Idx := TempVal + Length(SP_KEYWORDS[KeyWord]) + SpaceMod;
                  If KeyWord + SP_KEYWORD_BASE = SP_KW_REM Then Begin
                    StoreText := Copy(Line, Idx, Length(Line));
                    Result := Result + aChar(SP_TEXT) + LongWordToString(Length(StoreText)) + StoreText;
                    Goto Finish;
                  End;
                End Else Begin
                  KeyWord := SP_IsFunction(Upper(StoreText));
                  If KeyWord > -1 Then Begin
                    If KeyWord + SP_FUNCTION_BASE = SP_FN_HEX Then Begin
                      Result := Result + aChar(SP_FUNCTION) + LongWordToString(KeyWord + SP_FUNCTION_BASE);
                      Inc(Idx);
                      SP_SkipSpaces(Line, Idx);
                      StoreText := '';
                      StoreLen := 0;
                      While Line[Idx] in ['A'..'F', 'a'..'f', '0'..'9'] Do Begin
                        StoreText := StoreText + Line[Idx];
                        Inc(StoreLen);
                        Inc(Idx);
                      End;
                      If StoreLen <> 0 Then
                        Result := Result + aChar(SP_NUMVAR) + LongWordToString(0) + LongWordToString(StoreLen) + StoreText;
                      TempVal := Idx;
                      Dec(Idx, 3);
                    End Else
                      If KeyWord + SP_FUNCTION_BASE = SP_FN_AND Then
                        Result := Result + aChar(SP_SYMBOL) + SP_CHAR_AND
                      Else
                        If KeyWord + SP_FUNCTION_BASE = SP_FN_OR Then
                          Result := Result + aChar(SP_SYMBOL) + SP_CHAR_OR
                        Else
                          If KeyWord + SP_FUNCTION_BASE = SP_FN_MOD Then
                            Result := Result + aChar(SP_SYMBOL) + SP_CHAR_MOD
                          Else
                            If KeyWord + SP_FUNCTION_BASE = SP_FN_XOR Then
                              Result := Result + aChar(SP_SYMBOL) + SP_CHAR_XOR
                            Else
                              If KeyWord + SP_FUNCTION_BASE = SP_FN_SHL Then
                                Result := Result + aChar(SP_SYMBOL) + SP_CHAR_SHL
                              Else
                                If KeyWord + SP_FUNCTION_BASE = SP_FN_SHR Then
                                  Result := Result + aChar(SP_SYMBOL) + SP_CHAR_SHR
                                Else
                                  If KeyWord + SP_FUNCTION_BASE = SP_FN_MUL Then
                                    Result := Result + aChar(SP_SYMBOL) + SP_CHAR_MUL
                                  Else
                                    If KeyWord + SP_FUNCTION_BASE = SP_FN_DIV Then
                                      Result := Result + aChar(SP_SYMBOL) + SP_CHAR_DIV
                                    Else
                                      If KeyWord + SP_FUNCTION_BASE = SP_FN_ADD Then
                                        Result := Result + aChar(SP_SYMBOL) + SP_CHAR_ADD
                                      Else
                                        If KeyWord + SP_FUNCTION_BASE = SP_FN_SUB Then
                                          Result := Result + aChar(SP_SYMBOL) + SP_CHAR_SUB
                                        Else
                                          Result := Result + aChar(SP_FUNCTION) + LongWordToString(KeyWord + SP_FUNCTION_BASE);
                    Idx := TempVal + Length(SP_FUNCTIONS[KeyWord]);
                  End Else Begin
                    If StoreText[Length(StoreText)] = '$' Then
                      StoreText := Copy(StoreText, 1, Length(StoreText) -1);
                    StoreLen := Length(StoreText);
                    If (Idx <= LineLen +1) And (Line[Idx-1] = '$') Then
                      Result := Result + aChar(SP_STRVAR) + LongWordToString(0) + LongWordToString(StoreLen) + StoreText
                    Else
                      Result := Result + aChar(SP_NUMVAR) + LongWordToString(0) + LongWordToString(StoreLen) + StoreText;
                  End;
                End;
              End;
            End Else Begin
              // must be a random character - a comma or period, perhaps.
              If Result <> '' Then
                If Line[Idx] = ':' Then Begin
                  Result := Result + aChar(SP_SYMBOL) + ':';
                End Else
                  If Line[Idx] = '@' Then Begin
                    Inc(Idx);
                    StoreText := '@';
                    While (Idx <= LineLen) And (Line[Idx] in ['0'..'9', 'A'..'Z', 'a'..'z', '_']) Do Begin
                      StoreText := StoreText + Line[Idx];
                      Inc(Idx);
                    End;
                    Result := Result + aChar(SP_LABEL) + LongWordToString(Length(StoreText)) + StoreText;
                    Dec(Idx);
                  End Else
                    If (Result[Length(Result) -1] = aChar(SP_SYMBOL)) And (Line[Idx] = '=') Then Begin
                      Case Result[Length(Result)] of
                        '>': Result[Length(Result)] := SP_CHAR_GTE;
                        '<': Result[Length(Result)] := SP_CHAR_LTE;
                      Else
                        Result := Result + aChar(SP_SYMBOL) + Line[Idx];
                      End;
                    End Else Begin
                      If (Result[Length(Result) -1] = aChar(SP_SYMBOL)) Then Begin
                        If (Line[Idx] = '>') And (Result[Length(Result)] = '<') Then
                          Result[Length(Result)] := SP_CHAR_DNE
                        Else
                          Result := Result + aChar(SP_SYMBOL) + Line[Idx];
                      End Else
                        Result := Result + aChar(SP_SYMBOL) + Line[Idx];
                    End Else
                      Result := Result + aChar(SP_SYMBOL) + Line[Idx];
                  Inc(Idx);
              End;

    End;

Finish:

  End;

End;

Function SP_IsConstant(Line: aString): Integer;
Var
  Idx: Integer;
Begin

  Result := -1;

  For Idx := 0 To High(SP_CONSTANTS) Do
    If Upper(SP_CONSTANTS[Idx].Name) = Line Then Begin
      Result := Idx;
      Exit;
    End;

End;

Function SP_IsKeyWord(Line: aString): Integer;
Var
  Idx: Integer;
Begin

  Result := -1;

  For Idx := 0 To High(SP_KEYWORDS) Do
    If Upper(SP_KEYWORDS[Idx]) = Line Then Begin
      Result := Idx;
      Exit;
    End;

End;

Function SP_IsFunction(Line: aString): Integer;
Var
  Idx: Integer;
Begin

  Result := -1;

  For Idx := 0 To High(SP_FUNCTIONS) Do
    If Upper(SP_FUNCTIONS[Idx]) = Line Then Begin
      Result := Idx;
      Exit;
    End;

End;

Function SP_IsFunctionEx(Line: aString): Integer;
Var
  Idx: Integer;
Begin

  Result := -1;

  For Idx := 0 To High(SP_FUNCTIONS_EX) Do
    If Upper(SP_FUNCTIONS_EX[Idx]) = Line Then Begin
      Result := Idx;
      Exit;
    End;

End;

Function SP_GetNumber(Line: aString; Var Idx: Integer; Var Number: aFloat): Boolean;
Var
  NegExp: Boolean;
  OldIdx, NewIdx, NumDigits: Integer;
  Dec_Value: aFloat; Dec_Count, Exponent: Integer;
Begin

  // Returns a number from a String, at the supplied position.
  // Binary is prefixed by '%', Hex by '$' or '0x' and decimal has
  // no prefix.

  Result := False;
  Number := 0;

  If Line[Idx] = '%' Then Begin

    // Binary.

    Inc(Idx);
    OldIdx := Idx;
    NumDigits := 1;
    While (Idx <= Length(Line)) And (Line[Idx] in [' ', '1', '0']) Do Inc(Idx);
    If Idx > OldIdx Then Begin
      NewIdx := Idx;
      Dec(Idx);
      While Idx >= OldIdx Do Begin
        If Line[Idx] in ['0', '1'] Then Begin
          Result := True;
          If Line[Idx] = '1' Then
            Number := Number + NumDigits;
          Inc(NumDigits, NumDigits);
        End;
        Dec(Idx);
      End;
      Idx := NewIdx;
    End;

  End Else

    If (Line[Idx] = '$') or (Lower(Copy(Line, Idx, 2)) = '0x') Then Begin

      // Hex.

      Inc(Idx);
      If Idx <= Length(Line) Then
        If Line[Idx] in ['x', 'X'] Then
          Inc(Idx);

      While (Idx <= Length(Line)) and (Line[Idx] in ['0'..'9', 'A'..'F', 'a'..'f']) Do Begin
        If Line[Idx] <> ' ' Then
          Result := True;
        If Line[Idx] in ['A'..'F'] Then
          Number := (Number * 16) + Ord(Line[Idx]) - 55
        Else
          If Line[Idx] in ['a'..'f'] Then
            Number := (Number * 16) + Ord(Line[Idx]) - 87
          Else
            If Line[Idx] <> ' ' Then
              Number := (Number * 16) + Ord(Line[Idx]) - 48;
        Inc(Idx);
      End;

    End Else Begin

      // Decimal.

      While (Idx <= Length(Line)) And (Line[Idx] in ['0'..'9']) Do Begin
        Result := True;
        Number := (Number * 10) + Ord(Line[Idx]) - 48;
        Inc(Idx);
      End;

      If Idx <= Length(Line) Then Begin
        If Line[Idx] = '.' Then Begin
          Inc(Idx);
          Dec_Value := 0;
          Dec_Count := 10;
          While (Idx <= Length(Line)) And (Line[Idx] in ['0'..'9']) Do Begin
            Result := True;
            Dec_Value := ((Ord(Line[Idx]) - 48) / Dec_Count) + Dec_Value;
            Dec_Count := Dec_Count * 10;
            Inc(Idx);
          End;
          If Dec_Value <> 0 Then
            Number := Number + Dec_Value;
        End;

        If Result And (Idx <= Length(Line)) Then Begin
          If Line[Idx] in ['E','e'] Then Begin
            Inc(Idx);
            Exponent := 0;
            NegExp := False;
            If Idx <= Length(Line) Then
              If Line[Idx] in ['+', '-'] Then Begin
                If Line[Idx] = '-' Then NegExp := True;
                Inc(Idx);
              End;
            While (Idx <= Length(Line)) And (Line[Idx] in ['0'..'9']) Do Begin
              Exponent := (Exponent * 10) + Ord(Line[Idx]) - 48;
              Inc(Idx);
            End;
            If NegExp Then
              Number := Number / Power(10, Exponent)
            Else
              Number := Number * Power(10, Exponent);
          End;
        End;

      End;

    End;

End;

Function SP_Highlight(hlClr, NewWord, nClr: aString): aString;
Var
  Idx, wLen, LastSpace: Integer;
Begin

  If NewWord <> '' Then Begin
    Idx := 1;
    Result := '';
    wLen := Length(NewWord);
    While Idx <= wLen Do Begin
      If NewWord[Idx] <> ' ' Then Begin
        Result := Result + hlClr;
        Break;
      End;
      Result := Result + ' ';
      Inc(Idx);
    End;
    LastSpace := wLen;
    While NewWord[LastSpace] = ' ' Do Dec(LastSpace);
    While Idx <= LastSpace Do Begin
      Result := Result + NewWord[Idx];
      Inc(Idx);
    End;
    Result := Result + nClr;
    While Idx <= wLen Do Begin
      Result := Result + ' ';
      Inc(Idx);
    End;
  End;

End;

Function SP_Detokenise(Tokens: aString; Var cPos: Integer; Highlight, UseDoubles: Boolean): aString;
Var
  Idx, Idx2, cIdx, Token, LastToken, TokenStart, LabelLen, lC: Integer;
  NewWord, Str, strClr, remClr, kwClr, nClr, numClr, vClr, ConstClr, Clr, BreakCharPre, BreakCharPost, Txt: aString;
  KeyWordID, LongWordPtr: pLongWord;
  LastChar: aChar;
  LastKeyWord: LongWord;
  FoundPos, FirstChar: Boolean;
Begin

  If Highlight Then Begin
    nClr := #16#17#0#0#0;
    kwClr := #16#0#0#0#0;
    numClr := #16#1#0#0#0;
    strClr := #16#1#0#0#0;
    vClr := #16#2#0#0#0;
    remClr := #16#20#0#0#0;
    constClr := #16#3#0#0#0;
  End Else Begin
    nClr := '';
    kwClr := '';
    numClr := '';
    vCLr := '';
    remClr := '';
    strClr := '';
    constClr := '';
  End;

  Idx := 1;
  TokenStart := 1;
  LastToken := -1;
  Result := '';
  BreakCharPre := '';
  BreakCharPost := '';
  FirstChar := False;

  If Tokens[Idx] = aChar(SP_LINE_NUM) Then Begin
    Inc(Idx);
    Result := IntToString(pLongWord(@Tokens[Idx])^) + ' ';
    Inc(Idx, SizeOf(LongWord));
    FirstChar := True;
  End;

  If Tokens[Idx] = aChar(SP_LINE_LEN) Then
    Inc(Idx, SizeOf(LongWord) +1);

  If Tokens[Idx] = aChar(SP_STATEMENTS) Then
    Inc(Idx, 1 +((1 + pLongWord(@Tokens[Idx +1])^) * SizeOf(LongWord)));

  LastKeyWord := 0;
  FoundPos := False;

  While Idx <= Length(Tokens) Do Begin

    LastChar := #0;
    If Length(Result) > 0 Then Begin
      lC := Length(Result);
      While lC > 0 Do Begin
        If Result[lC] >= ' ' Then Begin
          LastChar := Result[lC];
          Break;
        End;
        Dec(lC);
      End;
    End;

    If (Not FoundPos) And (Idx >= cPos) Then Begin
      cPos := Length(Result) +1;
      FoundPos := True;
    End;

    TokenStart := Length(Result) +1;
    Token := Ord(Tokens[Idx]);

    Case Token Of

      SP_STATEMENTS:
        Break;

      SP_LINE_LEN:
        Break;

      SP_LABEL:
        Begin
          Inc(Idx);
          LabelLen := pLongWord(@Tokens[Idx])^;
          Inc(Idx, SizeOf(LongWord));
          Result := Result + SP_Highlight(numClr, Copy(Tokens, Idx, LabelLen), nClr);
          Inc(Idx, LabelLen);
          LastKeyWord := 0;
        End;

      SP_KEYWORD, SP_FUNCTION, SP_FUNCTION_EX:
        Begin
          Inc(Idx);
          KeyWordID := @Tokens[Idx];
          LastKeyWord := KeyWordID^;
          { fix for legacy programs that use FN and POINT }
            If Ord(Tokens[Idx -1]) = SP_FUNCTION_EX Then
              Tokens[Idx -1] := aChar(SP_FUNCTION);
            If KeyWordID^ = 3000 Then KeyWordID^ := SP_FN_POINT;
            If KeyWordID^ = 3001 Then KeyWordID^ := SP_FN_FN;
          { End fix}
          If Highlight Then Begin
            BreakCharPre := '';
            BreakCharPost := '';
            If KeyWordID^ = SP_KW_THEN Then Begin
              BreakCharPre := '';
              BreakCharPost := #13;
            End;
            If KeyWordID^ = SP_KW_ELSE Then Begin
              BreakCharPre := #13;
              BreakCharPost := #13;
            End;
          End;
          If Ord(Tokens[Idx -1]) = SP_KEYWORD Then Begin
            If KeyWordID^ - SP_KEYWORD_BASE <= Length(SP_KEYWORDS_EXTRA) Then
              NewWord := SP_KEYWORDS_EXTRA[KeyWordID^ - SP_KEYWORD_BASE]
            Else
              NewWord := ' ';
            If NewWord = ' s** ' Then
              NewWord := ' SUB ';
          End Else
            If Ord(Tokens[Idx -1]) = SP_FUNCTION Then
              NewWord := Copy(SP_FUNCTIONS_EXTRA[KeyWordID^ - SP_FUNCTION_BASE], 2, 999);
          If NewWord <> '' Then
            If NewWord[1] = ' ' Then
              If (Result <> '') And (Result[Length(Result)] = ' ') Then
                NewWord := Copy(NewWord, 2, Length(NewWord));
          If (Length(Result) > 0) And (LastChar in ['0'..'9', 'A'..'Z', 'a'..'z', ':', ')']) And
             (LastChar <> ' ') And ((NewWord <> '') And (NewWord[1] <> ' ')) Then
            Result := Result + ' ' + BreakCharPre + SP_Highlight(kwClr, NewWord, nClr) + BreakCharPost
          Else
            Result := Result + BreakCharPre + SP_Highlight(KwClr, NewWord, nClr) + BreakCharPost;
          Inc(Idx, SizeOf(LongWord));
          LastChar := NewWord[Length(NewWord)];
          If KeyWordID^ = SP_KW_REM Then Begin
            If  LastChar = ' ' Then
              Result := Copy(Result, 1, Length(Result) -1);
            If Byte(Tokens[Idx]) = SP_TEXT Then Inc(Idx, 1 + SizeOf(LongWord));
            Result := Result + SP_Highlight(remClr, Copy(Tokens, Idx, Length(Tokens)), nClr);
            If Pos(#255, Result) > 0 Then
              Result := Copy(Result, 1, Pos(#255, Result) -1);
            Break;
          End;
        End;

      SP_VALUE:
        Begin
          If UseDoubles Then
            Inc(Idx, SizeOf(Double) +1)
          Else
            Inc(Idx, SizeOf(aFloat)+1);
          If LastToken in [SP_KEYWORD, SP_FUNCTION] Then
            If LastChar <> ' ' Then
              Result := Result + ' ';
          LastKeyWord := 0;
        End;

      SP_VALUE10:
        Begin
          Inc(Idx, 11);
          LastKeyWord := 0;
        End;

      SP_STRING:
        Begin
          Inc(Idx);
          LongWordPtr := @Tokens[Idx];
          Inc(Idx, SizeOf(LongWord));
          Str := '';
          Idx2 := Idx;
          While LongWord(Idx2) < LongWord(Idx) + LongWordPtr^ Do Begin
            If Tokens[Idx2] = '"' Then
              Str := Str + '""'
            Else
              Str := Str + Tokens[Idx2];
            Inc(Idx2);
          End;
          Result := Result + SP_Highlight(strClr, '"'+Str+'"', nClr);
          Inc(Idx, LongWordPtr^);
          LastKeyWord := 0;
        End;

      SP_STRINGCHAR:
        Begin
          Result := Result + '#';
          If UseDoubles Then
            Inc(Idx, SizeOf(Double)+1)
          Else
            Inc(Idx, SizeOf(aFloat)+1);
          LastKeyWord := 0;
        End;

      SP_NUMVAR:
        Begin
          pLongWord(@Tokens[Idx +1])^ := 0;
          Inc(Idx, 1 + SizeOf(LongWord));
          LongWordPtr := @Tokens[Idx];
          Inc(Idx, SizeOf(LongWord));
          NewWord := Copy(Tokens, Idx, LongWordPtr^);
          If NewWord[1] = '!' Then
            NewWord := Copy(NewWord, 2, Length(NewWord));
          If LastChar in ['a'..'z', 'A'..'Z', '0'..'9'] Then
            NewWord := ' ' + NewWord;
          Result := Result + SP_Highlight(vClr, NewWord, nClr);
          Inc(Idx, LongWordPtr^);
          LastKeyWord := 0;
        End;

      SP_STRVAR:
        Begin
          pLongWord(@Tokens[Idx +1])^ := 0;
          Inc(Idx, 1 + SizeOf(LongWord));
          LongWordPtr := @Tokens[Idx];
          Inc(Idx, SizeOf(LongWord));
          NewWord := Copy(Tokens, Idx, LongWordPtr^);
          If NewWord[1] = '!' Then
            NewWord := Copy(NewWord, 2, Length(NewWord));
          If LastChar in ['a'..'z', 'A'..'Z', '0'..'9'] Then
            NewWord := ' ' + NewWord;
          Result := Result + SP_Highlight(vClr, NewWord + '$', nClr);
          Inc(Idx, LongWordPtr^);
          LastKeyWord := 0;
        End;

      SP_TEXT, SP_STRUCT_MEMBER_N, SP_STRUCT_MEMBER_S:
        Begin
          Inc(Idx);
          LongWordPtr := @Tokens[Idx];
          Inc(Idx, SizeOf(LongWord));
          Txt := Copy(Tokens, Idx, LongWordPtr^);
          Clr := nClr;
          If Token in [SP_STRUCT_MEMBER_N, SP_STRUCT_MEMBER_S] Then
            Result := Result + '.'
          Else Begin
            cIdx := 0;
            While cIdx < Length(SP_Constants) -1 Do Begin
              If SP_Constants[cIdx].Name = Txt Then Begin
                Clr := ConstClr;
                Break;
              End;
              Inc(cIdx);
            End;
          End;
          Result := Result + SP_Highlight(Clr, Txt, nClr);
          Inc(Idx, LongWordPtr^);
          LastKeyWord := 0;
        End;

      SP_SYMBOL:
        Begin
          Inc(Idx);
          Case Tokens[Idx] of
            SP_CHAR_GTE: NewWord := '>=';
            SP_CHAR_LTE: NewWord := '<=';
            SP_CHAR_DNE: NewWord := '<>';
            SP_CHAR_AND: NewWord := ' AND ';
            SP_CHAR_OR:  NewWord := ' OR ';
            SP_CHAR_MOD: NewWord := ' MOD ';
            SP_CHAR_XOR: NewWord := ' XOR ';
            SP_CHAR_SHL: NewWord := ' SHL ';
            SP_CHAR_SHR: NewWord := ' SHR ';
            SP_CHAR_ADD: NewWord := ' ADD ';
            SP_CHAR_SUB: NewWord := ' SUB ';
            SP_CHAR_MUL: NewWord := ' MUL ';
            SP_CHAR_DIV: NewWord := ' DIV ';
            '&': NewWord := ' & ';
            '|': NewWord := ' | ';
            '(':
              Begin
                If Length(Result) > 0 Then Begin
                  If (LastChar = ' ') And (Not (LastKeyWord < SP_FUNCTION_BASE) And (Not (Result[Length(Result)-1] in ['A'..'Z'])) or (LastKeyWord >= SP_FUNCTION_BASE)) Then
                    Result := Copy(Result, 1, Length(Result) -1) + '('
                  Else
                    Result := Result + '(';
                End Else
                    Result := Result + '(';
                NewWord := '';
              End;
            ':':
              Begin
                If Not FirstChar Then
                  While (Copy(Result, Length(Result), 1) <= ' ') And (Result <> '') Do
                    Result := Copy(Result, 1, Length(Result) -1);
                FirstChar := False;
                Result := Result + ':';
                NewWord := '';
              End;
            ')':
              Begin
                If Length(Result) > 0 Then Begin
                  If (LastChar = ' ') And (NewWord[Length(NewWord)-1] in ['A'..'Z']) Then
                    Result := Copy(Result, 1, Length(Result) -1) + ')'
                  Else
                    Result := Result + ')';
                End Else
                    Result := Result + ')';
                NewWord := '';
              End;
          Else
            Result := Result + Tokens[Idx];
            NewWord := '';
          End;
          Inc(Idx);
          If NewWord <> '' Then Begin
            If NewWord[1] = ' ' Then
              If (Result = '') or (Result[Length(Result)] = ' ') Then
                NewWord := Copy(NewWord, 2, Length(NewWord));
            Result := Result + NewWord;
          End;
          LastKeyWord := 0;
        End;

      SP_TERMINAL:
        Begin
          If Not FoundPos Then
            If cPos < Length(Tokens) Then
              cPos := Length(Result)
            Else
              cPos := cPos - Length(Tokens);
          Exit;
        End;

    End;

    LastToken := Token;

  End;

  If Not FoundPos Then
    If cPos < Length(Tokens) Then
      cPos := Length(Result)
    Else
      cPos := cPos - Length(Tokens);

End;

Function SP_Store_Line(Tokens: aString): Integer;
Var
  LineNum: LongWord;
  Idx, LineIdx, Position: Integer;
Begin

  // Stores a line in the SP_Program list. First, get the line number:

  LineNum := pLongWord(@Tokens[2])^;
  Position := 3 + SizeOf(LongWord);

  // Now skip the statements list.

  Inc(Position, (1 + pLongWord(@Tokens[Position])^) * SizeOf(LongWord));

  If Tokens[Position] = SP_TERMINAL_CHAR Then Begin

    // Delete operation

    Idx := SP_FindLine(LineNum, True);
    If Idx <> -1 Then Begin
      SP_Program_Delete(Idx);
      SP_MakeEditorList;
      LineIdx := SP_FindLine(PROGLINE, True);
      If LineIdx > -1 Then Begin
        If LineIdx < SHOWLINE Then SHOWLINE := pLongWord(@SP_Program[LineIdx][2])^;
        PROGLINE := pLongWord(@SP_Program[LineIdx][2])^;
      End;
    End;

  End Else Begin

    Idx := 0;
    While (Idx < SP_Program_Count) And (pLongWord(@SP_Program[Idx][2])^ < LineNum) Do
      Inc(Idx);

    If Idx = SP_Program_Count Then
      SP_Program_Add(Tokens)
    Else
      If LineNum = pLongWord(@SP_Program[Idx][2])^ Then
        SP_Program[Idx] := Tokens
      Else
        SP_Program_Insert(Idx, Tokens);

    SP_MakeEditorList;

  End;

  Result := LINENUM;

End;

// String array handlers

Procedure SP_Program_Add(Str: aString);
Begin
  Inc(SP_Program_Count);
  SetLength(SP_Program, SP_Program_Count);
  SP_Program[SP_Program_Count -1] := Str;
  PROGCHANGED := True;
End;

Procedure SP_Program_AddStrings(List: TStringlist);
Var
  Idx: Integer;
Begin
  For Idx := 0 To List.Count -1 Do
    SP_Program_Add(aString(List[Idx]));
  PROGCHANGED := True;
End;

Procedure SP_Program_AddStringsArray(Var List: Array of aString);
Var
  Idx: Integer;
Begin
  For Idx := 0 To Length(List) -1 Do
    SP_Program_Add(List[Idx]);
  PROGCHANGED := True;
End;

Procedure SP_Program_Insert(Index: Integer; Str: aString);
Var
  Idx: Integer;
Begin
  Inc(SP_Program_Count);
  SetLength(SP_Program, SP_Program_Count);
  For Idx := SP_Program_Count -1 DownTo Index +1 Do
    SP_Program[Idx] := SP_Program[Idx -1];
  SP_Program[Index] := Str;
  PROGCHANGED := True;
End;

Procedure SP_Program_Delete(Index: Integer);
Var
  Idx: Integer;
Begin
  For Idx := Index To SP_Program_Count -2 Do
    SP_Program[Idx] := SP_Program[Idx +1];
  Dec(SP_Program_Count);
  SetLength(SP_Program, SP_Program_Count);
  PROGCHANGED := True;
End;

Procedure SP_Program_Delete_Line(Line: Integer);
Begin

  Line := SP_FindLine(Line, True);
  If Line >= 0 Then SP_Program_Delete(Line);

End;

Procedure SP_Program_Clear;
Begin
  SetLength(SP_Program, 0);
  SP_Program_Count := 0;
  SP_DeleteIncludes;
  PROGCHANGED := True;
End;

Initialization

  SetLength(SP_KEYWORDS, High(SP_KEYWORDS_EXTRA) +1);
  For Idx_t := 0 To High(SP_KEYWORDS_EXTRA) Do Begin

    TempStr := SP_KEYWORDS_EXTRA[Idx_t];
    If TempStr[1] = ' ' Then TempStr := Copy(TempStr, 2, Length(TempStr));
    If TempStr[Length(TempStr)] = ' ' Then TempStr := Copy(TempStr, 1, Length(TempStr) -1);

    SP_KEYWORDS[Idx_t] := TempStr;

  End;

  SetLength(SP_FUNCTIONS, High(SP_FUNCTIONS_EXTRA) +1);
  SetLength(SP_OPTIMISE_FLAGS, Length(SP_FUNCTIONS));

  For Idx_t := 0 To High(SP_FUNCTIONS_EXTRA) Do Begin

    TempStr := SP_FUNCTIONS_EXTRA[Idx_t];
    If TempStr[1] = 'o' Then SP_OPTIMISE_FLAGS[Idx_t] := True Else SP_OPTIMISE_FLAGS[Idx_t] := False;
    TempStr := Copy(TempStr, 2, Length(TempStr));
    If TempStr[1] = ' ' Then TempStr := Copy(TempStr, 2, Length(TempStr));
    If TempStr[Length(TempStr)] = ' ' Then TempStr := Copy(TempStr, 1, Length(TempStr) -1);

    SP_FUNCTIONS[Idx_t] := TempStr;

  End;

end.
