// Copyright (C) 2010 By Paul Dunn
//
// This file is part of the SpecBAS BASIC Interpreter, which is in turn
// part of the SpecOS project.
//
// SpecBAS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// SpecBAS is distributed in the hope that it will be entertaining,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SpecBAS.  If not, see <http://www.gnu.org/licenses/>.

{$INCLUDE SpecBAS.inc}

{$IFDEF PANDORA}
program pSpecBAS;
{$ELSE}
program SDLBas;
{$ENDIF}
{$IFDEF RASPICONSOLE}
{$DEFINE RASPI}
{$ENDIF}

{$IFDEF FPC}
{$MODE Delphi}{$H+}
{$ENDIF}

uses
  {$IFDEF FPC}
  cThreads,
  {$IFDEF RASPICONSOLE}
  NoGUIInt,
  {$ELSE}
  Interfaces,
  {$ENDIF}
  LCLIntf, Classes, SysUtils, crt,
  {$ELSE}
  Windows, Classes, SysUtils, SHFolder, ShellAPI,
  {$ENDIF}
  { you can add units after this }
  SDL, SDL_Image, SP_Graphics32,
  Math, SP_Main, SP_SysVars, SP_Graphics, SP_Input, SP_BankManager, SP_Errors,
  SP_Sound, BASS, SP_Util, SP_Tokenise, SP_FileIO, SP_Editor;

Type

  TSpecBAS_Thread = Class(TThread)
    Procedure Execute; Override;
  end;

Var

  ScrWidth, ScrHeight: Integer;
  BASThread: TSpecBAS_Thread;
  Quitting: Boolean = False;
  Screen: pSDL_Surface;
  Event: TSDL_Event;
  ConvertKeys: Array[0..65535] of Byte;
  Key, Button: Integer;
  TickCount, LastTickCount, InitTime: Integer;
  ScrResChange, ScrNewFull, DoPalette: Boolean;
  ScrNewWidth, ScrNewHeight: Integer;
  ImgResource: pSDL_Surface;
  HiddenCursor: pSDL_Cursor;
  K_Flag: Boolean;
  {$IFDEF PANDORA}
  Nub1: pSDL_Joystick;
  Nub2: pSDL_Joystick;
  {$ENDIF}

Procedure TSpecBAS_Thread.Execute;
Begin

  Priority := tpHigher;
  SP_MainLoop;
  Terminate;

End;

Procedure LoadImage(Filename: aString; Var Error: TSP_ErrorCode);
Var
  Idx: Integer;
  r, g, b: Byte;
Begin

  // Loads (using arch-specific routines - GraphicEx under Win32, SDL_Image under linux) an image,
  // and fills the supplied pointer and w/h/s vars with info about the image. Use the "FreeImageResource"
  // proc when we're done with the image.

  If FileExists(Filename) Then Begin
    Try
      ImgResource := IMG_Load(pChar(Filename));
    Except
      On E: Exception Do Begin
        If Assigned(ImgResource) Then SDL_FreeSurface(ImgResource);
        Error.Code := SP_ERR_UNSUPPORTED_IMAGE_FORMAT;
        Exit;
      End;
    End;
    If ImgResource^.format.BitsPerPixel <> 8 Then
      Error.Code := SP_ERR_INVALID_IMAGE_FORMAT
    Else Begin
      For Idx := 0 To 255 Do Begin
        r := ImgResource^.Format.palette.colors[Idx].r;
        g := ImgResource^.Format.palette.colors[Idx].g;
        b := ImgResource^.Format.palette.colors[Idx].b;
        ImgPalette[Idx].R := r;
        ImgPalette[Idx].G := g;
        ImgPalette[Idx].B := b;
      End;
      ImgPtr := pByte(ImgResource^.pixels);
      ImgStride := ImgResource^.pitch;
      ImgHeight := ImgResource^.h;
      ImgWidth := ImgResource^.w;
    End;
  End Else Begin
    Error.Code := SP_ERR_FILE_MISSING;
  End;

End;

Procedure FreeImageResource;
Begin

  // Removes an image from memory after loading.

  SDL_FreeSurface(ImgResource);

End;

Procedure SetPalette(Index: Integer; R, G, B: Byte);
Var
  BankScreen: pSP_Bank_Screen;
Begin

  While DRAWING Do Sleep(1);
  While Length(SP_BankList) = 0 Do Sleep(1);
  BankScreen := @SP_BankList[0]^.Info[0];
  BankScreen^.Palette[Index].r := R;
  BankScreen^.Palette[Index].g := G;
  BankScreen^.Palette[Index].b := B;

end;

Function  GetPalette(Index: Integer): LongWord;
Var
  BankScreen: pSP_Bank_Screen;
  Clr: TP_Colour;
Begin

  While Length(SP_BankList) = 0 Do Sleep(1);
  BankScreen := @SP_BankList[0]^.Info[0];
  Clr := BankScreen^.Palette[Index];
  Result := (Clr.r Shl 16) + (Clr.g Shl 8) + Clr.b;

end;

Procedure FinalisePalette;
Begin

  DoPalette := True;

end;

Procedure SDL_FinalisePalette;
Var
  BankScreen: pSP_Bank_Screen;
  Clr: Array[0..255] of TSDL_Color;
  Idx: Integer;
Begin

  While DRAWING Do Sleep(1);
  While Length(SP_BankList) = 0 Do Sleep(1);

  BankScreen := @SP_BankList[0]^.Info[0];
  For Idx := 0 To 255 Do Begin
    Clr[Idx].r := BankScreen^.Palette[Idx].r;
    Clr[Idx].g := BankScreen^.Palette[Idx].g;
    Clr[Idx].b := BankScreen^.Palette[Idx].b;
  end;
  SDL_SetColors(screen, @Clr, 0, 256);
  DoPalette := False;

end;

Function SetScreen(Width, Height: Integer; FullScreen: Boolean): Integer;
Begin

  DISPLAYPOINTER := nil;
  DPtrBackup := nil;

  ScrResChange := True;
  ScrNewWidth := Width;
  ScrNewHeight := Height;
  ScrNewFull := FullScreen;

end;

Function SetScreenResolution(Width, Height: Integer; FullScreen: Boolean): Boolean;
Begin

  // Pandora SpecBAS never runs in a window

  {$IFDEF FPC}
  {$IFDEF PANDORA}
  FullScreen := True;
  {$ENDIF}

  {$IFDEF RASPICONSOLE}
  FullScreen := True;
  {$ELSE}
  SDL_putenv('SDL_VIDEO_CENTERED=center');
  {$IFDEF RASPI}
  SDL_WM_SetCaption('SpecBAS Pi build', nil);
  {$ELSE}
  SDL_WM_SetCaption('SpecBAS Linux build', nil);
  {$ENDIF}
  {$ENDIF}
  {$ELSE}
  SDL_WM_SetCaption('SpecBAS Windows Build', nil);
  {$ENDIF}

  SDL_FreeSurface(Screen);

  If FullScreen Then Begin

    Screen := SDL_SetVideoMode(Width, Height, 32, SDL_FULLSCREEN or SDL_HWSURFACE);
    SPFULLSCREEN := True;
    Result := Screen <> nil;

  end Else Begin

    Screen := SDL_SetVideoMode(Width, Height, 32, SDL_HWSURFACE);
    SPFULLSCREEN := False;
    Result := Screen <> nil;

  end;

  FinalisePalette;

  DISPLAYSTRIDE := Screen^.pitch;
  DISPLAYPOINTER := Screen^.pixels;
  DPtrBackup := DISPLAYPOINTER;

end;

Function  TestScreenResolution(Width, Height: Integer; FullScreen: Boolean): Boolean;
Var
  CurVidInfo: PSDL_VideoInfo;
  ModeList : PPSDL_Rect;
Begin
  {$IFDEF RASPICONSOLE}
  CurVidInfo := SDL_GetVideoInfo;
  ModeList := SDL_ListModes(CurVidInfo^.vfmt, SDL_OPENGL or SDL_HWSURFACE);
  Result := False;
  While Assigned (ModeList^) do
	Begin
		{ Don't allow modes wider than 1024 pixels else the Pi can't keep up }
		If (Width = ModeList^.w) and (Height = ModeList^.h) and (ModeList^.w < 1024) Then Begin
			 Result := True;
		End;
	Inc(ModeList^);
	End;
  {$ELSE}
  Result := True;
  {$ENDIF}
end;

Procedure YieldProc;
Begin

  SDL_Delay(1);

end;

Procedure GetKeyState;
Var
  Keys: Integer;
Begin
  Keys := SDL_GetModState;
  CAPSLOCK := Integer((Keys And $2000) > 0);
  NUMLOCK := Integer((Keys And $1000) > 0);
end;

Procedure Refresh_Display;
Var
  X1, Y1, X2, Y2: Integer;
Begin

  DRAWING := True;
  If Not (Quitting or SCREENCHANGE or SetDR) Then Begin

    If MOUSEVISIBLE Then
      SP_RestoreMouseRegion;
    If DISPLAYPOINTER <> nil Then Begin
      If NUMSPRITES > 0 Then Begin
        SCMINX := 0; SCMINY := 0;
        SCMAXX := DISPLAYWIDTH; SCMAXY := DISPLAYHEIGHT;
      End;
      If SCMAXX >= SCMINX Then Begin
        X1 := SCMINX; Y1 := SCMINY; X2 := SCMAXX +1; Y2 := SCMAXY +1;
        SCMAXX := 0; SCMAXY := 0; SCMINX := DISPLAYWIDTH; SCMINY := DISPLAYHEIGHT;
        SetDR := False;
        SP_Composite32(DISPLAYPOINTER, X1, Y1, X2, Y2);
        With DispRects[GfxUpdRect] Do Begin
          x := Max(0, X1);
          y := Max(0, Y1);
          w := Min(DISPLAYWIDTH, Min(DISPLAYWIDTH -X1, (X2 - X1) +1));
          h := Min(DISPLAYHEIGHT, Min(DISPLAYHEIGHT -Y1, (Y2 - Y1) +1));
          Inc(GfxUpdRect);
        End;
      End;
      If MOUSEVISIBLE Then SP_DrawMouseImage;

      If GfxUpdRect > 0 Then Begin
        SDL_UpdateRects(Screen, GfxUpdRect, @DispRects[0]);
        GfxUpdRect := 0;
      End;

    End;
  End;
  DRAWING := False;

end;

Function  GetTicks: LongWord;
Begin

  Result := GetTickCount;

end;

Procedure Quit;
Begin

  Quitting := True;

end;

Procedure ReleaseSticks;
Begin

  {$IFDEF PANDORA}
  If Nub1 <> Nil Then SDL_JoystickClose(Nub1);
  If Nub2 <> Nil Then SDL_JoystickClose(Nub2);
  Nub1 := Nil;
  Nub2 := Nil;
  SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
  {$ENDIF}

End;

Procedure InitSticks;
Var
  Idx: Integer;
Begin

  {$IFDEF PANDORA}
  SDL_InitSubSystem(SDL_INIT_JOYSTICK);
  For Idx := 0 To SDL_NumJoySticks -1 Do Begin
    If SDL_JoystickName(Idx) = 'nub0' Then Begin
      Nub1 := SDL_JoystickOpen(Idx);
    End Else
      If SDL_JoyStickName(Idx) = 'nub1' Then Begin
        Nub2 := SDL_JoyStickOpen(Idx);
      End;
  End;
  SDL_JoystickUpdate;
  {$ENDIF}

end;

Function GetAxis(NubID, Axis: Integer): Integer;
Begin

  Result := 0;

  {$IFDEF PANDORA}

    If NubID = 1 Then Begin
      If Nub1 <> Nil Then
         Result := SDL_JoystickGetAxis(Nub1, Axis)
      Else
         Result := 0;
    End Else
      If Nub2 <> Nil Then
        Result := SDL_JoystickGetAxis(Nub2, Axis)
      Else
        Result := 0;

  {$ENDIF}

End;

Function GetBatteryStatus: Integer;
Begin

  {$IFDEF PANDORA}
  BATTLEVEL := StrToInt(ReadLinuxFile('/sys/class/power_supply/bq27500-0/capacity'));
  {$ELSE}
  BATTLEVEL := 100;
  {$ENDIF}

End;

Procedure Init;
Var
  Error: TSP_ErrorCode;
  HiddenCursorData: Byte;
  CurVidInfo: PSDL_VideoInfo;
  ModeList : PPSDL_Rect;
  Path: Array [0..MAX_PATH] of Char;
Begin

  {$IFDEF FPC}
  {$IFDEF PANDORA}
  BUILDSTR := '0.820';
  HOMEFOLDER := GetCurrentDir;
  If Copy(HOMEFOLDER, Length(HOMEFOLDER), 1) <> '/' Then
     HOMEFOLDER := HOMEFOLDER + '/';
  HOMEFOLDER := HOMEFOLDER + 'home';
  {$ELSE}
  BUILDSTR := '0.820';
  HOMEFOLDER := GetEnvironmentVariable('HOME')+'/specbas';
  {$ENDIF}
  {$ELSE} // Windows
  BUILDSTR := '0.820';
  SHGetFolderPath(0,$0028,0,SHGFP_TYPE_CURRENT,@path[0]);
  HOMEFOLDER := Path + aString('\specbas');
  SetCurrentDir(String(HOMEFOLDER));
  HOMEFOLDER := Lower(HOMEFOLDER);
  If HOMEFOLDER[Length(HOMEFOLDER)] <> '\' Then HOMEFOLDER := HOMEFOLDER + '\';
  {$ENDIF}

  {$IF DEFINED(LINUX) OR DEFINED(RASPI)}
  If DirectoryExists('/usr/share/specbas/skel/specbas') and
     (Not DirectoryExists(HOMEFOLDER)) then
     CopyDir('/usr/share/specbas/skel/specbas/',HOMEFOLDER + '/');
  {$IFEND}

  If Not DirectoryExists(HOMEFOLDER) Then
    CreateDir(HOMEFOLDER);
  If Not DirectoryExists(HOMEFOLDER + '/temp') Then
    CreateDir(HOMEFOLDER + '/temp');
  TEMPDIR := HOMEFOLDER + '/temp/';

  {$IFDEF FPC}
  HOMEFOLDER := HOMEFOLDER + '/';
  {$ENDIF}

  HiddenCursorData := 0;
  HiddenCursor := SDL_CreateCursor(@hiddenCursorData, @hiddenCursorData, 8, 1, 0, 0);
  SDL_ShowCursor(SDL_ENABLE);
  SDL_SetCursor(HiddenCursor);

  AUTOSAVE := True;

  {$IFDEF BOLLOX}
  CurVidInfo := SDL_GetVideoInfo;
  ModeList := SDL_ListModes(CurVidInfo^.vfmt, SDL_OPENGL or SDL_HWSURFACE);

  While Assigned (ModeList^) do
        Begin
                { Don't allow modes wider than 800 pixels else the Pi can't keep up }
                If (ModeList^.w < 800) Then Begin
                         Break;
                End;
        Inc(ModeList^);
        End;

  ScrWidth := ModeList^.w;
  ScrHeight := ModeList^.h;
  {$ELSE}
  ScrWidth := 800;
  ScrHeight := 480;
  {$IFDEF RASPICONSOLE}
  ScrWidth := 720;
  ScrHeight := 480;
  {$ENDIF}
  {$ENDIF}

  // Initialise callbacks

  CB_GetKeyLockState := GetKeyState;
  CB_Refresh_Display := Refresh_Display;
  CB_Quit := Quit;
  CB_SetScreenRes := SetScreen;
  CB_Test_Resolution := TestScreenResolution;
  CB_GetTicks := GetTicks;
  CB_Yield := YieldProc;
  CB_Load_Image := LoadImage;
  CB_Free_Image := FreeImageResource;
  CB_GetAxis := GetAxis;
  CB_InitSticks := InitSticks;
  CB_ReleaseSticks := ReleaseSticks;

  SP_InitialGFXSetup(ScrWidth, ScrHeight);
  SetScreenResolution(ScrWidth, ScrHeight, False);

  // Launch the interpreter

  SP_CLS(CPAPER);
  EDITLINE := '';
  CURSORPOS := 0;
  CURSORCHAR := 32;
  SYSTEMSTATE := SS_IDLE;
  GetBatteryStatus;

  BASThread := TSpecBAS_Thread.Create(False);

End;

Procedure SetupKeys;
Var
  Idx: Integer;
Begin

  For Idx := 0 To 65535 Do ConvertKeys[Idx] := Idx And $FF;

  ConvertKeys[SDLK_BACKSPACE] := K_BACK;
  ConvertKeys[SDLK_TAB] := K_TAB;
  ConvertKeys[SDLK_RETURN] := K_RETURN;
  ConvertKeys[SDLK_KP_ENTER] := K_RETURN;
  ConvertKeys[SDLK_PAUSE] := K_PAUSE;
  ConvertKeys[SDLK_ESCAPE] := K_ESCAPE;
  ConvertKeys[SDLK_SPACE] := K_SPACE;
  ConvertKeys[SDLK_COMMA] := K_OEM_COMMA;
  ConvertKeys[SDLK_MINUS] := K_OEM_MINUS;
  ConvertKeys[SDLK_PERIOD] := K_OEM_PERIOD;
  ConvertKeys[SDLK_PLUS] := K_OEM_PLUS;
  ConvertKeys[SDLK_SLASH] := K_OEM_2;
  ConvertKeys[SDLK_0] := K_0;
  ConvertKeys[SDLK_1] := K_1;
  ConvertKeys[SDLK_2] := K_2;
  ConvertKeys[SDLK_3] := K_3;
  ConvertKeys[SDLK_4] := K_4;
  ConvertKeys[SDLK_5] := K_5;
  ConvertKeys[SDLK_6] := K_6;
  ConvertKeys[SDLK_7] := K_7;
  ConvertKeys[SDLK_8] := K_8;
  ConvertKeys[SDLK_9] := K_9;
  ConvertKeys[SDLK_SEMICOLON] := K_OEM_1;
  ConvertKeys[SDLK_EQUALS] := K_OEM_PLUS;
  ConvertKeys[SDLK_BACKSLASH] := K_OEM_5;
  ConvertKeys[SDLK_a] := K_A;
  ConvertKeys[SDLK_b] := K_B;
  ConvertKeys[SDLK_c] := K_C;
  ConvertKeys[SDLK_d] := K_D;
  ConvertKeys[SDLK_e] := K_E;
  ConvertKeys[SDLK_f] := K_F;
  ConvertKeys[SDLK_g] := K_G;
  ConvertKeys[SDLK_h] := K_H;
  ConvertKeys[SDLK_i] := K_I;
  ConvertKeys[SDLK_j] := K_J;
  ConvertKeys[SDLK_k] := K_K;
  ConvertKeys[SDLK_l] := K_L;
  ConvertKeys[SDLK_m] := K_M;
  ConvertKeys[SDLK_n] := K_N;
  ConvertKeys[SDLK_o] := K_O;
  ConvertKeys[SDLK_p] := K_P;
  ConvertKeys[SDLK_q] := K_Q;
  ConvertKeys[SDLK_r] := K_R;
  ConvertKeys[SDLK_s] := K_S;
  ConvertKeys[SDLK_t] := K_T;
  ConvertKeys[SDLK_u] := K_U;
  ConvertKeys[SDLK_v] := K_V;
  ConvertKeys[SDLK_w] := K_W;
  ConvertKeys[SDLK_x] := K_X;
  ConvertKeys[SDLK_y] := K_Y;
  ConvertKeys[SDLK_z] := K_Z;
  ConvertKeys[SDLK_DELETE] := K_DELETE;
  ConvertKeys[SDLK_KP0] := K_NUMPAD0;
  ConvertKeys[SDLK_KP1] := K_NUMPAD1;
  ConvertKeys[SDLK_KP2] := K_NUMPAD2;
  ConvertKeys[SDLK_KP3] := K_NUMPAD3;
  ConvertKeys[SDLK_KP4] := K_NUMPAD4;
  ConvertKeys[SDLK_KP5] := K_NUMPAD5;
  ConvertKeys[SDLK_KP6] := K_NUMPAD6;
  ConvertKeys[SDLK_KP7] := K_NUMPAD7;
  ConvertKeys[SDLK_KP8] := K_NUMPAD8;
  ConvertKeys[SDLK_KP9] := K_NUMPAD9;
  ConvertKeys[SDLK_KP_PERIOD] := K_DECIMAL;
  ConvertKeys[SDLK_KP_DIVIDE] := K_DIVIDE;
  ConvertKeys[SDLK_KP_MULTIPLY] := K_MULTIPLY;
  ConvertKeys[SDLK_KP_MINUS] := K_SUBTRACT;
  ConvertKeys[SDLK_KP_PLUS] := K_ADD;
  ConvertKeys[SDLK_UP] := K_UP;
  ConvertKeys[SDLK_DOWN] := K_DOWN;
  ConvertKeys[SDLK_RIGHT] := K_RIGHT;
  ConvertKeys[SDLK_LEFT] := K_LEFT;
  ConvertKeys[SDLK_INSERT] := K_INSERT;
  ConvertKeys[SDLK_HOME] := K_HOME;
  ConvertKeys[SDLK_END] := K_END;
  ConvertKeys[SDLK_PAGEUP] := K_PRIOR;
  ConvertKeys[SDLK_PAGEDOWN] := K_NEXT;
  ConvertKeys[SDLK_F1] := K_F1;
  ConvertKeys[SDLK_F2] := K_F2;
  ConvertKeys[SDLK_F3] := K_F3;
  ConvertKeys[SDLK_F4] := K_F4;
  ConvertKeys[SDLK_F5] := K_F5;
  ConvertKeys[SDLK_F6] := K_F6;
  ConvertKeys[SDLK_F7] := K_F7;
  ConvertKeys[SDLK_F8] := K_F8;
  ConvertKeys[SDLK_F9] := K_F9;
  ConvertKeys[SDLK_F10] := K_F10;
  ConvertKeys[SDLK_F11] := K_F11;
  ConvertKeys[SDLK_F12] := K_F12;
  ConvertKeys[SDLK_F13] := K_F13;
  ConvertKeys[SDLK_F14] := K_F14;
  ConvertKeys[SDLK_F15] := K_F15;
  ConvertKeys[SDLK_NUMLOCK] := K_NUMLOCK;
  ConvertKeys[SDLK_CAPSLOCK] := K_CAPITAL;
  ConvertKeys[SDLK_SCROLLOCK] := K_SCROLL;
  ConvertKeys[SDLK_RSHIFT] := K_SHIFT;
  ConvertKeys[SDLK_LSHIFT] := K_SHIFT;
  ConvertKeys[SDLK_RCTRL] := K_CONTROL;
  ConvertKeys[SDLK_LCTRL] := K_CONTROL;
  ConvertKeys[SDLK_RALT] := K_ALT;
  ConvertKeys[SDLK_LALT] := K_ALT;
  ConvertKeys[SDLK_LSUPER] := K_LWIN;
  ConvertKeys[SDLK_RSUPER] := K_RWIN;
  ConvertKeys[SDLK_MODE] := K_ALT;
  ConvertKeys[SDLK_HELP] := K_HELP;
  ConvertKeys[SDLK_PRINT] := K_PRINT;

  {$IFDEF PANDORA}
  ConvertKeys[308] := K_ALT;
  ConvertKeys[306] := K_CONTROL;
  {$ENDIF}

End;

// {$IFDEF WINDOWS}{$R SpecBAS.rc}{$ENDIF}

// {$R *.res}

{$R *.res}

begin

  InitTime := GetTickCount;
  {$IFDEF PANDORA}
  SDL_Init(SDL_INIT_VIDEO or SDL_INIT_TIMER or SDL_INIT_JOYSTICK);
  InitSticks;
  {$ELSE}
  SDL_Init(SDL_INIT_VIDEO or SDL_INIT_TIMER);
  {$ENDIF}
  SDL_EnableUNICODE(1);

  SP_Init_Sound;
  SetUpKeys;
  Init;
  FRAME_MS := 20;
  FPS := 50;
  LastTickCount := GetTickCount;
  TickCount := LastTickCount;

  While Not Quitting Do Begin

    Inc(FRAMES);

    SP_FrameUpdate;
    If CauseUpdate Then Begin
      CB_Refresh_Display;
      CauseUpdate := False;
    end;

    If ScrResChange Then Begin
      SetScreenResolution(ScrNewWidth, ScrNewHeight, ScrNewFull);
      ScrResChange := False;
    end;

    While SDL_PollEvent(@Event) > 0 Do Begin

      Case Event.type_ of
        SDL_KeyDown:
          Begin
            K_DOWNFLAG := True;
            Key := Event.Key.Keysym.sym;
            DEBUGVAL := Key;
            K_Flag := False;
            LASTKEYCHAR := #0;
            {$IFDEF SPECCYKEYS}
            If KEY = K_SHIFT Then
              KEYSTATE[K_SHIFT] := 1;

            If (Key in [K_5, K_6, K_7, K_8, K_0, K_SPACE]) And (KEYSTATE[K_SHIFT] = 1) Then Begin
              Case Key Of
                K_5: Key := K_LEFT;
                K_6: Key := K_DOWN;
                K_7: Key := K_UP;
                K_8: Key := K_RIGHT;
                K_0: Key := K_BACK;
                K_SPACE: Key := K_ESCAPE;
              End;
              KEYSTATE[K_SHIFT] := 0;
              K_Flag := True;
            End;
            If (LASTKEYCHAR > #127) Then Key := Key and $7f;
            {$ENDIF}

            If LASTKEYCHAR = #0 Then LASTKEYCHAR := aChar(Event.Key.Keysym.unicode);
            If Ord(LASTKEYCHAR[1]) >= $7F Then LASTKEYCHAR := #0;
            If Event.Key.Keysym.Unicode = 163 Then LASTKEYCHAR := '';
            If Key = 96 Then Begin Key := 127; LASTKEYCHAR := #127; End;
            {$IFDEF PANDORA}
            If Key in [64, 40, 41, 33, 95, 39, 34, 45, 43, 61, 163, 59, 58, 47, 63, 92, 124, 35, 36, SDLK_0..SDLK_9] Then Begin
              Case Key of
                64:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_OEM_3); End;
                40:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_9); End;
                41:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_0); End;
                33:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_1); End;
                95:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_OEM_MINUS); End;
                39:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_3); End;
                34:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_2); End;
                45:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_MINUS); End;
                43:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_OEM_PLUS);  End;
                61:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_PLUS); End;
                163:    Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_3); End;
                59:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_1); End;
                58:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_OEM_1); End;
                47:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_2); End;
                63:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_OEM_2); End;
                92:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_5); End;
                124:    Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_OEM_5); End;
                35:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyDown(K_OEM_7); End;
                36:     Begin KEYSTATE[K_SHIFT] := 1; SP_KeyDown(K_4); End;
                SDLK_1: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 32; SP_KeyDown(Key); End;
                SDLK_2: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 219; SP_KeyDown(Key); End;
                SDLK_3: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 221; SP_KeyDown(Key); End;
                SDLK_4: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 222; SP_KeyDown(Key); End;
                SDLK_5: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_5; SP_KeyDown(ConvertKeys[Key]); End;
                SDLK_6: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_6; SP_KeyDown(ConvertKeys[Key]); End;
                SDLK_7: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_7; SP_KeyDown(ConvertKeys[Key]); End;
                SDLK_8: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_8; SP_KeyDown(ConvertKeys[Key]); End;
                SDLK_9: Begin If KEYSTATE[K_SHIFT] = 1 Then Begin KEYSTATE[K_SHIFT] := 0; Key := 219; End; SP_KeyDown(Key); End;
                SDLK_0: Begin If KEYSTATE[K_SHIFT] = 1 Then Begin KEYSTATE[K_SHIFT] := 0; Key := 221; End; SP_KeyDown(Key); End;
              End;
            End Else
              SP_KeyDown(ConvertKeys[Key]);
            {$ELSE}
            SP_KeyDown(ConvertKeys[Key]);
            {$ENDIF}
            If K_Flag Then
              KEYSTATE[K_SHIFT] := 1;
          End;
        SDL_KeyUp:
          Begin
            K_UPFLAG := True;
            Key := Event.Key.Keysym.sym;
            LASTKEYCHAR := #0;
            {$IFDEF PANDORA}
            If Key in [64, 40, 41, 33, 95, 39, 34, 45, 43, 61, 163, 59, 58, 47, 63, 92, 124, 35, 36, SDLK_0..SDLK_9] Then Begin
              Case Key of
                64:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_3); End;
                40:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_9); End;
                41:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_0); End;
                33:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_1); End;
                95:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_MINUS); End;
                39:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(192); End;
                34:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_2); End;
                45:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_MINUS); End;
                43:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_PLUS);  End;
                61:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_PLUS); End;
                163:    Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_3); End;
                59:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_1); End;
                58:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_1); End;
                47:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_2); End;
                63:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_2); End;
                92:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_5); End;
                124:    Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_5); End;
                35:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_OEM_7); End;
                36:     Begin KEYSTATE[K_SHIFT] := 0; SP_KeyUp(K_4); End;
                SDLK_1: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 32; SP_KeyUp(Key); End;
                SDLK_2: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 219; SP_KeyUp(Key); End;
                SDLK_3: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 221; SP_KeyUp(Key); End;
                SDLK_4: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := 222; SP_KeyUp(Key); End;
                SDLK_5: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_5; SP_KeyUp(ConvertKeys[Key]); End;
                SDLK_6: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_6; SP_KeyUp(ConvertKeys[Key]); End;
                SDLK_7: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_7; SP_KeyUp(ConvertKeys[Key]); End;
                SDLK_8: Begin If KEYSTATE[K_SHIFT] = 1 Then Key := SDLK_8; SP_KeyUp(ConvertKeys[Key]); End;
                SDLK_9: Begin If KEYSTATE[K_SHIFT] = 1 Then Begin Key := 219; End; SP_KeyUp(Key); End;
                SDLK_0: Begin If KEYSTATE[K_SHIFT] = 1 Then Begin Key := 221; End; SP_KeyUp(Key); End;
              End;
            End Else
              SP_KeyUp(ConvertKeys[Key]);
            {$ELSE}
            SP_KeyUp(ConvertKeys[Key]);
            {$ENDIF}
          End;
        SDL_QuitEv:
          Begin
            Quitting := True;
          end;
        {$IFDEF PANDORA}
        SDL_JOYAXISMOTION:
          Begin
          End;
        {$ENDIF}
        SDL_MouseButtonDown:
          Begin
            Button := Event.button.button;
            If Button in [1, 2, 3] Then Begin
              If Button = 1 Then MOUSEBTN := MOUSEBTN or 1; // Left
              If Button = 2 Then MOUSEBTN := MOUSEBTN or 4; // Middle
              If Button = 3 Then MOUSEBTN := MOUSEBTN or 2; // Right
              M_DOWNFLAG := True;
            End Else
              If Button = SDL_BUTTON_WHEELUP Then Begin
                M_WHEELUPFLAG := True;
                Dec(MouseWheel);
              End Else
                If Button = SDL_BUTTON_WHEELDOWN Then Begin
                  M_WHEELDNFLAG := True;
                  Dec(MouseWheel);
                End;
          end;
        SDL_MouseButtonUp:
          Begin
            Button := Event.button.button;
            If Button in [1, 2, 3] Then Begin
              If Button = 1 Then MOUSEBTN := MOUSEBTN And 254; // or 1
              If Button = 2 Then MOUSEBTN := MOUSEBTN And 251; // or 4
              If Button = 3 Then MOUSEBTN := MOUSEBTN And 253; // or 2
              M_UPFLAG := True;
            End;
          end;
        SDL_MouseMotion:
          Begin
            M_DELTAX := Event.motion.x - MOUSEX;
            M_DELTAY := Event.motion.y - MOUSEY;
            M_MOVEFLAG := True;
            MOUSEX := Event.motion.x;
            MOUSEY := Event.motion.y;
            If MOUSEVISIBLE Then
              SP_NeedDisplayUpdate := True;
          end;
      End;

    end;

    TickCount := FRAME_MS - (GetTickCount - LastTickCount);
    If TickCount >= 0 Then
      SDL_Delay(TickCount);

    Inc(LastTickCount, FRAME_MS);

  End;

  {$IFDEF PANDORA}
  SDL_JoystickClose(Nub1);
  SDL_JoystickClose(Nub2);
  {$ENDIF}
  SDL_FreeCursor(hiddenCursor);
  SDL_Quit;
  BASS_Free;

end.

